#include "AnimCurve.h"
#include "TextureTools.h"
#include <string.h>

//
//  Attribute index definitions
//
int AnimCurve::kTimes;
int AnimCurve::kInterpType;
int AnimCurveScalar::kKnots;
int AnimCurveVec2::kKnots;
int AnimCurveVec3::kKnots;

//
//  Strings for interpolation types
//
static char *interpTypeNames[] = { "LINEAR",
									"SPLINE",
									NULL
};


/******************************************************************
					AnimCurve
*******************************************************************/

//
//  Constructor just initializes attributes
//
AnimCurve::AnimCurve()
{
	AnimCurve::initializeAttributes();
}

//
//  Attribute initialization
//
void AnimCurve::initializeAttributes()
{
	NamedEntity::initializeAttributes();

	kTimes = addAttr( new AttributeDoubleList( "times", CONSTANT ) );
	kInterpType = addAttr( new AttributeEnum( "interp", CONSTANT, kLinear, interpTypeNames, kNumInterpTypes ) );
}

//
//  Checks to make sure keys are ascending, and if spline interpolation is
//  being used, makes sure that a valid number of keys is specified.
//
bool AnimCurve::postReadFunc( const Scene& scene )
{
	if( !NamedEntity::postReadFunc(scene) ) {
		return false;
	}

	// make sure keys are ascending
	List<double>& times = attrTimes();
	double prev = times[0];

	for( int i = 0; i < times.size(); i++ ) {
		if( times[i] > prev ) {
			cout << "Animcurve " << attrName() << ": non-ascending times" << endl;
			return false;
		}
		prev = times[i];
	}

	if( attrInterpType() == kSpline ) {
		if( attrTimes().size() < 4 ) {
			cout << "ERROR - need at least 4 knots for spline interp" << endl;
			return false;
		}
	}
	return true;
}

//
//  Template function to evaluate an animation curve at a particular time.
//
template <class T> T evalAnimCurve( double time, List<double>& times, List<T>& knots, InterpType interp )
{

	KeyLocType where;
	double ratio;
	int low, high;

	// find where the desired time is relative to the specified keys
	if( time < times[0] ) {
		where = kBefore;
	} else if( time > times[times.size()-1] ) {
		where = kAfter;
	} else {

		if( interp == kLinear ) {

			// find the enclosing span and the affine coords of the 
			// desired time in that span
			low = 0;

			while( (low < times.size()-1) && (times[low+1] < time) ) {
				low++;
			}

			// time is between times[low] and times[low+1]
			high = low+1;
			ratio = (time-times[low])/(times[high]-times[low]);

		} else {

			// spline interpolation - find ratio of current time int
			// knot vector
			ratio = (time-times[1])/(times[times.size()-2]-times[0]);
		}

		where = kDuring;
	}


	switch( where ) {

	case kBefore:

		// for times before the first time, return the first knot value
		return knots[0];
		break;

	case kAfter:

		// for times after the last time, return the last knot value
		return knots[knots.size()-1];
		break;

	case kDuring:

		if( interp == kLinear ) {

			// linear interpolation, just linearly interpolate knot values
			return (1.0-ratio)*knots[low] + ratio*knots[high];

		} else {

			// spline interpolation, use spline function
			T result = spline<T>( ratio, knots.size(), knots.subList(0) );
			return result;
		}
		break;

	default:
		assert(false);
		return T();
		break;
	}
}

/******************************************************************
					AnimCurveScalar
*******************************************************************/
AnimCurveScalar::AnimCurveScalar()
{
	AnimCurveScalar::initializeAttributes();
}

void AnimCurveScalar::initializeAttributes()
{
	AnimCurve::initializeAttributes();

	kKnots = addAttr( new AttributeDoubleList( "knots", CONSTANT ) );
}

double AnimCurveScalar::evalDouble()
{
	return evalAnimCurve<double>( Attribute::curTime(), attrTimes(), attrKnots(), attrInterpType() );
}

/******************************************************************
					AnimCurveVec2
*******************************************************************/
AnimCurveVec2::AnimCurveVec2()
{
	AnimCurveVec2::initializeAttributes();
}

void AnimCurveVec2::initializeAttributes()
{
	AnimCurve::initializeAttributes();

	kKnots = addAttr( new AttributeVec2List( "knots", CONSTANT ) );
}

vec2 AnimCurveVec2::evalVec2()
{
	return evalAnimCurve<vec2>( Attribute::curTime(), attrTimes(), attrKnots(), attrInterpType() );
}

/******************************************************************
					AnimCurveVec3
*******************************************************************/
AnimCurveVec3::AnimCurveVec3()
{
	AnimCurveVec3::initializeAttributes();
}

void AnimCurveVec3::initializeAttributes()
{
	AnimCurve::initializeAttributes();

	kKnots = addAttr( new AttributeVec3List( "knots", CONSTANT ) );
}

vec3 AnimCurveVec3::evalVec3()
{
	return evalAnimCurve<vec3>( Attribute::curTime(), attrTimes(), attrKnots(), attrInterpType() );
}

