#ifndef AnimCurve_h
#define AnimCurve_h

#include "NamedEntity.h"

// types of interpolation used in AnimCurves
typedef enum { kLinear=0, kSpline, kNumInterpTypes } InterpType;

// places where a key can be located relative to the keys for the 
// AnimCurve
typedef enum { kBefore, kDuring, kAfter } KeyLocType;

/****************************************************************

	Class:		AnimCurve
	Purpose:	Base class for all animation curves (double, vec2, 
				vec3).  Defines attributes and evaluation methods
				common to all animation curves.

				An AnimCurve is just a time-dependant function
				specified by a set of keys of the form (time,value)
				and an interpolation method.

*****************************************************************/
class AnimCurve : public NamedEntity
{
public:

	AnimCurve();

	// attribute initialization
	void initializeAttributes();

	// attribute indices 
	static int kTimes;
	static int kInterpType;

	// attribute accessors
	List<double>& attrTimes() { return findAttr(kTimes)->asDoubleList(); };
	InterpType    attrInterpType() { return (InterpType)findAttr(kInterpType)->asInt(); };

	// checks to make sure the specified curve is valid
	bool postReadFunc( const Scene& scene );

	// animcurve evaluators
	virtual double  evalDouble() { assert(false);  return *(double *)NULL;	};
	virtual vec2	evalVec2() { assert(false);  return *(vec2 *)NULL;		};
	virtual vec3    evalVec3() { assert(false);  return *(vec3 *)NULL;		};

};

/****************************************************************

	Class:		AnimCurveScalar
	Purpose:	Animation curve for scalar (double, int) values.	

*****************************************************************/
class AnimCurveScalar : public AnimCurve
{
public:

	AnimCurveScalar();

	// attribute initialization
	void initializeAttributes();

	// attribute indices
	static int kKnots;

	// attribute accessors
	List<double>&	attrKnots() { return findAttr(kKnots)->asDoubleList(); };

	// animcurve evaluator
	virtual double evalDouble();

};

/****************************************************************

	Class:		AnimCurveVec2
	Purpose:	Animation curve for vec2 values.	

*****************************************************************/
class AnimCurveVec2 : public AnimCurve
{
public:

	AnimCurveVec2();

	// attribute initialization
	void initializeAttributes();

	// attribute indices
	static int kKnots;

	// attribute accessors
	List<vec2>&	attrKnots() { return findAttr(kKnots)->asVec2List(); };

	// animcurve evaluator
	virtual vec2 evalVec2();

};

/****************************************************************

	Class:		AnimCurveVec3
	Purpose:	Animation curve for vec3 values.	

*****************************************************************/
class AnimCurveVec3 : public AnimCurve
{
public:

	AnimCurveVec3();

	// attribute initialization
	void initializeAttributes();

	// attribute indices
	static int kKnots;

	// attribute accessors
	List<vec3>&	attrKnots() { return findAttr(kKnots)->asVec3List(); };

	// animcurve evaluator
	virtual vec3 evalVec3();

};

#endif