#include <iostream.h>
#include <ctype.h>
#include <string.h>
#include "Scene.h"
#include "Attributes.h"
#include "Texture.h"
#include "AnimCurve.h"
#include "algebra3.h"

Sample Attribute::fCurSample;
double Attribute::fCurTime = -999.9;	// initialize cur time to -999.9


/*****************************************************************
						Attribute
******************************************************************/

//
//  Intialize attributes to dirty
//
Attribute::Attribute( const char *n, DrivenCode d )
: fDriveable(d), fDriverName(NULL), fTexture(NULL), fAnimCurve(NULL), fIsDirty(true)
{
	fName = new char[strlen(n)+1];
	strcpy( fName, n );
}

//
//  Read the attribute from a file.  Textured and animated attributes can be
//  resolved here, in which case DRIVEN is returned.  Otherwise, GETITYOURSELF
//  is returned to indicate that the attribute type-specific read method must
//  be invoked to read the value of the attribute.
//
ReadCode Attribute::read( istream& s ) 
{
	// if no more input, error
	if( s.eof() ) {
		return ERROR;
	}

	if( (fDriveable == TEXTURABLE) || (fDriveable == ANIMATABLE) ) {
		// if driveable, see if it is a texture
		char temp[100];
		char c;
		eatWS(s);

		// get the first char of the value, textures must be of the
		// form <texturename>
		c = s.peek();

		if( c != '<' ) {
			// not a texture, attr type must read the value itself
			return GETITYOURSELF;

		} else {

			// this should be a texture or animCurve
			s >> c;
			s.get( temp, 100, '>' );

			// make sure last char is '>'
			if( s.get() != '>' ) {
				return ERROR;

			} else {
				// texture is valid, so set the texture name field of the
				// attribute, to be resolved later once the whole scene is
				// read in
				fDriverName = new char[strlen(temp)+1];
				strcpy( fDriverName, temp );
				return DRIVEN;
			}
		}

	} else {
		return GETITYOURSELF;
	}
}


void Attribute::write( ostream& s )
{
	s << fName << " ";
}

//
//  postReadFunc tries to resolve the driver for an attribute from
//  the driver name that was read earlier.
//
bool Attribute::postReadFunc( const Scene& scene )
{
	bool result = false;

	if( fDriverName != NULL ) {

		// see if it is a texture 
		Texture *tex = scene.findTexture( fDriverName );

		// see if it is an anim curve
		AnimCurve *anim = scene.findAnimCurve( fDriverName );
		if( tex != NULL ) {
			fDriven = TEXTURABLE;
			fTexture = tex;
			result = true;
		} else if( anim != NULL ) {
			fDriven = ANIMATABLE;
			fAnimCurve = anim;
		    result = true;
		} 

		else {
			cout << "Can't resolve driver " << fDriverName << endl;
			result = false;
		}
	} else {
		fDriven = CONSTANT;
		result = true;
	}

	return result;
}

/*****************************************************************
						AttributeBool
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeBool::AttributeBool( const char *name, DrivenCode driveable, bool defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
	DB_ASSERT( (driveable == CONSTANT), "ERROR - cannot drive boolean attr" );
}

//
//	Evaluates a boolean attribute
//
bool& AttributeBool::asBool()
{
	return( fValue );
}

//
//	Read the boolean value of an attribute
//
ReadCode AttributeBool::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		int temp;
		s >> temp;
		fValue = temp ? true : false;
	}

	return result;
}

void AttributeBool::write( ostream& s )
{
	Attribute::write( s );
	s << (fValue ? "true" : "false");
}


/*****************************************************************
						AttributeInt
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeInt::AttributeInt( const char *name, DrivenCode driveable, int defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
}

//
//	Evaluates an integer attribute
//
int& AttributeInt::asInt()
{
	switch( fDriven ) {

	case CONSTANT:
		break;

	case ANIMATABLE:

		if( fIsDirty ) {
			// evaluate the AnimCurve
			fValue = fastDblToInt(fAnimCurve->evalDouble());
		}
		fIsDirty = false;
		break;

	case TEXTURABLE:

		assert( fTexture != NULL );
		fValue = fastDblToInt(fTexture->scalar());
		break;

	}

	return fValue;
}

//
//	Read the boolean value of an attribute
//
ReadCode AttributeInt::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		s >> fValue;
	}

	return result;
}

void AttributeInt::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}


/*****************************************************************
						AttributeDouble
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeDouble::AttributeDouble( const char *name, DrivenCode driveable, double defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
}

//
//	Evaluates an integer attribute
//
double& AttributeDouble::asDouble()
{
	switch( fDriven ) {

	case CONSTANT:
		break;

	case ANIMATABLE:

		if( fIsDirty ) {
			// evaluate the AnimCurve
			fValue = fAnimCurve->evalDouble();
		}
		fIsDirty = false;
		break;

	case TEXTURABLE:

		assert( fTexture != NULL );
		fValue = fTexture->scalar();
		break;

	}

	return fValue;
}

//
//	Read the boolean value of an attribute
//
ReadCode AttributeDouble::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		s >> fValue;
	}

	return result;
}

void AttributeDouble::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}

/*****************************************************************
						AttributeVec2
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeVec2::AttributeVec2( const char *name, DrivenCode driveable, const vec2& defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
}

//
//	Evaluates a vec2 attribute
//
vec2& AttributeVec2::asVec2()
{
	switch( fDriven ) {

	case CONSTANT:
		break;

	case ANIMATABLE:

		if( fIsDirty ) {
			// evaluate the AnimCurve
			fValue = fAnimCurve->evalVec2();
		}
		fIsDirty = false;
		break;

	case TEXTURABLE:

		assert( fTexture != NULL );
		fValue = vec2( fTexture->vector(), VZ );
		break;

	}

	return fValue;
}

//
//	Read the vec2 value of an attribute
//
ReadCode AttributeVec2::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		s >> fValue;
	}

	return result;
}

void AttributeVec2::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}

/*****************************************************************
						AttributeVec3
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeVec3::AttributeVec3( const char *name, DrivenCode driveable, const vec3& defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
}

//
//	Evaluates a vec3 attribute
//
vec3& AttributeVec3::asVec3()
{
	switch( fDriven ) {

	case CONSTANT:
		break;

	case ANIMATABLE:

		if( fIsDirty ) {
			// evaluate the AnimCurve
			fValue = fAnimCurve->evalVec3();
		}
		fIsDirty = false;
		break;

	case TEXTURABLE:

		assert( fTexture != NULL );
		fValue = fTexture->vector();
		break;

	}

	return fValue;
}


//
//	Read the vec3 value of an attribute
//
ReadCode AttributeVec3::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		s >> fValue;
	}

	return result;
}

void AttributeVec3::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}


/*****************************************************************
						AttributeVec4
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeVec4::AttributeVec4( const char *name, DrivenCode driveable, const vec4& defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
	DB_ASSERT( (driveable == CONSTANT), "ERROR - cannot drive vec4 attr" );
}

//
//	Evaluates a vec4 attribute
//
vec4& AttributeVec4::asVec4()
{
	return( fValue );
}

//
//	Read the vec3 value of an attribute
//
ReadCode AttributeVec4::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		s >> fValue;
	}

	return result;
}

void AttributeVec4::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}

/*****************************************************************
						AttributeMat3
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeMat3::AttributeMat3( const char *name, DrivenCode driveable, const mat3& defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
	DB_ASSERT( (driveable == CONSTANT), "ERROR - cannot drive mat3 attr" );
}

//
//	Evaluates a mat3 attribute
//
mat3& AttributeMat3::asMat3()
{
	return( fValue );
}

//
//	Read the mat3 value of an attribute
//
ReadCode AttributeMat3::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		result = read2dTransform(s,fValue) ? result : ERROR;
	}

	return result;
}

void AttributeMat3::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}

/*****************************************************************
						AttributeMat4
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeMat4::AttributeMat4( const char *name, DrivenCode driveable, const mat4& defaultVal )
: Attribute( name, driveable ), fValue(defaultVal)
{
	DB_ASSERT( (driveable == CONSTANT), "ERROR - cannot drive mat4 attr" );
}

//
//	Evaluates a mat4 attribute
//
mat4& AttributeMat4::asMat4()
{
	return( fValue );
}

//
//	Read the mat4 value of an attribute
//
ReadCode AttributeMat4::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		result = read3dTransform(s,fValue) ? result : ERROR;
	}

	return result;
}

void AttributeMat4::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}

/*****************************************************************
						AttributeDoubleList
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeDoubleList::AttributeDoubleList( const char *name, DrivenCode driveable )
: Attribute( name, driveable ), fValue(1)
{
}

//
//	Evaluates a vec2List attribute
//
List<double>& AttributeDoubleList::asDoubleList()
{
	return( fValue );
}

//
//	Read the vec2 list value of an attribute
//
ReadCode AttributeDoubleList::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		result = readList<double>(s,fValue) ? result : ERROR;
		fValue.shrinkWrap();
	}

	return result;
}

void AttributeDoubleList::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}


/*****************************************************************
						AttributeVec2List
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeVec2List::AttributeVec2List( const char *name, DrivenCode driveable )
: Attribute( name, driveable ), fValue(1)
{
}

//
//	Evaluates a vec2List attribute
//
List<vec2>& AttributeVec2List::asVec2List()
{
	return( fValue );
}

//
//	Read the vec2 list value of an attribute
//
ReadCode AttributeVec2List::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		result = readList<vec2>(s,fValue) ? result : ERROR;
		fValue.shrinkWrap();
	}

	return result;
}

void AttributeVec2List::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}

/*****************************************************************
						AttributeVec3List
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeVec3List::AttributeVec3List( const char *name, DrivenCode driveable )
: Attribute( name, driveable ), fValue(1)
{
	DB_ASSERT( (driveable == CONSTANT), "ERROR - cannot drive vec3 list attr" );
}

//
//	Evaluates a vec3List attribute
//
List<vec3>& AttributeVec3List::asVec3List()
{
	return( fValue );
}

//
//	Read the vec3 list value of an attribute
//
ReadCode AttributeVec3List::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		result = readList<vec3>(s,fValue) ? result : ERROR;
		fValue.shrinkWrap();
	}

	return result;
}

void AttributeVec3List::write( ostream& s )
{
	Attribute::write( s );
	s << fValue;
}

/*****************************************************************
						AttributeString
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeString::AttributeString( const char *name, DrivenCode driveable, const char *defaultVal )
: Attribute( name, driveable )
{
	fValue = new char[strlen(defaultVal)];
	strcpy( fValue, defaultVal );
}

//
//	Evaluates a string attribute
//
char *&AttributeString::asString()
{
	return( fValue );
}

//
//	Read the string value of an attribute
//
ReadCode AttributeString::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		char temp[100];
		eatWS(s);
		s >> temp;
		
		if( strlen(temp) > 0 ) {
			fValue = new char[strlen(temp)+1];
			strcpy( fValue, temp );
		} else {
			result = ERROR;
		}
	}

	return result;
}

void AttributeString::write( ostream& s )
{
	Attribute::write( s );
	s << (fValue ? fValue : "");
}

/*****************************************************************
						AttributeIntListList
******************************************************************/

//
//	Constructor - just call Attribute constructor
//
AttributeIntListList::AttributeIntListList( const char *name, DrivenCode driveable )
: Attribute( name, driveable )
{
	DB_ASSERT( (driveable == CONSTANT), "ERROR - cannot drive int list list attr" );
}

//
//	Evaluates an IntListList attribute
//
IntListList& AttributeIntListList::asIntListList()
{
	return( fValue );
}

//
//	Read the string value of an attribute
//
ReadCode AttributeIntListList::read( istream& s ) 
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		result = fValue.read(s) ? result : ERROR;
		fValue.shrinkWrap();
	}
	return result;
}

void AttributeIntListList::write( ostream& s )
{
	Attribute::write( s );
	fValue.write(s);
}


/*****************************************************************
						AttributeEnum
******************************************************************/

AttributeEnum::AttributeEnum( const char *name, DrivenCode driveable, int defaultVal, char **keys, int numKeys )
: AttributeInt( name, driveable, defaultVal ), fKeys(keys), fNumKeys(numKeys)
{
	DB_ASSERT( (driveable == CONSTANT), "ERROR - cannot drive enumerated attr" );
}

ReadCode AttributeEnum::read( istream& s )
{
	ReadCode result = Attribute::read(s);

	if( result == GETITYOURSELF ) {
		char temp[100];
		s >> temp;

		for( int i = 0; i < fNumKeys; i++ ) {
			if( stricmp( temp, fKeys[i] ) == 0 ) {
				fValue = i;
				break;
			}
		}

		if( i == fNumKeys ) {
			return ERROR;
		} else {
			return GETITYOURSELF;
		}
	} else {

		return result;

	}
}

void AttributeEnum::write( ostream& s )
{
	Attribute::write(s);
	if( fValue < fNumKeys ) {
		s << fKeys[fValue];
	} else {
		s << "INVALID VALUE";
	}
}