#ifndef Attributes_h
#define Attributes_h

#include "List.h"
#include "algebra3.h"
#include "IntListList.h"
#include "Sample.h"

class Texture;
class AnimCurve;
class Scene;

// result codes for attribute reads
typedef enum {ERROR, DRIVEN, GETITYOURSELF} ReadCode;

// codes for driving attributes - attributes can be constant, animated,
// or textured
typedef enum {CONSTANT, ANIMATABLE, TEXTURABLE} DrivenCode;

/****************************************************************

	Class:		Attribute
	Purpose:	Represents an attribute of an entity.  Attributes
				are stored in the scene file format.

				In theory, all attributes can be constant, animated,
				or textured.  Animation and texturing actually only
				work on double, vec2, and vec3 attributes.

				When an attribute is created, you specify whether they
				can be animated or textured.  When the attribute is
				read from a file, it is determined whether the attribute
				is constant, animated, or textured.

*****************************************************************/

class Attribute
{
public:

	Attribute( const char *name, DrivenCode driveable );

	char *getName() const { return fName; };

	// value accessors - use one of these methods depending on the type
	// of the attribute.  If the attribute is textured or animated, the
	// static animation time and shading sample data is used to evaluate
	// the animation curve or texture that is driving the attribute.  Animation
	// curves are only re-evaluated if the 'fDirty' flag is true.
	virtual bool&			asBool()		{ assert(false); return *(bool *)		NULL;	};
	virtual int&			asInt()			{ assert(false); return *(int *)		NULL;	};
	virtual double&			asDouble()		{ assert(false); return *(double *)		NULL;	};
	virtual vec2&			asVec2()		{ assert(false); return *(vec2 *)		NULL;	};
	virtual vec3&			asVec3()		{ assert(false); return *(vec3 *)		NULL;	};
	virtual vec4&			asVec4()		{ assert(false); return *(vec4 *)		NULL;	};
	virtual mat3&			asMat3()		{ assert(false); return *(mat3 *)		NULL;	};
	virtual mat4&			asMat4()		{ assert(false); return *(mat4 *)		NULL;	};
	virtual List<double>&	asDoubleList()  { assert(false); return *(List<double> *)NULL;  };
	virtual List<vec2>&		asVec2List()	{ assert(false); return *(List<vec2> *)	NULL;	};
	virtual List<vec3>&		asVec3List()	{ assert(false); return *(List<vec3> *)	NULL;	};
	virtual char *&			asString()		{ assert(false); return *(char **)		NULL;	};
	virtual IntListList&	asIntListList()	{ assert(false); return *(IntListList *)NULL;	};

	// read/write the attribute
	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

	// sets the current shading sample and current animation time
	static void setCurSample( const Sample& newSample )		{ fCurSample = newSample; };
	static void setCurTime( double newTime )				{ fCurTime = newTime; };

	// accessors for the current shading sample and current animation time
	static Sample& curSample() { return fCurSample; };
	static double  curTime()   { return fCurTime; };

	// for resolving textures and animCurves
	bool postReadFunc( const Scene& scene );

	void setDirty() { fIsDirty = true; };
	bool isConstant() { return (fDriven == CONSTANT); };
	bool isAnimated() { return (fDriven == ANIMATABLE); };
	bool isTextured() { return (fDriven == TEXTURABLE); };

	// comparison operator for Attributes
	int operator==( const char *rhs ) { return ( strcmp(fName,rhs) == 0 ); };


protected:

	char *fName;			// name of the attribute;

	bool fIsDirty;			// for animated attributes, does this attribute need to be
							// re-evaluated?  Animated attributes need only be evaluated
							// once per frame

	DrivenCode fDriveable;	// can this attribute be textured and/or animated?
	DrivenCode fDriven;		// is this attribute actually textured or animated?

	char *fDriverName;		// name of driving entity (texture or AnimCurve) - read
							// from file

	Texture *fTexture;		// driving texture
	AnimCurve *fAnimCurve;	// driving AnimCurve

	static Sample fCurSample;	// current shading sample - global data for all attributes
	static double fCurTime;		// current animation time - global data for all attributes
};



/****************************************************************

	Class:		AttributeBool
	Purpose:	Boolean-valued attribute class.	

*****************************************************************/

class AttributeBool : public Attribute
{
public:

	AttributeBool( const char *name, DrivenCode driveable, bool defaultVal );

	virtual bool& asBool();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	bool fValue;
};


/****************************************************************

	Class:		AttributeInt
	Purpose:	Integer-valued attribute class.	

*****************************************************************/

class AttributeInt : public Attribute
{
public:

	AttributeInt( const char *name, DrivenCode driveable, int defaultVal );

	virtual int& asInt();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	int fValue;
};

/****************************************************************

	Class:		AttributeDouble
	Purpose:	Double-valued attribute class.	

*****************************************************************/

class AttributeDouble : public Attribute
{
public:

	AttributeDouble( const char *name, DrivenCode driveable, double defaultVal );

	virtual double& asDouble();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	double fValue;
};

/****************************************************************

	Class:		AttributeVec2
	Purpose:	vec2-valued attribute class.	

*****************************************************************/

class AttributeVec2 : public Attribute
{
public:

	AttributeVec2( const char *name, DrivenCode driveable, const vec2& defaultVal );

	virtual vec2& asVec2();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	vec2 fValue;
};


/****************************************************************

	Class:		AttributeVec3
	Purpose:	vec3-valued attribute class.	

*****************************************************************/

class AttributeVec3 : public Attribute
{
public:

	AttributeVec3( const char *name, DrivenCode driveable, const vec3& defaultVal );

	virtual vec3& asVec3();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	vec3 fValue;
};


/****************************************************************

	Class:		AttributeVec4
	Purpose:	vec3-valued attribute class.	

*****************************************************************/

class AttributeVec4 : public Attribute
{
public:

	AttributeVec4( const char *name, DrivenCode driveable, const vec4& defaultVal );

	virtual vec4& asVec4();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	vec4 fValue;
};


/****************************************************************

	Class:		AttributeMat3
	Purpose:	mat3-valued attribute class.	

*****************************************************************/

class AttributeMat3 : public Attribute
{
public:

	AttributeMat3( const char *name, DrivenCode driveable, const mat3& defaultVal );

	virtual mat3& asMat3();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	mat3 fValue;
};


/****************************************************************

	Class:		AttributeMat4
	Purpose:	mat4-valued attribute class.	

*****************************************************************/

class AttributeMat4 : public Attribute
{
public:

	AttributeMat4( const char *name, DrivenCode driveable, const mat4& defaultVal );

	virtual mat4& asMat4();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	mat4 fValue;
};

/****************************************************************

	Class:		AttributeDoubleList
	Purpose:	List<double>-valued attribute class.	

*****************************************************************/

class AttributeDoubleList : public Attribute
{
public:

	AttributeDoubleList( const char *name, DrivenCode driveable );

	virtual List<double>& asDoubleList();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	List<double> fValue;
};

/****************************************************************

	Class:		AttributeVec2List
	Purpose:	List<vec2>-valued attribute class.	

*****************************************************************/

class AttributeVec2List : public Attribute
{
public:

	AttributeVec2List( const char *name, DrivenCode driveable );

	virtual List<vec2>& asVec2List();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	List<vec2> fValue;
};

/****************************************************************

	Class:		AttributeVec3List
	Purpose:	List<vec3>-valued attribute class.	

*****************************************************************/

class AttributeVec3List : public Attribute
{
public:

	AttributeVec3List( const char *name, DrivenCode driveable );

	virtual List<vec3>& asVec3List();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	List<vec3> fValue;
};


/****************************************************************

	Class:		AttributeString
	Purpose:	String-valued attribute class.	

*****************************************************************/

class AttributeString : public Attribute
{
public:

	AttributeString( const char *name, DrivenCode driveable, const char *defaultVal );

	virtual char *& asString();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	char *fValue;
};


/****************************************************************

	Class:		AttributeIntListList
	Purpose:	IntListList-valued attribute class.	

*****************************************************************/

class AttributeIntListList : public Attribute
{
public:

	AttributeIntListList( const char *name, DrivenCode driveable );

	virtual IntListList& asIntListList();

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	IntListList fValue;
};

/****************************************************************

	Class:		AttributeEnum
	Purpose:	Enumeration-valued attribute class.	

*****************************************************************/

class AttributeEnum : public AttributeInt
{
public:

	AttributeEnum( const char *name, DrivenCode driveable, int defaultVal, char **keys, int numKeys );

	virtual ReadCode read( istream& s );
	virtual void write( ostream& s );

protected:

	char **fKeys;
	int fNumKeys;
};

#endif