#include "BoundingBox.h"
#include "TriageMask.h"
#include "algebra3.h"
#include "ostream.h"


//
//  Constructors
//
BoundingBox::BoundingBox() : low(0.0), high(0.0), init(false)
{
}

BoundingBox::BoundingBox( const vec3& l, const vec3& h )
: low(l), high(h), init(true)
{
}

BoundingBox::BoundingBox( const BoundingBox& bb ) 
: low(bb.low), high(bb.high), init(bb.init)
{
}

//
//  Expand the box to fit the new point
//
void BoundingBox::expand( vec3& newPoint )
{
	if( init ) {
		low = min( low, newPoint );
		high = max( high, newPoint );
	} else {
		low = high = newPoint;
		init = true;
	}
}

//
//  Take the union of this box with the new one
//
void BoundingBox::expand( const BoundingBox& newBox )
{
	if( init ) {
		low = min( low, newBox.lowBound() );
		high = max( high, newBox.highBound() );
	} else {
		low = newBox.lowBound();
		high = newBox.highBound();
		init = true;
	}
}

//
//  Just transforms the 8 corner points of the box and
//  computes the bounding box of the transformed points
//
void BoundingBox::applyTransform( const mat4& tform )
{
	BoundingBox newBox;

	if( init ) {

		// loop through the corner points
		for( OctantType i = 0; i < 8; i++ ) {

			vec3 curPoint;
			vec3 newPoint;

			// compute the current corner point
			curPoint[VX] = (i&4) ? low[VX] : high[VX];
			curPoint[VY] = (i&2) ? low[VY] : high[VY];
			curPoint[VZ] = (i&1) ? low[VZ] : high[VZ];

			// transform the current point
			newPoint = tform * curPoint;
			newBox.expand( newPoint );
		}
	}

	(*this) = newBox;
}

//
//  Compute the 3 midplanes for this box
//
void BoundingBox::clipPlanes( OctantType octant, vec4 planes[3] ) 
{
	vec3 mid = 0.5 * (high + low);

	planes[VX] = vec4( (octant&4) ? 1.0 : -1.0, 0.0, 0.0, -mid[VX] );
	planes[VY] = vec4( 0.0, (octant&2) ? 1.0 : -1.0, 0.0, -mid[VY] );
	planes[VZ] = vec4( 0.0, 0.0, (octant&1) ? 1.0 : -1.0, -mid[VZ] );
}

//
//  Compute a bounding box for one octant of the current box
//
BoundingBox BoundingBox::subBox( OctantType octant )
{
	BoundingBox result;

	vec3 mid = 0.5 * (high + low);
	
	result.low[VX] = (octant&4) ? mid[VX] : low[VX];
	result.high[VX] = (octant&4) ? high[VX] : mid[VX];

	result.low[VY] = (octant&2) ? mid[VY] : low[VY];
	result.high[VY] = (octant&2) ? high[VY] : mid[VY];

	result.low[VZ] = (octant&1) ? mid[VZ] : low[VZ];
	result.high[VZ] = (octant&1) ? high[VZ] : mid[VZ];

	return result;
}

//
//  Compute the corner points of the box
//
void BoundingBox::cornerPoints( vec3 points[8] )
{
	points[0] = vec3( low[VX],  low[VY],  high[VZ] );
	points[1] = vec3( high[VX], low[VY],  high[VZ] );
	points[2] = vec3( high[VX], high[VY], high[VZ] );
	points[3] = vec3( low[VX],  high[VY], high[VZ] );
	points[4] = vec3( low[VX],  low[VY],  low[VZ] );
	points[5] = vec3( high[VX], low[VY],  low[VZ] );
	points[6] = vec3( high[VX], high[VY], low[VZ] );
	points[7] = vec3( low[VX],  high[VY], low[VZ] );
}

//
//	Determine which octant this point lies in
//
OctantType BoundingBox::whichOctant( vec3& point ) 
{
	vec3 mid = 0.5 * (high+low);
	unsigned char result;

	// compare point to each coordinate of midpoint to
	// see if it lies in the upper or lower region for
	// that coordinate
	result = ((point[VX] >= mid[VX]) ? 4 : 0) |
			 ((point[VY] >= mid[VY]) ? 2 : 0) |
			 ((point[VZ] >= mid[VZ]) ? 1 : 0);

	return result;
}


//
//  Compute a bitmask for this bounding box in a particular region
//  of image space.  Assume that the first 2 coordinates of the box are
//  screenspace x and y.
//
Bitmask BoundingBox::bitMask( vec4& imageRegion )
{
	vec2 regionLow = vec2( imageRegion[XL], imageRegion[YL] );
	vec2 regionHigh = vec2( imageRegion[XU], imageRegion[YU] );

	double regionSize = imageRegion[XU] - imageRegion[XL];

	// calculate low and high bounds of the box in the image region, normalized to
	// the range [0..8].
	vec2 boxLow =  Bitmask::maskSize*((max( vec2(low, VZ), regionLow  ) - regionLow) / regionSize);
	vec2 boxHigh = Bitmask::maskSize*((min( vec2(high,VZ), regionHigh ) - regionLow) / regionSize); 

	// Calculate the first and last sample positions covered by the box in each dimension
	int xLow = fastDblToInt(ceil(boxLow[VX]-0.5));
	int xHigh = fastDblToInt(floor(boxHigh[VX]-0.5));
	int yLow = fastDblToInt(ceil(boxLow[VY]-0.5));
	int yHigh = fastDblToInt(floor(boxHigh[VY]-0.5));

	// construct the result
	Bitmask result;
	result.setBits( xLow, yLow, xHigh, yHigh );
	return result;
}

// 
//  Calculate a triage mask for this bounding box in a particular region
//  of image space.  Assumes first two coordinates of box represent image
//  space x and y.
//
TriageMask BoundingBox::triageMask( vec4& imageRegion )
{
	vec2 regionLow = vec2( imageRegion[XL], imageRegion[YL] );
	vec2 regionHigh = vec2( imageRegion[XU], imageRegion[YU] );

	double regionSize = imageRegion[XU] - imageRegion[XL];

	// normalize box bounds to [0..8]
	vec2 boxLow = TriageMask::maskSize*((max( vec2(low,VZ), regionLow ) - regionLow) / regionSize);
	vec2 boxHigh = TriageMask::maskSize*((min( vec2(high,VZ), regionHigh ) - regionLow) / regionSize); 

	// calculate first and last subregions completely covered by the box
	int CxLow = fastDblToInt(ceil(boxLow[VX]));
	int CxHigh = fastDblToInt(floor(boxHigh[VX])-1);
	int CyLow = fastDblToInt(ceil(boxLow[VY]));
	int CyHigh = fastDblToInt(floor(boxHigh[VY])-1);

	// calculate first and last subregions partially covered by the box
	int AxLow = fastDblToInt(floor(boxLow[VX]));
	int AxHigh = fastDblToInt(ceil(boxHigh[VX])-1);
	int AyLow = fastDblToInt(floor(boxLow[VY]));
	int AyHigh = fastDblToInt(ceil(boxHigh[VY])-1);


	// construct the TriageMask from the C and A masks
	Bitmask C;
	Bitmask A;
	Bitmask V;

	C.setBits( CxLow, CyLow, CxHigh, CyHigh );
	A.setBits( AxLow, AyLow, AxHigh, AyHigh );

	V = ~( C | A );

	return TriageMask( C, V );
}


//
//  Output routines
//
void BoundingBox::write( ostream& s )
{
	s << "Min = " << low << ", Max = " << high << endl;
}

void printOctantType( ostream& s, OctantType o )
{
	char *names[] = { "xNyNzN",
					  "xNyNzP",
					  "xNyPzN",
					  "xNyPzP",
					  "xPyNzN",
					  "xPyNzP",
					  "xPyPzN", 
					  "xPyPzP"
	};

	s << names[o];
}

