#ifndef BoundingBox_h
#define BoundingBox_h

#include "algebra3.h"

class ostream;
class TriageMask;
class Bitmask;

//
//  This enumeration is arranged such that the first bit
//  represents the z coordinate, the second bit represents
//  the y coordinate, and the third bit represents the x
//  coordinate.  A zero bit represents the lower value
//  for that coordinate, a one bit represents the higher
//  value.
//
enum {	
		xNyNzN=0,
		xNyNzP=1,
		xNyPzN=2,
		xNyPzP=3,
		xPyNzN=4,
		xPyNzP=5,
		xPyPzN=6, 
		xPyPzP=7 
};
typedef unsigned short OctantType;

void printOctantType( ostream& s, OctantType o );


/****************************************************************

	Class:		BoundingBox
	Purpose:	Represents a 3d bounding box.  Methods for computing
				octree subboxes and for testing points withing boxes

*****************************************************************/
class BoundingBox
{
public:

	BoundingBox();
	BoundingBox( const vec3& low, const vec3& high );
	BoundingBox( const BoundingBox& bb );

	// expand by adding a new point
	void expand( vec3& newPoint );

	// union of two bounding boxes
	void expand( const BoundingBox& newBox );

	// transform box - boudning box of transformed points
	void applyTransform( const mat4& tform );

	// methods for accessing bounds
	vec3 lowBound() const { return low; };
	vec3 highBound() const { return high; };
	vec3 midBound() const { return 0.5 * (high + low); };

	// get the plane equations for the 3 axis-aligned planes centred
	// at the centre of the box
	void clipPlanes( OctantType octant, vec4 planes[3] );

	// get the corner points of the box
	void cornerPoints( vec3 points[8] );

	// get the bounding box for one octant of the current box
	BoundingBox subBox( OctantType octant );

	// determine in which octant of the box the point lies
	OctantType whichOctant( vec3& point );

	// compute bitmask/triage masks of the box for a particular image region
	Bitmask bitMask( vec4& imageRange );
	TriageMask triageMask( vec4& imageRange );

	// output methods
	void write( ostream& s );
	friend ostream& operator<<( ostream& s, BoundingBox& bb ) { bb.write(s); return s; };

private:

	vec3 low;		// low bound
	vec3 high;		// high bound

	bool init;		// is initialized?
};

#endif