#include "BspTree.h"
#include "Tessellation.h"

//
//  Constructor - initializes members and constructs the tree
//
BspTree::BspTree( Tessellation *t, List<int>& faceList )
: tess(t), nodes(3*faceList.size()), size(0)
{
	// how many faces to add?
	int numFaces = faceList.size();
	
	// initialize the tree, -1 means 'empty'
	nodes.setList( (unsigned char)-1 );

	// insert each of the faces into the BSP tree
	for( int i = 0; i < numFaces; i++ ) {
		insertFace( faceList[i] );
	}

	// shrink the list to fit the tree size exactly
	nodes.shrinkWrap();
}

//
//  Appends a leaf to the tree
//
int BspTree::addFace( int face )
{
	nodes.append(face);
	nodes.append(-1);
	nodes.append(-1);

	size++;
	return size-1;
}

//
//  Non-recursive face insert function.  Starting from the top of
//  the tree, clip the face to the current node face, and insert the
//  pieces into the child nodes of the node.    
//
//  To avoid recursion, we use a stack-based algorithm to insert a
//  face into the tree.  We cannot use recursion, because it will 
//  inevitably overflow the system call stack for larger trees.
//
void BspTree::insertFace( int face, int nodeIndex )
{
	// Stack used to keep track of which faces need to be inserted into
	// which nodes.
	//
	// The stack is a list of (face,node) pairs.
	static List<int> faceStack(20);

	if( size == 0 ) {

		// if the tree is empty, just append the face as the only leaf
		addFace( face );

	} else {

		// reset the insert stack 
		faceStack.resize(0);
		
		// push the current (face,node) combination onto the stack
		faceStack.push( face );
		faceStack.push( nodeIndex );

		// loop through the stack, popping off (face,node) pairs and
		// performing the appropriate insertion.  Any new insertions that
		// must be performed are pushed on the stack
		while( faceStack.size() > 0 ) {

			// get the current face and the node into which it is to be inserted
			int curNode = faceStack.pop();
			int curFace = faceStack.pop();

			// get the indices in the BSP tree of the current node and its left and
			// right children
			int faceIndex = 3*curNode;
			int negIndex = faceIndex+1;
			int posIndex = faceIndex+2;

			// indices of split faces
			int negFace, posFace;

			// get the plane equation of the current node's face, and split the
			// face that we are inserting based on this equation.  The split
			// function returns the indices of the new faces in 'posFace' and
			// 'negFace'
			vec4 planeEq = tess->eList[nodes[faceIndex]];
			tess->splitFace( curFace, planeEq, negFace, posFace );

			// insert the neg-size face (if it exists) into the negative subtree 
			// of the current node
			if( negFace != -1 ) {
				if( nodes[negIndex] == -1 ) {

					// this is the first node in the negative subtree, so just add it
					nodes[negIndex] = addFace( negFace );

				} else {

					// the negative subtree is non-empty, so we must insert the new face
					// into the subtree.  Just add the (face,node) pair to the stack
					// and we'll get to it sometime
					faceStack.push( negFace );
					faceStack.push( nodes[negIndex] );
				}
			}

			// same as above, but fpr [psotopm face
			if( posFace != -1 ) {
				if( nodes[posIndex] == -1 ) {
					nodes[posIndex] = addFace( posFace );
				} else {
					faceStack.push( posFace );
					faceStack.push( nodes[posIndex] );
				}
			}
		}
	}
}

//
//  Computes a front-to-back face ordering from the given viewpoint for
//  the specified faces
//
void BspTree::orderFaces( vec3& viewPoint, List<int>& faces, int node )
{
	if( node != -1 ) {

		// get the indices of the current node and its children
		int faceIndex = 3*node;
		int negIndex = faceIndex+1;
		int posIndex = faceIndex+2;

		// determine which side of the current face the eyepoint lies
		// on
		vec4 planeEq = tess->eList[nodes[faceIndex]];
		SideType whichSide = inside( viewPoint, planeEq );

		switch( whichSide ) {

		case INSIDE:
		case INPLANE:

			// if the eyepoint is on the negative side, process the negative
			// subtree, then the node, then the positive subtree
			orderFaces( viewPoint, faces, nodes[negIndex] );
			faces.append( nodes[faceIndex] );
			orderFaces( viewPoint, faces, nodes[posIndex] );
			break;

		case OUTSIDE:

			// if the eyepoint is on the position side, process the positive
			// subtree, then the node, then the negative subtree
			orderFaces( viewPoint, faces, nodes[posIndex] );
			faces.append( nodes[faceIndex] );
			orderFaces( viewPoint, faces, nodes[negIndex] );
			break;
		}
	}
}


ostream& operator<<( ostream& s, BspTree& t )
{
	s << "Tree size = " << t.size << endl;

	for( int i = 0; i < t.size; i++ ) {
		s << "Node " <<  i   << ", face=" << t.nodes[3*i]
			                 << ", neg=" << t.nodes[3*i+1]
							 << ", pos=" << t.nodes[3*i+2]
							 << endl;
	}

	return s;
}

