#include <string.h>
#include "Scene.h"
#include "Cam.h"
#include "Light.h"
#include "Texture.h"
#include "BoundingBox.h"
#include "glut.h"

//
//  Attribute index definitions
//
int Cam::kOrigin;
int Cam::kLookAt; 
int Cam::kUpDir; 
int Cam::kImageRes;
int Cam::kFOV;
int Cam::kImageName;
int Cam::kRenderable;
int Cam::kBgColor;
int Cam::kShadingType;
int Cam::kClipPlanes;
int Cam::kDebugRegion;
int Cam::kAnimRange;

//
//  Enumeration strings for shading types attribute
//
static char *shadingTypeNames[] = { "FLAT",
									"FACE",
									"VERTEX",
									"PIXEL",
									NULL
};

//
//  Constructor - just initializes attributes and view transforms
//
Cam::Cam() : fViewTransform( identity3D() ), fInvViewTransform( identity3D() ) 
		     
{
	Cam::initializeAttributes();
}

//
//  Initialize the attributes for an instance of a camera
//
void Cam::initializeAttributes()
{
	Model::initializeAttributes();	

	kOrigin =		addAttr( new AttributeVec3(		"origin",		ANIMATABLE,	vec3(0.0)	) );
	kLookAt =		addAttr( new AttributeVec3(		"lookat",		ANIMATABLE,	vec3(0.0)	) );
	kUpDir =		addAttr( new AttributeVec3(		"updir",		ANIMATABLE,	vec3(0.0)	) );
	kImageRes =		addAttr( new AttributeVec2(		"imageres",		CONSTANT,	vec2(512.0) ) );
	kFOV =			addAttr( new AttributeVec2(		"fov",			ANIMATABLE,	vec2(57.0)	) );
	kImageName =	addAttr( new AttributeString(	"imagename",	CONSTANT,	"image.ppm" ) );
	kRenderable =	addAttr( new AttributeBool(		"renderable",	CONSTANT,	true		) );
	kBgColor =		addAttr( new AttributeVec3(		"bgcolor",		TEXTURABLE,	vec3(0.0)	) );
	kShadingType =	addAttr( new AttributeEnum(		"shadingtype",	CONSTANT,	kPerPixel, shadingTypeNames, kNumShadingTypes	) );
	kClipPlanes =	addAttr( new AttributeVec2(		"clipplanes",	ANIMATABLE,	vec2(0.1,100.0) ) );
	kDebugRegion =	addAttr( new AttributeVec4(		"debugregion",	CONSTANT,	vec4(-1.0)	) );
	kAnimRange =	addAttr( new AttributeVec2(		"animrange",	CONSTANT,	vec2(0.0)	) );
}

//
//  Called after a camera is read from a file.  Computes the transformation
//  matrices, and determines whether or not the camera is animated.
//
bool Cam::postReadFunc( const Scene& scene )
{
	if( Model::postReadFunc(scene) ) {
		computeMatrices();

		fCamAnimated = findAttr(kOrigin)->isAnimated() ||
					   findAttr(kLookAt)->isAnimated() || 
					   findAttr(kUpDir)->isAnimated();

		return true;
	} else {
		return false;
	}
}


// 
//	Before each frame, we update the camera transform/clipping planes, and have
//  the scene update any animated texture or light transforms.  We do not update
//  the geometry, since a particular camera may not support it.
//
void Cam::preFrameProcess( Scene& scene )
{
	// update transforms, clipping planes
	computeMatrices();

	// update texture/light transforms
	scene.updateTextures();
	scene.updateLights();

	// re-init RNG
	srand(0);
}

//
//  The main rendering algorithm.  Loop through the animation times and
//  render each frame.
//
void Cam::render( Scene& scene )
{
	// get the frame range
	vec2 range = attrAnimRange();
	int start = fastDblToInt(range[0]);
	int end = fastDblToInt(range[1]);

	for( int curFrame = start; curFrame <= end; curFrame++ ) {

		// set the current time
		scene.setCurTime( (double)curFrame );

		// do appropriate preprocessing for this frame
		cout << endl << "Camera " << attrName() << " preprocessing frame " << Attribute::curTime() << endl;
		preFrameProcess(scene);

		// render the frame
		cout << "Camera " << attrName() << " rendering frame " << Attribute::curTime() << endl;
		renderFrame(scene);

		// do appropriate postprocessing for this frame
		cout << "Camera " << attrName() << " postprocessing frame " << Attribute::curTime() << endl;
		postFrameProcess(scene);

	}
}

//
//	Takes the current camera position attributes and computes the view
//  transform, inverse view transform, and projective transform.  Uses
//  the algorithm from Watt+Watt.
//
void Cam::computeMatrices( void )
{
	vec2 clip = findAttr(kClipPlanes)->asVec2();
	vec2 fov = findAttr(kFOV)->asVec2();

	//
	// viewing transform
	//

	vec3 C(attrOrigin());
	vec3 V(attrUpDir());
	vec3 N(attrLookAt() - attrOrigin());
	
	mat4 T;
	mat4 B = identity3D();
	vec3 U, Vproj;

	N.normalize();
	V.normalize();

	Vproj = V - (V * N)*N;
	Vproj.normalize();
	U = N ^ Vproj;
	
	T = translation3D(-C);

	B[0] = vec4(U,0.0);
	B[1] = vec4(Vproj,0.0);
	B[2] = vec4(-N,0.0);
	
	fViewTransform = B*T;
	fInvViewTransform = fViewTransform.inverse();

	//
	// projective transform
	//

	double a, b, c, d, e, f;
	double left, right, top, bottom;

	double nearClip = clip[0];
	double farClip = clip[1];

	right = nearClip*tan(((0.5*fov[VX])/180.0)*M_PI);
	left = -right;
	
	top = nearClip*tan(((0.5*fov[VY])/180.0)*M_PI);
	bottom = -right;

	a = (right + left)/(right - left);
	b = (top + bottom) / (top - bottom);
	c = -(farClip + nearClip)/(farClip - nearClip);
	d = -2.0*farClip*nearClip/(farClip - nearClip);
	e = 2*nearClip / (right - left);
	f = 2*nearClip / (top - bottom);

	fProjTransform[0] = vec4(e, 0.0, a, 0.0);
	fProjTransform[1] = vec4(0.0, f, b, 0.0);
	fProjTransform[2] = vec4(0.0, 0.0, c, d);
	fProjTransform[3] = vec4(0.0, 0.0, -1.0, 0.0);

	// also recompute clipping plane equations
	computeClipPlanes();
}

//
//  Computes the plane equations for the faces of the viewing
//  frustum (in camera space).
//
void Cam::computeClipPlanes()
{
	vec2 clip = attrClipPlanes();
	vec2 fov = attrFOV();
	vec2 imageRes = attrImageRes(); 

	// compute view frustum in camera space 
	double right = clip[0]*tan(((0.5*fov[VX])/180.0)*M_PI);
	double left = -right;
	double top = clip[0]*tan(((0.5*fov[VY])/180.0)*M_PI);
	double bottom = -right;

	vec3 clampL( 0.0, 0.0, clip[0] );
	vec3 clampH( imageRes, clip[1] );

	vec3 tl( left, top, -clip[0] );
	vec3 tr( right, top, -clip[0] );
	vec3 bl( left, bottom, -clip[0] );
	vec3 br( right, bottom, -clip[0] );
	vec3 origin( 0.0 );

	// compute the clip plane equations 
	fClipPlanes[0] = vec4( 0.0, 0.0, -1.0, -clip[0] );
	fClipPlanes[1] = vec4( 0.0, 0.0, 1.0, clip[1] );
	fClipPlanes[2] = computeTriPlaneEq( tl, tr, origin );
	fClipPlanes[3] = computeTriPlaneEq( br, bl, origin );
	fClipPlanes[4] = computeTriPlaneEq( bl, tl, origin );
	fClipPlanes[5] = computeTriPlaneEq( tr, br, origin );
}

//
//  Method for transforming a point from world space to screen space
//
vec3 Cam::worldToScreen( const vec3& wsPoint ) 
{
	vec2 ir = attrImageRes();

	/* transform into camera space */
	vec3 csPoint = fViewTransform * wsPoint;

	/* project point */
	vec4 projPoint = fProjTransform * vec4( csPoint, 1.0 );

	/* normalize */
	projPoint /= projPoint[VW];

	/* convert to screen space */
	vec2 ss( projPoint[VX], projPoint[VY] );
	ss[VX] = (ss[VX] + 1.0)/2.0;
	ss[VY] = (ss[VY] + 1.0)/2.0;
	ss[VX] *= ir[VX];
	ss[VY] *= ir[VY];

	return vec3( ss, projPoint[VZ] );
}

//
//  Transforms a list of points from world to screen space
//
void Cam::worldToScreen( const List<vec3>& wsPoints, List<vec3>& ssPoints )
{
	vec2 ir = attrImageRes();
	int numPoints = wsPoints.size();

	// make sure output list is big enough to hold all points
	ssPoints.resize( numPoints );

	for( int i = 0; i < numPoints; i++ ) {
		/* transform into camera space */
		vec3 csPoint = fViewTransform * wsPoints[i];

		/* project point */
		vec4 projPoint = fProjTransform * vec4( csPoint, 1.0 );

		/* normalize */
		projPoint /= projPoint[VW];

		/* convert to screen space */
		vec2 ss( projPoint[VX], projPoint[VY] );
		ss[VX] = (ss[VX] + 1.0)/2.0;
		ss[VY] = (ss[VY] + 1.0)/2.0;

		ss[VX] *= ir[VX];
		ss[VY] *= ir[VY];

		ssPoints[i] = vec3( ss, projPoint[VZ] );
	}
}	

/*
IsectType Cam::clipToFrustum( Face& face )
{
	IsectType ist;
	bool anyNeg = false;
	bool allPos = true;

	for( int p = 0; p < 6; p++ ) {
	
		ist = face.clip( clipPlanes[p], CAMERA );

		anyNeg |= (ist == NEG );
		allPos &= (ist == POS );
		if( anyNeg ) break;
	}

	if( anyNeg ) {
		return NEG;
	} else if( allPos ) {
		return POS;
	} else {
		return BOTH;
	}
}

IsectType Cam::isectBBoxFrustum( BoundingBox& bb )
{
	vec3 corners[8];
	Outcode outside = 255;
	Outcode inside = 0;

	bb.cornerPoints( corners );

	for( int i = 0; i < 8; i++ ) {
		Outcode curOutcode = computeOutcode( corners[i], clipPlanes, 6 ); 

		outside = outside & curOutcode;
		inside = inside | curOutcode;
	}

	if( outside ) {
		return POS;
	} else if( inside ) {
		return NEG;
	} else {
		return BOTH;
	}
}
*/
