#ifndef Cam_h
#define Cam_h

#include "Model.h"
#include "List.h"

class Scene;

/****************************************************************

	Class:		Cam
	Purpose:	Base class for all cameras.  Provides attributes
				common to all cameras (view/image attributes, 
				transforms, etc).

*****************************************************************/
class Cam : public Model
{
public:
	Cam();
	virtual ~Cam() {};

	// valid values for the "shadingType" attribute
	typedef enum {kFlat=0,kPerFace,kPerVertex,kPerPixel,kNumShadingTypes} ShadingType;

	// attribute initialization
	virtual void initializeAttributes();

	// attribute indices
	static int kOrigin;			// where the camera is positioned
	static int kLookAt;			// the point the camera is looking at
	static int kUpDir;			// the camera up vector
	static int kImageRes;		// size of the image
	static int kFOV;			// camera field of view
	static int kImageName;		// image file name
	static int kRenderable;		// determines whether or not the camera gets rendered
	static int kBgColor;		// camera bg color
	static int kShadingType;	// shading model to use
	static int kClipPlanes;		// near/far clip planes
	static int kDebugRegion;	// region of interest for debugging
	static int kAnimRange;		// which frames to render

	// attribute accessors
	vec3& attrOrigin() { return findAttr(kOrigin)->asVec3(); };
	vec3& attrLookAt() { return findAttr(kLookAt)->asVec3(); };
	vec3& attrUpDir()  { return findAttr(kUpDir)->asVec3(); };
	vec2& attrImageRes() { return findAttr(kImageRes)->asVec2(); };
	vec2& attrFOV()	   { return findAttr(kFOV)->asVec2(); };
	char *attrImageName() { return findAttr(kImageName)->asString(); };
	bool attrRenderable() { return findAttr(kRenderable)->asBool(); };
	vec3& attrBgColor() { return findAttr(kBgColor)->asVec3(); };
	ShadingType attrShadingType() { return (ShadingType)findAttr(kShadingType)->asInt(); };
	vec2& attrClipPlanes() { return findAttr(kClipPlanes)->asVec2(); };
	vec4& attrDebugRegion() { return findAttr(kDebugRegion)->asVec4(); };
	vec2& attrAnimRange() { return findAttr(kAnimRange)->asVec2(); };

	// called after read
	virtual bool postReadFunc( const Scene& scene );

	// methods to access the various transformation matrices for this camera
	mat4 viewTransform( void ) const { return fViewTransform; };
	mat4 invViewTransform( void ) const { return fInvViewTransform; };
	mat4 projTransform( void ) const { return fProjTransform; };

	//
	// Rendering methods: 
	//
	//		preProcess		is called before rendering any frames
	//		preFrameProcess is called before rendering each frame
	//		renderFrame		is called to render each frame
	//		posFrameProcess is called after rendering each frame
	//		postProcess		is called after all frames have been rendered
	//
	//	The 'render' method invokes the previous 5 methods in the prescribed
	//  order.

	virtual void preProcess( Scene& scene ) {};
	virtual void postProcess( Scene& scene ) {};
	virtual void renderFrame( Scene& scene ) {};
	virtual void preFrameProcess( Scene& scene );
	virtual void postFrameProcess( Scene& scene ) {};

	virtual void render( Scene& scene );


	// point transformation routines 
	vec3 worldToScreen( const vec3& wsPoint );
	void worldToScreen( const List<vec3>& wsPoints, List<vec3>& ssPoints );

	//IsectType clipToFrustum( Face& face );
	//IsectType isectBBoxFrustum( BoundingBox& bb );

	// determines whether or not the camera is animated
	bool camAnimated() { return fCamAnimated; };

protected:

	void computeMatrices(void);		// computes view/proj matrices
	void computeClipPlanes(void);	// computes frustum plane equations

	// the transformation matrices
	mat4 fViewTransform;		// world -> camera transform
	mat4 fInvViewTransform;		// camera -> world transform
	mat4 fProjTransform;		// camera -> normalized screen transform

	vec4 fClipPlanes[6];		// plane equations (world) of view frustum

	bool fCamAnimated;			// is the camera animated?

	Timer *rootTimer;			// root of all timing probes for this camera
};

#endif