#ifndef HCMCam_h
#define HCMCam_h

#include "Cam.h"
#include "AffineTriangle.h"

class Image;
class Triangle;
class CoveragePyramid;
class Light;
class Tessellation;
class Octree;
class Shader;


/***************************************************************************************************

	Class:		CamHCM
	Purpose:	Camera that implements Ned Greene's Hierarchical
				Coverage Mask visibility algorithm (from Siggraph
				'96).

				The algorithm is:

					Before rendering starts:

						1) tessellate the scene into worldspace
						2) organize the tessellation into an octree in which
						   only the leaf nodes actually contain geometry.  Polys
						   that straddle several nodes are clipped to each node.
						3) organize each leaf node into a BSP tree

					For each frame:
						
						1) compute a complete front-to-back traversal order
						   for the faces.
						2) for each face, render it using Greene's hierarchical
						   coverage mask algorithm.  A face's bounding box is first
						   tested to see if any part of it is visible before the
						   actual face is rendered.

				For interpolative shading, each face is tessellated into triangles
				before rendering, and the AffineTri class is used to perform interpolation
				of appropriate shading parameters (in the case of per-pixel shading) or
				colours (in the case of per-vertex shading).  For each sample point, its affine
				coordinates within the current triangle are calculated.  This is expensive, as
				it does not take advantage of scanline coherence and is not incremental.

				The visibility algorithm can use either 1 or 64 visibility samples per pixel.
				This is controlled by the 'subpixel' attribute.

				This camera ignores all object animation, but camera
				animation is allowed.

*****************************************************************************************************/

class CamHCM : public Cam
{
public:

	CamHCM();
	virtual ~CamHCM() {};

	// attribute initialization
	virtual void initializeAttributes();

	// attribute indices
	static int kSubpixel;		// should visibility be resolved at the subpixel level?
	static int kCommand;		// command line mode for debugging

	// attribute accessors
	bool attrSubpixel() { return findAttr(kSubpixel)->asBool(); };
	bool attrCommand() { return findAttr(kCommand)->asBool(); };

	// pre-render function - constructs the visibility structure
	virtual void preProcess( Scene& scene );

	// pre-frame function - computes a front-to-back face ordering for the
	// current eyepoint
	virtual void preFrameProcess(Scene& scene );

	// renders the current frame using the HCM algorithm
	virtual void renderFrame( Scene& scene );

	// post-frame function - renders the background and writes the image
	virtual void postFrameProcess(Scene& scene );

	// post-render function - cleans up data structures
	virtual void postProcess( Scene& scene );

protected:

	Image *image;						// the computed image

	CoveragePyramid *coveragePyramid;	// the screenspace hierarchical visibility structure

	Tessellation *tess;					// tessellation of the entire scene

	Octree *octree;						// octree organization of the faces in the tessellation

	List<int> *faceOrder;				// used to store the front-to-back face ordering for the
										// current point of view

	List<vec3> *ssPoints;				// used to store screenspace projections

	// data for rendering a single face
	vec3 curFaceColor;					// colour of the current face (for flat and per-face shading
										// modes)
	AffineTriangle curTriData;			// for interpolative shading in the current triangle
	vec3 vertexColors[3];				// colours of the vertices for the current triangle (used in 
										// per-vertex shading)
	List<Light *> *lights;				// current light list

	//
	//	Protected methods for rendering faces
	//

	// renders a single triangle into a particular region of the image
	void renderTri( Triangle& tri, Shader *shader, int level, int x, int y );

	// renders a single face into a particular region of the image
	void renderFace( int face, bool reverse, int level, int x, int y );

	// renders the background colour into the image
	void renderBG( vec3& bgColor, int level, int x, int y );
};

#endif