#include "CamZB.h"
#include "Scene.h"
#include "Poly.h"
#include "Triangle.h"
#include "Face.h"
#include "Image.h"
#include "EdgeTable.h"
#include "Sample.h"
#include "Shader.h"
#include "Tessellation.h"
#include <stdio.h>

static Timer *CurCamTimer;
static Timer *CurFrameTimer;

//
//  Constructor - just initialize the attributes for the class
//
CamZB::CamZB() : fImage(NULL), fEdgeTable(NULL), fLights(NULL),
				 fWorldPoints(NULL), fScreenPoints(NULL), fWorldNormals(NULL)
{
	CamZB::initializeAttributes();
}

//
//  Attribute initialization
//
void CamZB::initializeAttributes()
{
	Cam::initializeAttributes();
}

//
//  Pre-render function.  Set up the image and the edge table, and
//  tessellate the scene
//
void CamZB::preProcess( Scene& scene )
{
	// set up main camera timer
	CurCamTimer = MainTimer->newChild( "ZB Camera Timer" );
	CurCamTimer->start();

	// set up preprocess timer
	Timer *ZBpreTimer = CurCamTimer->newChild( "ZB preprocess" );
	ZBpreTimer->start();

	// get current attribute values
	vec2 imageRes = attrImageRes();
	vec3 bgColor = attrBgColor();

	// set up image with depth buffer
	cout << endl << "Setting up image..." << endl;
	fImage = new ImageZB( imageRes, bgColor );
	cout << "Done image setup." << endl;

	// set up edge table
	fEdgeTable = new EdgeTable( vec2( 0.0, imageRes[VY]-1.0 ), false );

	//********************START TESSELLATION**************************
	//
	//  If an object is animated, we tessellate into object space,
	//  otherwise we tessellated into worldspace.  
	//
	cout << "Tessellating..." << endl;

	TessOptions localOptions = kTessPlaneEqs;

	// update the scene geometry transforms
	scene.setCurTime( attrAnimRange()[0] );
	scene.updateGeometry();

	bool someGeomAnimated = false;		// is some piece of geometry animated?  This
										// determines if we need to allocate the worldspace
										// point and normal arrays

	int maxAnimVert = -1;				// what is the max # of animated vertices?  This
										// determines the size of the worldspace point and
										// normal arrays

	// loop through all the objects in the scene, tessellating each one
	for( int curShape = 0; curShape < scene.objList.size(); curShape++ ) {

		Shape *shape = scene.objList[curShape];

		if( shape->geometryAnimated() ) {

			// if the object is animated, we must tessellate it into local
			// space

			someGeomAnimated = true;

			shape->localTessellate( localOptions, LOCAL );
			
			// update the max # of animated vertices
			int numVertices = shape->localTessellation()->vList.size();
			maxAnimVert = MAX(maxAnimVert,numVertices);

		} else {

			// if the object is not animated, we tessellate directly into
			// worldspace and compute plane equations
			shape->localTessellate( localOptions, WORLD );
		}
	}
	
	cout << "Done tessellation." << endl;

	//**********************END TESSELLATION****************************

	// allocate the screenspace vertex array
	int numVertices = scene.maxVertices();
	fScreenPoints = new List<vec3>( numVertices );

	// if something is animated, allocate the worldspace point and normal
	// arrays
	if( someGeomAnimated ) {
		fWorldPoints = new List<vec3>( maxAnimVert );
		fWorldNormals = new List<vec3>( maxAnimVert );
	}

	// set up light list
	fLights = &(scene.lightList);

	ZBpreTimer->stop();
}


//
//  Pre-frame function.  Clears the image/depth buffer and updates the geometry
//  transforms for scene objects.
//
void CamZB::preFrameProcess( Scene& scene )
{
	// set up the timer for this frame
	char timerName[50];
	sprintf( timerName, "Frame %d render", (int)scene.getCurTime() );
	CurFrameTimer = CurCamTimer->newChild( timerName );
	CurFrameTimer->start();

	Cam::preFrameProcess(scene);

	// clear the image/depth buffer
	fImage->clear();

	// update positions of geometry
	scene.updateGeometry();
}

//
//  Post-frame function.  Writes the image to a file
//
void CamZB::postFrameProcess( Scene& scene )
{
	Cam::postFrameProcess(scene);

	// write the image to file
	char fileName[100];
	sprintf(fileName, "%s.%d.ppm", attrImageName(), (int)scene.getCurTime());
	fImage->write( fileName );

	CurFrameTimer->stop();
}

//
//  Post-render function.  Cleans up data structures.
//
void CamZB::postProcess( Scene& scene )
{
	if( fEdgeTable ) delete fEdgeTable;
	if( fImage ) delete fImage;
	if( fWorldPoints ) delete fWorldPoints;
	if( fScreenPoints ) delete fScreenPoints;
	if( fWorldNormals ) delete fWorldNormals;

	CurCamTimer->stop();
}


//
//  Frame render function, gets called to render each frame
//
void CamZB::renderFrame( Scene& scene )
{
	int numObjects = scene.objList.size();
	mat4 worldToScreenMat = fProjTransform * fViewTransform;

	// get current attribute values
	vec3 origin = attrOrigin();
	ShadingType shadingType = attrShadingType();

	// fill in eyepoint for shading samples
	Sample& currentSample = Attribute::curSample();
	currentSample.eye = origin;

	// for tessellating faces
	Triangle triangles[10];


	// loop through the objects in the scene, rendering each one
	for( int i = 0; i < numObjects; i++ ) {

		// pointers to world space points and normals for the current object.
		// if the object is not animated, these point into its tessellation structure,
		// otherwise they point to the external point and normal arrays
		List<vec3> *curWorldPoints;
		List<vec3> *curWorldNormals;

		// skip hidden objects
		Shape *curShape = scene.objList[i];
		if( curShape->attrHidden() ) continue;

		// cumulative transform for this object
		mat4 objToWorldMat = curShape->getCtransform();

		// get tessellation and shader for this object
		Tessellation *curTess = curShape->localTessellation();
		Shader *curShader = curShape->getShader();

		// is the object animated?
		bool geomAnim = curShape->geometryAnimated();

		// update the tessellation (points, normals, plane equations
		if( geomAnim ) {

			// if this object is animated, the tessellation is in object space,
			// so we convert the points to worldspace, compute plane equations,
			// then convert the points to screen space.  Normals are transformed
			// into worldspace

			// resize arrays
			fWorldNormals->resize( curTess->vList.size() );
			fWorldPoints->resize( curTess->vList.size() );
			fScreenPoints->resize( curTess->vList.size() );

			// vertices to world space, compute plane eqs, storing them in
			// the local tessellation
			curTess->transformVertices( objToWorldMat, fWorldPoints );
			curTess->computePlaneEqs( fWorldPoints, NULL );

			// compute screenspace points
			worldToScreen( (*fWorldPoints), (*fScreenPoints) );

			// compute worldspace normals
			curTess->transformNormals( curShape->getCNtransform(), fWorldNormals );

			// initialize the point/normal array pointers for this object
			curWorldPoints = fWorldPoints;
			curWorldNormals = fWorldNormals;

		} else {

			// the object is not animated, so everything is already in world space.
			// Just recompute the screenspace points if the camera is animated
			worldToScreen( curTess->vList, (*fScreenPoints) );

			// initialize the point/normal array pointers for this object
			curWorldPoints = &(curTess->vList);
			curWorldNormals = &(curTess->nList);
		}


		// loop through the faces of the current object tessellation, rendering each one
		int numFaces = curTess->size();
		for( int curFace = 0; curFace < numFaces; curFace++ ) {

			// get the size and vertex indices of the current face
			int *vIndices = curTess->fList.getList( curFace );
			int faceSize = curTess->fList.getListSize( curFace );

			// figure out if this face is backfacing
			vec3 eyeToFace = (*curWorldPoints)[vIndices[0]]-origin;
			bool reverse = ( eyeToFace * vec3(curTess->eList[curFace],VW) > 0.0 );

			// tessellate the face into triangles
			int numTri = curTess->tessellateFace( curFace, reverse, curWorldPoints, 
												 (*fScreenPoints), curWorldNormals, triangles );

			// render each triangle
			for( int curTri = 0; curTri < numTri; curTri++ ) {

				// add this triangle to the edge table
				Triangle& tri = triangles[curTri];
				fEdgeTable->clear();			
				fEdgeTable->addTriangle( tri );

				// data for interpolation
				int scanLine;		// current scanline
				vec2 xRange;		// x range of current span
				LerpVec3 coords;	// for interpolating depth along a span - we interpolate
									// affine coords, and this allows us to interpolate anything
									// else


				// for flat and per-face shading, just render the triangle in a constant
				// color
				if( (shadingType == kPerFace) || (shadingType == kFlat) ) {

					vec3 currentColor;

					if( shadingType == kPerFace ) {

						// per-face shading, compute a random face color
						currentColor[0] = ((double)rand())/RAND_MAX;
						currentColor[1] = ((double)rand())/RAND_MAX;
						currentColor[2] = ((double)rand())/RAND_MAX;
						currentColor.clamp(0.0,1.0);
					} else {

						// flat shading, just use white for face color
						currentColor = vec3(1.0);
					}

					// traverse the edge table, rendering each span of the current
					// triangle
					fEdgeTable->startTraversal();
					while( fEdgeTable->getCurrentSpan( scanLine, xRange, coords ) ) {

						// traverse the current span
						int spanStart = fastDblToInt(xRange[0]);
						int spanEnd = fastDblToInt(xRange[1]);
						for( int i = spanStart; i <= spanEnd; i++ ) {
			
							// get affine coords of current point, interpolate depth from this
							vec3 currentCoords = coords.current().clamp(0.0,1.0);
							double currentDepth = currentCoords[0]*tri.screen(0)[VZ] +
												  currentCoords[1]*tri.screen(1)[VZ] +
												  currentCoords[2]*tri.screen(2)[VZ];

							// perform depth test and set pixel if appropriate
							fImage->setPixel( i, scanLine, (float)currentDepth, currentColor );

							// move to next pixel
							coords.increment();
						}
					}

				} else if( shadingType == kPerVertex ) {

					// per-vertex shading, shade at the vertices and interpolate the colors 
					vec3 vertexColors[3];

					// shade the vertices of the triangle
					for( int i = 0; i < 3; i++ ) {
						tri.fillSample( i, currentSample );
						vertexColors[i] = curShader->computeColor( scene.lightList );
					}

					// traverse the edge table, rendering each span of the current triangle
					fEdgeTable->startTraversal();
					while( fEdgeTable->getCurrentSpan( scanLine, xRange, coords ) ) {

						// interpolate color along the current span
						int spanStart = fastDblToInt(xRange[0]);
						int spanEnd = fastDblToInt(xRange[1]);
						for( int i = spanStart; i <= spanEnd; i++ ) {

							// get the affine coords of the current point
							vec3 currentCoords = coords.current().clamp(0.0,1.0);

							// interpolate the color using the affine coords
							vec3 currentColor = currentCoords[0]*vertexColors[0] +
												currentCoords[1]*vertexColors[1] +
												currentCoords[2]*vertexColors[2];
							currentColor.clamp(0.0,1.0);

							// interpolate the depth using the affine coords
							double currentDepth = currentCoords[0]*tri.screen(0)[VZ] +
												  currentCoords[1]*tri.screen(1)[VZ] +
												  currentCoords[2]*tri.screen(2)[VZ];

							// set the pixel, if it passes the depth test
							fImage->setPixel( i, scanLine, (float)currentDepth, currentColor );

							// move to the next pixel
							coords.increment();
						}
					}

				} else if( shadingType == kPerPixel ) {

					// per-pixel shading, interpolate shading parameters across each span
					// and shade at each pixel

					fEdgeTable->startTraversal();
					while( fEdgeTable->getCurrentSpan( scanLine, xRange, coords ) ) {

						int spanStart = fastDblToInt(xRange[0]);
						int spanEnd = fastDblToInt(xRange[1]);
						for( int i = spanStart; i <= spanEnd; i++ ) {

							// interpolate affine coords and depth
							vec3 currentCoords = coords.current().clamp(0.0,1.0);
							double currentDepth = currentCoords[0]*tri.screen(0)[VZ] + 
												  currentCoords[1]*tri.screen(1)[VZ] + 
												  currentCoords[2]*tri.screen(2)[VZ];

							// shade if the pixel passes the depth test
							if( fImage->testDepth( i, scanLine, (float)currentDepth ) ) {
								tri.fillSample( currentCoords, currentSample );
								vec3 currentColor = curShader->computeColor( scene.lightList ).clamp(0.0,1.0);

								fImage->setPixel( i, scanLine, (float)currentDepth, currentColor );
							}

							coords.increment();
						}
					}
				}
			}
		}
	}
}


