#ifndef ZBCam_h
#define ZBCam_h

#include "Cam.h"

class Scene;
class Face;
class Triangle;
class ImageZB;
class EdgeTable;
class Tessellation;
class Light;

/**********************************************************************************

	Class:		CamZB
	Purpose:	Camera that implements a software z-buffer rendering algorithm. 
				Visibility and shading are sampled once per pixel.

				The algorithm goes as follows:

					Before rendering starts:
						
						  1) each object is tessellated.  If the object is animated,
							 it is tessellated into object space, otherwise it is
							 tessellated into world space.

					For each frame:

						  1) for each object, if the object is animated, we update 
							 plane equations and normals in worldspace and vertices in
							 screen space.  If the object is not animated, but the camera
							 is, we just update the screen space vertices.

						  2) each object is rendered individually using a z-buffer 
						     visibility algorithm.

************************************************************************************/
class CamZB : public Cam
{
public:
	CamZB();
	virtual ~CamZB() {};

	// attribute initialization
	virtual void initializeAttributes();

	// attribute indices (none)

	// attribute accessors (none)

	// pre-render function - tessellates the scene
	virtual void preProcess( Scene& scene );

	// pre-frame function -	updates positions, plane eqs, and normals
	// (if necessary)
	virtual void preFrameProcess( Scene& scene );

	// renders the current frame using the z-buffer visibility algorithm
	virtual void renderFrame( Scene& scene );

	// post-frame function - writes the image
	virtual void postFrameProcess( Scene& scene );

	// post-render function - cleans up data structures
	virtual void postProcess( Scene& scene );

protected:

	EdgeTable     *fEdgeTable;		// edge table used for scan conversion
	ImageZB	      *fImage;			// image with depth buffer
	List<Light *> *fLights;			// light list
	List<vec3>    *fWorldPoints;	// stores points in world space
	List<vec3>    *fScreenPoints;	// stores points in screen space
	List<vec3>    *fWorldNormals;	// stores normals in world space

};

#endif