#include "CoveragePyramid.h"
#include "TriageMask.h"
#include "Triangle.h"
#include "BoundingBox.h"

//
//  Constructor initializes the coverage pyramid
//
CoveragePyramid::CoveragePyramid( int res, bool sub )
: subpixel(sub), debug(false), lvlOffsets(NULL)
{
	// round the given resolution to the nearest power of 8
	int power = (int)(3.0*ROUND((log10((double)res)/log10(2.0))/3.0));
	resolution = 1<<power;

	// determine how big the pyramid will be
	if( subpixel ) {
		maxlevel = power/3;
	} else {
		maxlevel = (power/3)-1;
	}

	// allocate the level offsets
	lvlOffsets = new int[maxlevel+1];

	// allocate the triage masks for all levels
	int arraySize = 1;
	pyramidSize = 0;
	for( int curLevel = 0; curLevel < maxlevel; curLevel++ ) {
		lvlOffsets[curLevel] = pyramidSize;
		arraySize *= 64;
		pyramidSize += arraySize;
	}

	// initialize number of leaf bitmasks
	leafSize = arraySize;
	
	// allocate the triage masks for all levels except the last
	pyramid = new TriageMask[pyramidSize];

	// initialize the last level of the pyramid
	samples = new Bitmask[arraySize];

	// initialize Bitmask precomputed tables
	if( !Bitmask::isInit() ) {
		Bitmask::initialize();
	}

	// initialize TriageMask precomputed tables
	if( !TriageMask::isInit() ) {
		TriageMask::initialize();
	}
}

//
//  Destructor cleans up the pyramid
//
CoveragePyramid::~CoveragePyramid()
{
	// delete the pyramid and the last level
	delete [] pyramid;
	delete [] samples;
}

// 
//  accessor methods
//
int CoveragePyramid::maxLevel()
{
	return maxlevel;
}

bool CoveragePyramid::subPixel()
{
	return subpixel;
}


//
//  Clears the pyramid.  Sets all TriageMasks and Bitmasks to empty
//
void CoveragePyramid::clear()
{
	// clear the TriageMasks
	for( int i = 0; i < pyramidSize; i++ ) {
		pyramid[i].clear();
	}

	// clear the last level of the pyramid
	memset( samples, '\0', sizeof(Bitmask)*leafSize );
}

//
//  Finds the single smallest pyramid node that completely covers the given bounding
//  box
//
void CoveragePyramid::enclosingLevel( BoundingBox& bb, int &level, int& x, int& y )
{
	// normalizing factor - normalizes entire image to [0..8]
	double fraction = (double)TriageMask::maskSize / (double)resolution;
	int ms = TriageMask::maskSize;

	vec3 low = bb.lowBound();
	vec3 high = bb.highBound();

	// get lower/upper bounds of box in normalizes [0..8] space
	double xLow = fraction * low[VX];
	double xHigh = fraction * high[VX];
	double yLow = fraction * low[VY];
	double yHigh = fraction * high[VY];

	// loop through the levels, at each level checking to see if the box
	// is still within a single square at that level
	for( int curLevel = 0; curLevel < maxlevel; curLevel++ ) {
		
		// if the box is contained in a single square at this level,
		// then move to the next level
		if( ((ceil(xHigh) - floor(xLow)) == 1.0) &&
			((ceil(yHigh) - floor(yLow)) == 1.0) ) {
			
			// multiply by the mask size to normalize the box coords to
			// the range of the next level ([0..8] -> [0..64] -> [0..512])
			xLow *= ms;
			xHigh *= ms;
			yLow *= ms;
			yHigh *= ms;

		} else {
			
			// the box straddles pyramid square boundaries, so the current level
			// is the best we can do, return the current level and current coordinates
			// at this level
			level = curLevel;
			x = fastDblToInt(floor( xLow / ms ));
			y = fastDblToInt(floor( yLow / ms ));
			return;

		}
	}

	// we made it all the way to the max level, so return coordinates relative to it
	level = maxlevel;
	x = fastDblToInt(floor( xLow / ms ));
	y = fastDblToInt(floor( yLow / ms ));
	return;


}

//
//  Finds the smallest region of pyramid squares (up to 2x2 in size) that completely
//  encloses the given bounding box.
//
//  This is much more useful than enclosingLevel.  If a very small poly straddles the
//  boundary between several very large pyramid regions, enclosingLevel will have to
//  return a very large bounding area.  However, enclosingSquares will return several
//  much smaller areas instead.
//
void CoveragePyramid::enclosingSquares( BoundingBox& bb, int &level, int& xLow, int& xHigh, int& yLow, int& yHigh )
{
	vec3 low = bb.lowBound();
	vec3 high = bb.highBound();

	double xWidth = high[VX] - low[VX];
	double yWidth = high[VY] - low[VY];

	// width of the box along its longest axis
	double maxWidth = MAX(xWidth,yWidth);

	// calculate the smallest power of 8 such that maxWidth is less than
	// that power.  This determines what level of the pyramid we should begin
	// looking at
	int lvl = fastDblToInt(ceil( (log10(maxWidth)/log10(2.0))/3.0 ));

	// adjust for non-subpixel rendering
	if( !subpixel ) {
		lvl--;
	}

	// bump to lowest level if necessary, for very small polys
	lvl = MAX(lvl,0);

	// bump to highest level if necessary, for very large polys
	lvl = MIN(lvl,maxlevel);

	// if the bounding box could fit in a 2x2 region of squares
	// from the next lowest level, we'll use that level instead
	if( (lvl > 0) && (maxWidth < 2*(1<<(3*(lvl-1)))) ) {
		lvl--;
	}

	// calculate the size of the pyramid nodes at the selected level
	int pow;
	if( subpixel ) {
		pow = 1<<(3*lvl);
	} else {
		pow = 1<<(3*(lvl+1));
	}

	// normalize the box coords into the appropriate range
	double xL = low[VX] / pow;
	double xH = high[VX] / pow;
	double yL = low[VY] / pow;
	double yH = high[VY] / pow;

	// find the coordinates of the squares enclosing the box
	xLow = fastDblToInt(floor(xL));
	xHigh = fastDblToInt(ceil(xH)-1);
	yLow = fastDblToInt(floor(yL));
	yHigh = fastDblToInt(ceil(yH)-1);

	// number of subregions in the image at this level
	int numSquares = 1<<(3*(maxlevel-lvl));

	xLow = MAX(0,xLow);
	xHigh = MIN(numSquares-1,xHigh);
	yLow = MAX(0,yLow);
	yHigh = MIN(numSquares-1,yHigh);

	DB_ASSERT( ((xHigh-xLow) <= 2), "ERROR - enclosingSquares screwed up" );
	DB_ASSERT( ((yHigh-yLow) <= 2), "ERROR - enclosingSquares screwed up" );

	// calculate the level of the pyramid to use
	level = maxlevel-lvl;
}


//
//  Access the TriageMask at position (x,y) at the specified level of
//  the pyramid
//
TriageMask& CoveragePyramid::index( int level, int x, int y )
{
	int rowSize = 1<<(3*level);
	int	rowIndex = y*rowSize+x;

	assert( level < maxlevel );

	return pyramid[lvlOffsets[level]+rowIndex];
}

//
//  Access the Bitmask at position (x,y) in the last level of the
//  pyramid.
//
Bitmask& CoveragePyramid::leafIndex( int x, int y )
{
	int rowIndex = y*(1<<(3*maxlevel)) + x;

	return samples[rowIndex];
}

//
//  Determines the image range of a particular node in the quadtree
//
vec4 CoveragePyramid::imageRange( int level, int x, int y )
{
	vec4 range;
	int squareSamples; 
		
	// find the image size of quadtree nodes at this level
	if( subpixel ) {
		squareSamples = 1<<(3*(maxlevel-level));
	} else {
		squareSamples = 1<<(3*(maxlevel-level+1));
	}

	range[XL] = x*squareSamples;
	range[XU] = range[XL]+squareSamples;
	range[YL] = y*squareSamples;
	range[YU] = range[YL]+squareSamples;

	return range;
}

//
//  Fills in all the visibility information below position (x,y)
//  at the specified level
//
void CoveragePyramid::fillBelow( int level, int x, int y )
{
	if( level == maxLevel() ) {

		// if this is the lowest level, just fill in the bitmask for the
		// specified position
		Bitmask& leafBitmask = leafIndex( x, y );
		leafBitmask.fill();

	} else {

		// otherwise, get the triage mask for the specified position and
		// fill it in
		TriageMask& tMask = index(level, x, y);
		tMask.fill();

		// loop through the children of the specified node, filling each one in
		int startX = 8*x;
		int startY = 8*y;

		for( int curX = startX; curX < startX+8; curX++ ) {
			for( int curY = startY; curY < startY+8; curY++ ) {
				fillBelow( level+1, curX, curY );
			}
		}
	}
}
				
//
//  Tiles a polygon having the specified triage mask into the pyramid at the specified position
//  in the specified level.  Returns the W mask, which indicates subregions completely covered
//  by the poly, and the A mask, which represents subregions that are potentially partially covered
//  by the poly
//
void CoveragePyramid::tilePolygon( int level, int x, int y, TriageMask& polyMask, Bitmask& W, Bitmask& A )
{
	// keep track of whether the current node in the pyramid becomes full or becomes active
	// after the addition of this poly.  These flags are used to propagate information back
	// up the quadtree.
	bool becameFull = false;
	bool becameActive = false;
	
	if( level == maxlevel ) {

		// just tile into the specified leaf position in the quadtree
		Bitmask& currentMask = leafIndex(x,y);

		// AND the bitmasks together to get the W mask for the poly
		W = polyMask.C() & ~currentMask;

		#ifdef DEBUG_RENDERDUDE
		if( debug ) {
			cout << "Level " << level << " (" << x << "," << y << ")" << endl;
			cout << "Poly Mask: " << endl << polyMask << endl;
			cout << "Current Mask: " << endl << currentMask << endl;
			cout << "W Mask: " << endl << W << endl;
		}
		#endif

		// if the mask went from empty to non-empty, register this
		if( currentMask.empty() && !W.empty() ) {
			becameActive = true;
		}

		// tile the poly into the coverage pyramid
		currentMask |= W;

		// if the mask went from non-full to full, register this (it must have
		// been non-full before, or we wouldn't be here at all!)
		if( currentMask.full() ) {
			becameFull = 1;
		}

	} else {

		// find the triage mask for the specified quadtree node
		TriageMask& currentMask = index(level,x,y);

		// compute the W and A masks, as in Greene '96
		W = currentMask.V() & polyMask.C();
		A = ~( W | polyMask.V() | currentMask.C() );

		#ifdef DEBUG_RENDERDUDE
		if( debug ) {
			cout << "Level " << level << " (" << x << "," << y << ")" << endl;
			cout << "Poly Mask: " << endl << polyMask << endl;
			cout << "Current Mask: " << endl << currentMask << endl;
			cout << "W Mask: " << endl << W << endl;
			cout << "A Mask: " << endl << A << endl;
		}
		#endif

		// if the mask goes from empty to non-empty, register this
		if( currentMask.V().full() && !((W | A).empty()) ) {
			becameActive = 1;
		}

		// tile the poly into this node
		currentMask.C() |= W;
		currentMask.V() &= ~W;

		// register if it went from non-full to full (again, if it were
		// full before we wouldn't be here)
		if( currentMask.C().full() ) {
			becameFull = 1;
		}
	}

	// propagate any visibility changes back up the pyramid.  This is only
	// necessary if the node went from empty to active or from non-full to full
	while( (becameFull || becameActive) && (level > 0) ) {

		// find the parent node of the current node
		TriageMask& prevMask = index( level-1, x/8, y/8 );

		// keep track, as we may need to propagate even further
		bool prevEmpty = prevMask.empty();

		if( becameFull ) {

			// if the region became full, update the appropriate bits in the parent
			prevMask.C().setBit(x%8,y%8);
			prevMask.V().clearBit(x%8,y%8);

		} else if( becameActive ) {

			// if it became active, clear the vacant bit in the parent
			prevMask.V().clearBit(x%8,y%8);
		}

		// determine whether the parent became full or active
		becameFull = prevMask.full();
		becameActive = prevEmpty;

		// back up to the parent again
		level--;
		x /= 8;
		y /= 8;
	}
}


// 
//  Test to see if any part of the given bounding box is potentially visible through
//  the given position at the given level.  Do not update the visibility information
//  in the pyramid, as we are not actually scan converting the bounding box.
//
//  A return value of false means that the box is completely occluded by the geometry
//  already tiled into the pyramid at the current node.
//
bool CoveragePyramid::isBoundingBoxVisible( BoundingBox& bb, int level, int x, int y )
{
	// determine the image range of the current node
	vec4 iRange = imageRange( level, x, y );

	#ifdef DEBUG_RENDERDUDE
	if( debug ) {
		cout << "Testing bounding box " << bb << " at level " << level 
			 << " (" << x << "," << y << ")" << endl;
		cout << "Image range = " << iRange << endl;
	}
	#endif

	if( level == maxlevel ) {

		// we are at the lowest level of the pyramid, so compare the bb mask against
		// the current mask
		Bitmask& currentMask = leafIndex( x, y );
		Bitmask bbMask = bb.bitMask( iRange );

		// W denotes where the bounding box is visible
		Bitmask W = bbMask & ~currentMask;

		// if the box is partially visible, return true, else false
		if( W.empty() ) {
			return false;
		} else {
			return true;
		}

	} else {

		// get triage masks for the current node and for the bounding box
		TriageMask& currentMask = index( level, x, y );
		TriageMask bbMask = bb.triageMask( iRange );

		// compute the write mask
		Bitmask W = currentMask.V() & bbMask.C();

		if( !W.empty() ) {
			// if a W bit is true, quit now
			return true;

		} else {

			// the W mask is empty, so compute the active mask
			Bitmask A = ~( W | bbMask.V() | currentMask.C() );

			if( A.empty() ) {

				// W and A masks are both empty, so the bounding box
				// is completely occluded (not visible).  Return false.
				return false;

			} else {

				// A mask has some bits set, so recurse on the A mask searching for
				// regions in which the bb is visible

				int numActive = A.numBits();
				int ms = TriageMask::maskSize;

				int curX, curY;
				curX = curY = 0;

				// loop through the active mask
				while( numActive ) {

					if( A.testBit(curX,curY) ) {
						
						if( isBoundingBoxVisible( bb, level+1, ms*x+curX, ms*y+curY ) ) {
							// bounding box is visible in this subregion, so bail
							return true;
						}

						// one less active bit
						numActive--;
					}

					// increment current position
					curX++;								// next col
					curY+=(curX/ms);	// bump to next row if necessary
					curX %= ms;		// wrap back to start if necessary
				}

				// didn't find any subregions in which the bb can be seen, so
				// return false
				return false;
			}
		}
	}
}

//
// debugging function
//
void CoveragePyramid::printMask( int level, int x, int y ) 
{
	if( level == maxlevel ) {
		Bitmask& currentMask = leafIndex(x,y);
		cout << "Level " << level << " (" << x << "," << y << ")" << endl << currentMask;
	} else {
		TriageMask& currentMask = index(level,x,y);
		cout << "Level " << level << " (" << x << "," << y << ")" << endl << currentMask;
	}
}

