#ifndef CoveragePyramid_h
#define CoveragePyramid_h

#include "algebra3.h"

class TriageMask;
class Bitmask;
class Triangle;
class BoundingBox;

/********************************************************************************************************

	Class:		CoveragePyramid
	Purpose:	Image-space structure representing visibility of
				a scene.  Uses quadtree image space subdivision and
				represents visibility using Triage masks and bitmasks.

				Based on Greene's Siggraph '96 paper.

				The pyramid is structured as follows:


									[ ]				level 0 - 1 TriageMask representing
															  the entire image
						[][][][]..........[][][][]	level 1 - 8x8 array of TriageMasks representing
															  the entire image
									 .
									 .
									 .
									 .

				[][][]............................[][][]   level n - 2^(3*n)x2^(3*n) array of Bitmasks
																	 representing the entire image

				If the 'subpixel' mode is turned on, level n is such
				that each bitmask represents a single pixel, otherwise
				each bitmask at the lowest level represents an 8x8 block
				of pixels.


*********************************************************************************************************/
class CoveragePyramid
{
public:

	// initialization - determines image resolution and number of
	// visibility samples per pixe (subpixel = true -> 64 visibility
	// samples per pixel, subpixe = false -> 1 sample per pixel)
	CoveragePyramid( int resolution = 512, bool subPixel = true );
	~CoveragePyramid();

	// returns the number of levels in the pyramid
	int maxLevel();

	// determines whether or not subpixel visibility determination is enabled
	bool subPixel();

	// returns the TriageMask representing accumulated visibility for
	// a particular node in the pyramid.  Use this function to access
	// all levels of the pyramid except the last one
	TriageMask& index( int level, int x, int y );

	// this function is used to get visibility info for the leaf nodes
	// in the coverage pyramid
	Bitmask& leafIndex( int x, int y );

	// fills in the visibility for all samples below a particular node in the
	// pyramid
	void fillBelow( int level, int x, int y );

	// returns the image space range represented by a particular node in the 
	// pyramid
	vec4 imageRange( int level, int x, int y );

	// tiles a polygon (represented by a TriageMask) into the pyramid at index (x,y) of
	// the specified level.  Returns the write and active masks through parameters W and A.
	void tilePolygon( int level, int x, int y, TriageMask& polyMask, Bitmask& W, Bitmask& A );

	// finds the smallest pyramid node whose image region completely covers that of the given
	// bounding box
	void enclosingLevel( BoundingBox& bb, int &level, int& x, int& y );

	// finds the smallest possible set of pyramid nodes (up to 2x2) that completely cover
	// the given bounding box.  
	void enclosingSquares( BoundingBox& bb, int &level, int& xL, int& xH, int& yL, int& yH );

	// determines whether a particular node is covered by starting at the root and descending down
	// the pyramid towards the specified node, looking for a "C" bit 
	bool isNodeVisible( int level, int x, int y );
	
	// starting at the given level and position in the pyramid, tests to see if any part of
	// the given bounding box is potentially visible
	bool isBoundingBoxVisible( BoundingBox& bb, int level, int x, int y );

	// clears the pyramid
	void clear();

	// debugging functions
	void setDebug( bool d ) { debug = d; };
	void printMask( int level, int x, int y );

private:

	int maxlevel;			// index of last level in the pyramid
	int resolution;			// image resolution (usually 512)
	bool subpixel;			// is subpixel visibility on?

	int pyramidSize;		// total number of triagemasks in the pyramid
	int leafSize;			// total numbert of leaf nodes
	int *lvlOffsets;		// offsets for various levels in the coverage
							// pyramid

	TriageMask *pyramid;	// Triage masks for all levels of the pyramid 
							// except the last
	Bitmask *samples;		// Bitmasks representing the last level of the pyramid

	bool debug;				// debugging flag

};

#endif