#include "Edge.h"
#include "TriageMask.h"
#include <limits>

// for numeric_limits stuff
using namespace std;

// 
//  Constructor initializes vertices, deltas
//
Edge::Edge( vec3& V1, vec3& V2 ) 
{
	// initialize vertices
	v[0] = vec2(V1,VZ); v[1] = vec2(V2,VZ);

	// initialize deltas
	deltaV = v[1] - v[0];
	dx = deltaV[VX];
	dy = deltaV[VY];

	// compute slope and inverse slope
	if( fabs(dx) < (numeric_limits<double>::epsilon()) ) {

		// horizontal line
		invSlope = 0.0;
		slope = numeric_limits<double>::infinity();

	} else if( fabs(dy) < (numeric_limits<double>::epsilon()) ) {

		// vertical line
		slope = 0.0;
		invSlope = numeric_limits<double>::infinity();

	} else {

		// regular line
		slope = dy / dx;
		invSlope = 1.0 / slope;
	}
}

//
//  Intersect the edge with the given box.  If 'correct' is set, we round the
//  intersection points to the outside of the area enclosed by the edge.
//
//	Edges are clipped to the box boundary and/or extended to the box boundary,
//  if necessary.  It also MUST maintain the correct ordering of the vertices 
//  on the edge.
//
OverlapType Edge::boxIntersect( vec4& box, int *coords, bool correct )
{
	int whichCoord = 0;				// keeps track of how many intersections we've found
	int tempCoords[4];				// stores intersections with each face
	int whichValid[4] = {0,0,0,0};	// which intersections are valid?
	vec2 boxCoords[4];				// coordinates of intersection points
	double tXL, tXU, tYL, tYU;		// parametric values to each box side
	double xYL, xYU, yXL, yXU;		// intersection points

	// if not vertical, intersect with the left and right of the box
	if( dx != 0.0 ) {

		// calculate parametric values to left and right sides
		tXU = (box[XU]-v[0][VX])/dx;
		tXL = (box[XL]-v[0][VX])/dx;

		// calculate y values of intersection points
		yXU = v[0][VY] + tXU*dy;
		yXL = v[0][VY] + tXL*dy;

		// mark left and right intersections valid
		whichValid[XU] = whichValid[XL] = 1;

	} 

	// if not horizontal, intersect with the top and bottom of the box
	if( dy != 0.0 ) {

		// calculate parametric values to top and bottom
		tYU = (box[YU]-v[0][VY])/dy;
		tYL = (box[YL]-v[0][VY])/dy;

		// calculate x values of top and bottom
		xYU = v[0][VX] + tYU*dx;
		xYL = v[0][VX] + tYL*dx;

		// mark top and bottom intersections as valid
		whichValid[YU] = whichValid[YL] = 1;

	} 

	// if right intersection point is valid, calculate the intersection point with the
	// perimeter
	if( whichValid[XU] && ((box[YL] <= yXU) && (box[YU] >= yXU)) ) {

		// store the intersection point
		boxCoords[whichCoord] = vec2(box[XU],yXU);

		// calculate the index
		tempCoords[whichCoord] = getIndex(box, boxCoords[whichCoord], XU);
		
		// store the index, if it is new
		if( (whichCoord == 0) || (tempCoords[whichCoord] != tempCoords[whichCoord-1]) ) {
			whichCoord++;
		}
	}

	// same thing for left intersection point
	if( whichValid[XL] && ((box[YL] <= yXL) && (box[YU] >= yXL)) ) {
		boxCoords[whichCoord] = vec2(box[XL],yXL);
		tempCoords[whichCoord] = getIndex(box, boxCoords[whichCoord], XL);

		if( (whichCoord == 0) || (tempCoords[whichCoord] != tempCoords[whichCoord-1]) ) {
			whichCoord++;
		}
	}

	// same thing for top intersection point
	if( whichValid[YU] && ((box[XL] <= xYU) && (box[XU] >= xYU)) ) {
		boxCoords[whichCoord] = vec2(xYU, box[YU]);
		tempCoords[whichCoord] = getIndex(box, boxCoords[whichCoord], YU);

		if( (whichCoord == 0) || (tempCoords[whichCoord] != tempCoords[whichCoord-1]) ) {
			whichCoord++;
		}
	}

	// same thing for bottom intersection point
	if( whichValid[YL] && ((box[XL] <= xYL) && (box[XU] >= xYL)) ) {
		boxCoords[whichCoord] = vec2(xYL, box[YL]);
		tempCoords[whichCoord] = getIndex(box, boxCoords[whichCoord], YL);

		if( (whichCoord == 0) || (tempCoords[whichCoord] != tempCoords[whichCoord-1]) ) {
			whichCoord++;
		}
	}

	// check to see how many box intersections we got
	if( whichCoord >= 2 ) {

		// if 2 or more, use the first two intersections (there really shouldn't be any
		// more, right?)

		// calculate dx, dy of clipped/extended edge
		double newDx = boxCoords[1][VX] - boxCoords[0][VX];
		double newDy = boxCoords[1][VY] - boxCoords[0][VY];

		// see if the clipping/extending operation screwed up the ordering of the vertices,
		// and flip them if it did
		if( ( ((newDx >= 0.0) && (dx >= 0.0)) || ((newDx <= 0.0) && (dx <= 0.0))) && 
			( ((newDy >= 0.0) && (dy >= 0.0)) || ((newDy <= 0.0) && (dy <= 0.0))) ) {
			coords[0] = tempCoords[0];
			coords[1] = tempCoords[1];
		} else {
			coords[0] = tempCoords[1];
			coords[1] = tempCoords[0];
		}

		// see if both intersection points are on the same side of the box
		OverlapType overLap = sameSide( coords[0], coords[1] );

		if( (overLap == SOME) && correct ) {

			// apply correction if necessary.  Round to outside of area enclosed by
			// edge.  It turns out that all we need to do is bump up the first intersection
			// and bump down the second intersection

			// bump up the first intersection
			coords[0] = (coords[0] + 1) % (NUMSUB-1);
			
			// bump down the second intersection
			if( coords[1] == 0 ) {
				coords[1] = 511;
			} else {
				coords[1] = (coords[1] - 1) % (NUMSUB-1);
			}

			// if the two coordinates became the same, then the whole square is covered
			if( coords[0] == coords[1] ) {
				return ALL;
			} else {
				return SOME;
			}

		} else {

			// both intersection points are on the same side, so the entire box
			// is either inside or outside the edge (already determined by the
			// call to sameSide)
			return overLap;

		}

	} else {

		// we only have one intersection point, so the entire region is either
		// inside or outside the edge.  Determine which is the case by taking the
		// cross product of the directed edge with the line from the first vertex to
		// the centre of the pixel.  If the cross product points "out" of the image plane,
		// then the whole region is covered.

		// calculate vector from first edge vertex to region centre
		vec2 midpt( 0.5*(box[XL]+box[XU]), 0.5*(box[YL]+box[YU]) );
		vec2 v1ToMidpt = midpt - vec2(v[0],VZ);

		// do cross product
		vec3 cross = vec2(deltaV,VZ) ^ v1ToMidpt;

		// check third coordinate of cross product to see if region is inside or outside
		if( cross[VZ] > 0.0 ) {
			return ALL;
		} else {
			return NONE;
		}	
	} 

}


//
//  Given a point and a screenspace box and a side of the box, finds
//  the index of the point on the box perimeter.  If the suggested 
//  side is -1, try all the sides
//
int Edge::getIndex( vec4& box, vec2& vertex, int whichSide  )
{
	int result;
	double xLen = box[XU] - box[XL];
	double yLen = box[YU] - box[YL];

	// how many subintervals on a side
	int side = TriageMask::subIntervals * TriageMask::maskSize;

	switch( whichSide ) {
	case YL:
		// bottom = [0..128]
		result = fastDblToInt(ROUND(((vertex[VX]-box[XL])/xLen)*side));
		break;
	case XU:
		// right = [128..256]
		result = side + fastDblToInt(ROUND(((vertex[VY]-box[YL])/yLen)*side));
		break;
	case YU:
		// top = [256..384]
		result = (3*side) - fastDblToInt(ROUND(((vertex[VX]-box[XL])/xLen)*side));
		break;
	case XL:
		// left = [384..512]
		result = (4*side) - fastDblToInt(ROUND(((vertex[VY]-box[YL])/yLen)*side));
		break;
	default:

		// try all 4 sides
		if( vertex[VY] == box[YL] ) {
			result = fastDblToInt(ROUND(((vertex[VX]-box[XL])/xLen)*side));
		} else if( vertex[VX] == box[XU] ) {
			result = side + fastDblToInt(ROUND(((vertex[VY]-box[YL])/yLen)*side));
		} else if( vertex[VY] == box[YU] ) {
			result = (3*side) - fastDblToInt(ROUND(((vertex[VX]-box[XL])/xLen)*side));
		} else if( vertex[VX] == box[XL] ) {
			result = (4*side) - fastDblToInt(ROUND(((vertex[VY]-box[YL])/yLen)*side));
		} else {
			assert(false);
		}
		break;
	}

	// make sure in range [0..512]
	return result%(4*side);
}


//
//  Are two points on the perimeter on the same side of the box?
//
OverlapType Edge::sameSide( int v1, int v2 )
{
	// how many subintervals on a side
	int side = TriageMask::subIntervals * TriageMask::maskSize;

	// sides that v1 and v2 are on
	int v1sides[2], v2sides[2];

	// figure out which sides v1 is on
	v1sides[0] = (v1 / side)%4;
	if( (v1 % side) == 0 ) {

		if( v1sides[0] == 0 ) {
			v1sides[1] = 3;
		} else {
			v1sides[1] = v1sides[0]-1;
		}
	} else {
		v1sides[1] = -100;
	}

	// figure out which sides v2 is on
	v2sides[0] = (v2 / side)%4;
	if( (v2 % side) == 0 ) {
		if( v2sides[0] == 0 ) {
			v2sides[1] = 3;
		} else {
			v2sides[1] = v2sides[0]-1;
		}
	} else {
		v2sides[1] = -99;
	}

	// see if there is a side that both v1 and v2 are on
	int whichSide = -1;
	for( int i = 0; i < 2; i++ ) {
		for( int j = 0; j < 2; j++ ) {

			if( v1sides[i] == v2sides[j] ) {

				whichSide = v1sides[i];

			}
		}
	}

	if( whichSide == -1 ) {

		// they are not on the same side
		return SOME;

	} else {

		// they are on the same side, so adjust the indices
		// to reflect which side they are on

		if( (v1 == 512) && (whichSide == 0) ) {
			v1 = 0;
		}

		if( (v2 == 512) && (whichSide == 0) ) {
			v2 = 0;
		}

		if( (v1 == 0) && (whichSide == 3) ) {
			v1 = 512;
		}

		if( (v2 == 0) && (whichSide == 3) ) {
			v2 = 512;
		}

		// if v2 > v1, the square is inside the edge, otherwise it is
		// outside the edge
		return v2 > v1 ? ALL : NONE;

	}
}
































