#include "EdgeTable.h"
#include "Triangle.h"
#include <limits>

using namespace std;

/****************************************************************

	Class:		EdgeTableEntry
	Purpose:		

*****************************************************************/
class EdgeTableEntry
{
public:
	EdgeTableEntry();
	EdgeTableEntry( int yMax, double xMin, double invSlope, vec3& bary, LerpVec3& coords );

	friend class EdgeTable;
private:
	int yMax;
	double xMin;
	double invSlope;
	vec3 bary;
	LerpVec3 coords;
};



EdgeTableEntry::EdgeTableEntry() 
: yMax(0), xMin(0.0), invSlope(0.0), bary(0.0)
{
}

EdgeTableEntry::EdgeTableEntry( int ymax, double xmin, double invslope, vec3& b, LerpVec3& c ) 
: yMax(ymax), xMin(xmin), invSlope(invslope), bary(b), coords(c)
{
}


//
//  Intialize edge table, active edge table
//
EdgeTable::EdgeTable( vec2& yRange, bool d )
{
	debug = d;

	// initialize table range
	tableYMin = (int)yRange[0];
	tableYMax = (int)yRange[1];
	numScanlines = tableYMax-tableYMin+1;

	// initialize edge table entries
	table = new EdgeTableEntry[2*numScanlines];
	tableIndices = new int[numScanlines];

	// initialize active edge table
	activeEdges = new EdgeTableEntry[2];

	// clear table
	clear();
}

//
//  Destructor - clean up structures
//
EdgeTable::~EdgeTable()
{
	if( table ) {
		delete [] table;
	}

	if( tableIndices ) {
		delete [] tableIndices;
	}

	if( activeEdges ) {
		delete [] activeEdges;
	}
}

//
//  Clear the edge table - just zero the index array
//
void EdgeTable::clear(void)
{
	memset( tableIndices, 0, numScanlines*sizeof(int) );
}

//
//  Adds a triangle to the edge table.  Each edge has an EdgeTableEntry in the table
//  position corresponding to the lowest y coordinate of the edge. 
//
void EdgeTable::addTriangle( Triangle& tri )
{
	vec3 affineCoords[3];

	// affine coordinates of the triangle vertices
	affineCoords[0] = vec3(1.0,0.0,0.0);
	affineCoords[1] = vec3(0.0,1.0,0.0);
	affineCoords[2] = vec3(0.0,0.0,1.0);

	for( int curEdge = 0; curEdge < 3; curEdge++ ) {

		// denotes whether this edge should be shortened
		bool shorten = false;

		// get current and next two vertices in counterclockwise order
		int i1 = curEdge;
		int i2 = (curEdge+1)%3;
		int i3 = (curEdge+2)%3;
		vec3& v1 = tri.screen(i1);
		vec3& v2 = tri.screen(i2);
		vec3& v3 = tri.screen(i3);

		// get current edge deltas
		double dy = v2[VY] - v1[VY];
		double dx = v2[VX] - v1[VX];

		// info about where edge lies relative to scanlines
		bool v1OnScanline = (floor(v1[VY]) == v1[VY]);
		bool v2OnScanline = (floor(v2[VY]) == v2[VY]);
		bool crossesScanline = (dy!=0.0) && ((v1OnScanline || v2OnScanline) || (floor(v1[VY]) != floor(v2[VY]) ));

		#ifdef DEBUG_RENDERDUDE
		if( debug ) {
			cout << "Edge table: adding new edge " << v1 << " -> " << v2 << endl;
			cout << "V1 is " << (v1OnScanline ? "" : "not") << " on scanline" << endl;
			cout << "V2 is " << (v2OnScanline ? "" : "not") << " on scanline" << endl;
			cout << "Edge does " << (crossesScanline ? "" : "not") << " cross scanline" << endl;
		}
		#endif

		// skip edge if it doesn't cross a scanline
		if( !crossesScanline ) {
			continue;
		}

		// calculate slopes
		double slope, invSlope;
		if( fabs(dx) < (numeric_limits<double>::epsilon()) ) {
			invSlope = 0.0;
			slope = numeric_limits<double>::infinity();
		} else {
			slope = dy / dx;
			invSlope = 1.0 / slope;
		}

		double yMax, yMin, xMin;
		int tableRow, tableEntry;
		double bumpUp;
		vec3 lowCoords, deltaCoords;

		// if v2 lies on a scanline, shorten the edge if the slope doesn't change
		// signs at v2
		if( v2[VY] == floor(v2[VY]) ) {
			double nextDy = v3[VY] - v2[VY];
			shorten = sameSign( dy, nextDy );
		}

		// next, we round the lowest edge to the first scanline it crosses, 
		// and apply any edge shortening

		if( dy > 0.0 ) {
			
			// shorten edge if necessary
			yMax = floor( v2[VY] );
			if( shorten ) {
				yMax -= 1.0;
			}

			// bump lower vertex up to first scanline it crosses
			yMin = ceil( v1[VY] );

			// bump lower vertex to bottom of image, if necessary
			yMin = MAX( yMin, 0.0 );

			// discard if edge switches orientation
			if( yMax < yMin ) {
				continue;
			}

			// calculate amount edge gets bumped up to scanline
			bumpUp = yMin - v1[VY];

			// bump x coord of low point up
			xMin = v1[VX] + bumpUp * invSlope;

			// compute change in affine coords on current edge
			deltaCoords = (affineCoords[i2] - affineCoords[i1])/dy;

			// store affine coords of low vertex
			lowCoords = affineCoords[i1];
		
		} else {  

			// shorten edge if necessary
			yMax = floor( v1[VY] );
			yMin = ceil( v2[VY] );
			if( shorten ) {
				yMin += 1.0;
			}

			// bump lower vertex to bottom of image, if necessary
			yMin = MAX( yMin, 0.0 );

			// skip if orientation changes
			if( yMax < yMin ) {
				continue;
			}

			// bump v2 up to lowest scanline it crosses
			bumpUp = yMin - v2[VY];

			// calculate x coord at lowest scanline
			xMin = v2[VX] + bumpUp * invSlope;

			// compute affine coord change for this edge
			deltaCoords = (affineCoords[i2] - affineCoords[i1])/dy;

			// store affine coords of low point
			lowCoords = affineCoords[i2];

		}

		// initialize linear interpolator of affine coords along the current
		// edge, and bump up to lowest scanline
		LerpVec3 lerpCoords( lowCoords, deltaCoords );
		lerpCoords.increment(bumpUp);

		// determine which row in which to store the entry for this edge
		tableRow = fastDblToInt(( yMin - tableYMin ));

		// if above image range, discard
		if( tableRow > (numScanlines-1) ) {
			continue;
		}

		// find the index of the EdgeTableEntry for the edge
		tableEntry = 2*tableRow + tableIndices[tableRow];

		// store the edge data in the edge table
		table[tableEntry].invSlope = invSlope;
		table[tableEntry].yMax = fastDblToInt(yMax);
		table[tableEntry].xMin = xMin;
		table[tableEntry].coords = lerpCoords;

		// increment the number of edges starting on this scanline
		tableIndices[tableRow] += 1;

		DB_ASSERT( (tableIndices[tableRow] <= 2), "ERROR - too many edges on scanline" );

		// if we have 2 edge table entries for this scanline, make sure they are sorted in 
		// x
		if( tableIndices[tableRow] == 2 ) {
			EdgeTableEntry& edge1 = table[2*tableRow];
			EdgeTableEntry& edge2 = table[2*tableRow+1];

			if( edge1.xMin > edge2.xMin ) {
				EdgeTableEntry temp( edge1 );
				edge1 = edge2;
				edge2 = temp;
			}
		}
			
		#ifdef DEBUG_RENDERDUDE
		if( debug ) {
			cout << endl << "New Entry: " << endl;
			cout << "Scanline " << tableRow + tableYMin << endl;
			cout << "invSlope = " << table[tableEntry].invSlope << endl;
			cout << "yMax = " << table[tableEntry].yMax << endl;
			cout << "xMin = " << table[tableEntry].xMin << endl;
		}
		#endif

	}

}

//
//  Begins a traversal of the edge table for the current triangle
//
bool EdgeTable::startTraversal()
{
	currentScanline = 0;

	// search through the table looking for the first edge table entry
	for( currentScanline = 0; currentScanline < numScanlines; currentScanline++ ) {
		if( tableIndices[currentScanline] != 0 ) {
			break;
		}
	}

	if( currentScanline == numScanlines ) {
		// if no edges, return
		return false;

	} else {
		// initialize the active edge table - there must be 2 edge table entries
		// at the lowest scanline
		DB_ASSERT( (tableIndices[currentScanline] == 2), "ERROR - inconsistent edge table" );

		activeEdges[0] = table[2*currentScanline];
		activeEdges[1] = table[2*currentScanline+1];

		return true;
	}
}

//
//  Gets information about the current span in an edge table traversal.  
//  'xRange' gives the x-extent of the span, and 'coords' can be used to
//  interpolate affine coordinates across the span.  
//
bool EdgeTable::getCurrentSpan( int& scanLine, vec2& xRange, LerpVec3& coords )
{
	// find screenspace y value of current scanline
	scanLine = currentScanline + tableYMin;

	// determine whether either of the left and right active edges are finished
	int leftDone = scanLine > activeEdges[0].yMax;
	int rightDone = scanLine > activeEdges[1].yMax;

	// return if no more edges to process
	if( (scanLine > tableYMax) || 
		(leftDone && rightDone && (tableIndices[currentScanline] == 0)) ) {
		return false;
	} 

	// if the left edge is done, we need to replace it, so there must be a new edge starting
	// on this scanline
	if( leftDone ) {
		DB_ASSERT( (tableIndices[currentScanline] != 0), "ERROR - edge table left error" );
		activeEdges[0] = table[2*currentScanline];
	}

	// if the right edge is done, we need to replace it, so there must be a new edge starting 
	// on this line
	if( rightDone ) {
		DB_ASSERT( (tableIndices[currentScanline] != 0), "ERROR - edge table right error" );
		activeEdges[1] = table[(2*currentScanline)+(tableIndices[currentScanline]-1)];
	}

	// swap the active edges if they get out of order in x
	if( activeEdges[0].xMin > activeEdges[1].xMin ) {
		EdgeTableEntry temp = activeEdges[0];
		activeEdges[0] = activeEdges[1];
		activeEdges[1] = temp;
	}

	// fill in the x range
	xRange[0] = activeEdges[0].xMin;
	xRange[1] = activeEdges[1].xMin;

	// create a linear interpolant for the affine coords across this scanline from the left and
	// right edge interpolants
	coords = LerpVec3( activeEdges[0].coords.current(), activeEdges[1].coords.current(), xRange );

	// round the left end of the span to the nearest sample point and increment the 
	// interpolant, if necessary
	if( fabs( xRange[0] - ROUND(xRange[0]) ) < 0.000001 ) {
		xRange[0] = ROUND(xRange[0]);
	} else {
		double leftCeil = ceil(xRange[0]);
		coords.increment( leftCeil - xRange[0] );
		xRange[0] = leftCeil;
	}

	// round the right end of the span to the nearest sample point
	if( fabs( xRange[1] - ROUND(xRange[1]) ) < 0.000001 ) {
		xRange[1] = ROUND(xRange[1]);
	} else {
		xRange[1] = floor(xRange[1]);
	}

	// increment the left interpolant to the next scanline
	activeEdges[0].xMin += activeEdges[0].invSlope;
	activeEdges[0].coords.increment();

	// increment the right interpolant to the next scanline
	activeEdges[1].xMin += activeEdges[1].invSlope;
	activeEdges[1].coords.increment();

	// clear the table as we go through it
	tableIndices[currentScanline] = 0;

	// increment the scanline pointer
	currentScanline += 1;
	return true;
}


