#include "Face.h"
#include "Cam.h"


Face::Face( int numPoints )
: wPoints(numPoints), cPoints(numPoints), lPoints(numPoints), 
  sPoints(numPoints), tPoints(numPoints), normals(numPoints), 
  pEq(0.0), shader(NULL)
{
}

Face::Face( const Face& other )
{
	(*this) = other;
}

Face& Face::operator=( const Face& rhs )
{
	wPoints = rhs.wPoints;
	cPoints = rhs.cPoints;
	lPoints = rhs.lPoints;
	sPoints = rhs.sPoints;
	tPoints = rhs.tPoints;
	normals = rhs.normals;
	pEq = rhs.pEq;
	shader = rhs.shader;

	return (*this);
}

void Face::setTexCoords( const vec2& tc ) 
{
	int numPoints = size();

	for( int i = 0; i < numPoints; i++ ) {
		tPoints[i] = tc;
	}
}

void Face::setNormals( const vec3& n ) 
{
	int numPoints = size();

	for( int i = 0; i < numPoints; i++ ) {
		normals[i] = n;
	}
}

void Face::cameraSpace( const Cam& camera )
{
	int numPoints = size();
	for( int i = 0; i < numPoints; i++ ) {
		cPoints[i] = camera.worldToCamera( wPoints[i] );
	}
}

void Face::screenSpace( const Cam& camera )
{
	int numPoints = size();
	for( int i = 0; i < numPoints; i++ ) {
		sPoints[i] = camera.cameraToScreen( cPoints[i] );
	}
}

void Face::resize( int numPoints ) 
{
	wPoints.resize(numPoints);
	cPoints.resize(numPoints);
	lPoints.resize(numPoints);
	sPoints.resize(numPoints);
	tPoints.resize(numPoints);
	normals.resize(numPoints);
}

vec3& Face::point( int i, SpaceType space ) 
{
	switch( space ) {

	case WORLD:
		return wPoints[i];
		break;
	case LOCAL:
		return lPoints[i];
		break;
	case CAMERA:
		return cPoints[i];
		break;
	default:
		assert(false);
		return vec3(0.0);
		break;
	}
}

IsectType Face::clip( const vec4& clipPlane, SpaceType space )
{
	int numPoints = size();

	List<SideType> inOuts;
	List<vec3> points;
	bool allInside = true;
	bool allOutside = true;

	List<vec3> newWorld;
	List<vec3> newCamera;
	List<vec3> newLocal;
	List<vec2> newTexture;
	List<vec3> newNormal;

	for( int i = 0; i < numPoints; i++ ) {

		points.append( point( i, space ) );

		SideType curInOut = inside( points[i], clipPlane );

		inOuts.append( curInOut );

		allInside &= (curInOut == INSIDE);
		allOutside &= (curInOut == OUTSIDE);
	}

	if( allInside ) {
		return POS;
	} 

	if( allOutside ) {
		return NEG;
	}

	for( i = 0; i < numPoints; i++ ) {

		double t;
		int i1 = i;
		int i2 = (i+1)%numPoints;

		int v1where = inOuts[i1];
		int v2where = inOuts[i2];

		vec3 v1 = points[i1];
		vec3 v2 = points[i2];

		if( linePlaneIsect( v1, v2, clipPlane, t ) ) {

			vec3 isectWorld = (1.0-t)*wPoints[i1] + t*wPoints[i2];
			vec3 isectCamera = (1.0-t)*cPoints[i1] + t*cPoints[i2];
			vec3 isectLocal = (1.0-t)*lPoints[i1] + t*lPoints[i2];
			vec2 isectTexture = (1.0-t)*tPoints[i1] + t*tPoints[i2];
			vec3 isectNormal = (1.0-t)*normals[i1] + t*normals[i2];

			if( v1where == INSIDE ) {
					
				newWorld.append( wPoints[i1] );
				newCamera.append( cPoints[i1] );
				newLocal.append( lPoints[i1] );
				newTexture.append( tPoints[i1] );
				newNormal.append( normals[i1] );

				if( v2where == OUTSIDE ) {

					newWorld.append( isectWorld );
					newCamera.append( isectCamera );
					newLocal.append( isectLocal );
					newTexture.append( isectTexture );
					newNormal.append( isectNormal );

				}

			} else if( v1where == INPLANE ) {

				newWorld.append( wPoints[i1] );
				newCamera.append( cPoints[i1] );
				newLocal.append( lPoints[i1] );
				newTexture.append( tPoints[i1] );
				newNormal.append( normals[i1] );

			} else if( v1where == OUTSIDE ) {

				if( v2where == INSIDE ) {

					newWorld.append( isectWorld );
					newCamera.append( isectCamera );
					newLocal.append( isectLocal );
					newTexture.append( isectTexture );
					newNormal.append( isectNormal );

				}
			}

		} else {

			if( v1where == INSIDE ) {

				newWorld.append( wPoints[i1] );
				newCamera.append( cPoints[i1] );
				newLocal.append( lPoints[i1] );
				newTexture.append( tPoints[i1] );
				newNormal.append( normals[i1] );

			}
		}
	}

	wPoints = newWorld;
	cPoints = newCamera;
	lPoints = newLocal;
	tPoints = newTexture;
	normals = newNormal;
	sPoints.resize( newWorld.size() );
	
	return BOTH;
}

IsectType Face::clip( const vec4& clipPlane, SpaceType space, Face& newFace )
{
	newFace = (*this);
	return newFace.clip( clipPlane, space );
}

IsectType Face::split( const vec4& clipPlane, SpaceType space, Face& posFace, Face& negFace )
{
	int numPoints = size();

	List<SideType> inOuts( numPoints );
	List<vec3> points( numPoints );
	bool allInside = true;
	bool allOutside = true;

	List<vec3> posWorld, negWorld;
	List<vec3> posCamera, negCamera;
	List<vec3> posLocal, negLocal;
	List<vec2> posTexture, negTexture;
	List<vec3> posNormal, negNormal;

	for( int i = 0; i < numPoints; i++ ) {

		points.append( point( i, space ) );
		
		SideType curInOut = inside( points[i], clipPlane );

		inOuts.append( curInOut );

		allInside &= ((curInOut == INSIDE) || (curInOut == INPLANE));
		allOutside &= ((curInOut == OUTSIDE) || (curInOut == INPLANE));
	}

	if( allInside ) {
		return POS;
	} 

	if( allOutside ) {
		return NEG;
	}

	posFace.pEq = pEq;
	negFace.pEq = pEq;
	posFace.shader = shader;
	negFace.shader = shader;

	for( i = 0; i < numPoints; i++ ) {

		double t;
		int i1 = i;
		int i2 = (i+1)%numPoints;

		int v1where = inOuts[i1];
		int v2where = inOuts[i2];

		vec3 v1 = point( i1, space );
		vec3 v2 = point( i2, space );

		if( linePlaneIsect( v1, v2, clipPlane, t ) ) {

			vec3 isectWorld = (1.0-t)*wPoints[i1] + t*wPoints[i2];
			vec3 isectCamera = (1.0-t)*cPoints[i1] + t*cPoints[i2];
			vec3 isectLocal = (1.0-t)*lPoints[i1] + t*lPoints[i2];
			vec2 isectTexture = (1.0-t)*tPoints[i1] + t*tPoints[i2];
			vec3 isectNormal = (1.0-t)*normals[i1] + t*normals[i2];


			if( v1where == INSIDE ) {
					
				posWorld.append( wPoints[i1] );
				posCamera.append( cPoints[i1] );
				posLocal.append( lPoints[i1] );
				posTexture.append( tPoints[i1] );
				posNormal.append( normals[i1] );

				if( v2where == OUTSIDE ) {

					posWorld.append( isectWorld );
					posCamera.append( isectCamera );
					posLocal.append( isectLocal );
					posTexture.append( isectTexture );
					posNormal.append( isectNormal );

					negWorld.append( isectWorld );
					negCamera.append( isectCamera );
					negLocal.append( isectLocal );
					negTexture.append( isectTexture );
					negNormal.append( isectNormal );
				}

			} else if( v1where == INPLANE ) {

				posWorld.append( wPoints[i1] );
				posCamera.append( cPoints[i1] );
				posLocal.append( lPoints[i1] );
				posTexture.append( tPoints[i1] );
				posNormal.append( normals[i1] );

				negWorld.append( wPoints[i1] );
				negCamera.append( cPoints[i1] );
				negLocal.append( lPoints[i1] );
				negTexture.append( tPoints[i1] );
				negNormal.append( normals[i1] );


			} else if( v1where == OUTSIDE ) {

				negWorld.append( wPoints[i1] );
				negCamera.append( cPoints[i1] );
				negLocal.append( lPoints[i1] );
				negTexture.append( tPoints[i1] );
				negNormal.append( normals[i1] );

				if( v2where == INSIDE ) {

					posWorld.append( isectWorld );
					posCamera.append( isectCamera );
					posLocal.append( isectLocal );
					posTexture.append( isectTexture );
					posNormal.append( isectNormal );

					negWorld.append( isectWorld );
					negCamera.append( isectCamera );
					negLocal.append( isectLocal );
					negTexture.append( isectTexture );
					negNormal.append( isectNormal );
				}
			}

		} else {

			if( v1where == INSIDE ) {

				posWorld.append( wPoints[i1] );
				posCamera.append( cPoints[i1] );
				posLocal.append( lPoints[i1] );
				posTexture.append( tPoints[i1] );
				posNormal.append( normals[i1] );

			} else {

				negWorld.append( wPoints[i1] );
				negCamera.append( cPoints[i1] );
				negLocal.append( lPoints[i1] );
				negTexture.append( tPoints[i1] );
				negNormal.append( normals[i1] );

			}
		}
	}

	if( posWorld.size() > 0 ) {

		posFace.wPoints = posWorld;
		posFace.cPoints = posCamera;
		posFace.lPoints = posLocal;
		posFace.tPoints = posTexture;
		posFace.normals = posNormal;
		posFace.sPoints.resize( posWorld.size() );

	}

	if( negWorld.size() > 0 ) {

		negFace.wPoints = negWorld;
		negFace.cPoints = negCamera;
		negFace.lPoints = negLocal;
		negFace.tPoints = negTexture;
		negFace.normals = negNormal;
		negFace.sPoints.resize( negWorld.size() );

	}

	if( (posWorld.size() > 0) && (negWorld.size() > 0) ) {
		return BOTH;
	} else if( posWorld.size() > 0 ) {
		return POS;
	} else if( negWorld.size() > 0 ) {
		return NEG;
	}

	return BOTH;
}

	
IsectType Face::bboxIsect( const BoundingBox& bb, SpaceType space )
{
	bool inside = false;
	bool outside = false;
	int numPoints = wPoints.size();
	vec3 low = bb.lowBound();
	vec3 high = bb.highBound();

	for( int i = 0; i < numPoints; i++ ) {

		vec3 curPoint( point( i, space ) );

		if( curPoint[VX] < low[VX] ||
			curPoint[VY] < low[VY] ||
			curPoint[VZ] < low[VZ] ||
			curPoint[VX] > high[VX] ||
			curPoint[VY] > high[VY] ||
			curPoint[VZ] > high[VZ] ) {

			outside = true;
		} else {
			inside = true;
		}

		if( inside && outside ) {
			return BOTH;
		}
	}

	if( inside ) {
		return POS;
	} else {
		return NEG;
	}
}

IsectType Face::clipToBBoxOctant( BoundingBox& bb, unsigned char octant, SpaceType space, Face& newFace )
{
	IsectType isect = isectBBoxOctant( bb, octant, space );
	vec4 clipPlanes[3];

	bb.clipPlanes( octant, clipPlanes );

	switch( isect ) {

	case POS:
		newFace = (*this);
		break;

	case NEG:
		break;

	case BOTH:
		newFace = (*this);

		for( int i = 0; i < 3; i++ ) {
			(void)newFace.clip( clipPlanes[i], space );
		}
		break;
	}

	return isect;
}

IsectType Face::isectBBoxOctant( BoundingBox& bb, unsigned char octant, SpaceType space )
{
	BoundingBox octantBB = bb.subBox( octant );

	return bboxIsect( octantBB, space );
}


void Face::tessellate( List<Triangle>& newTriangles )
{
/*	int numPoints = size();

	if( numPoints == 3 ) {

		Triangle newTri;

		for( int i = 0; i < 3; i++ ) {
			newTri.world(i) = wPoints[i];
			newTri.camera(i) = cPoints[i];
			newTri.local(i) = lPoints[i];
			newTri.texture(i) = tPoints[i];
			newTri.screen(i) = sPoints[i];
			newTri.normal(i) = normals[i];
		}

		newTri.plane() = pEq;
		newTri.setShader( shader );

		newTriangles.append( newTri );

	} else {

		vec3 cenWorld(0.0);
		vec3 cenCamera(0.0);
		vec3 cenLocal(0.0);
		vec2 cenTexture(0.0);
		vec3 cenNormal(0.0);
		
		for( int i = 0; i < numPoints; i++ ) {
			cenWorld += (1.0/numPoints)*wPoints[i];
			cenCamera += (1.0/numPoints)*cPoints[i];
			cenLocal += (1.0/numPoints)*lPoints[i];
			cenTexture += (1.0/numPoints)*tPoints[i];
			cenNormal += (1.0/numPoints)*normals[i];
		}

		for( i = 0; i < numPoints; i++ ) {
			
			Triangle newTri;

			int i1 = i;
			int i2 = (i+1)%numPoints;

			newTri.world(0) = cenWorld;
			newTri.camera(0) = cenCamera;
			newTri.local(0) = cenLocal;
			newTri.texture(0) = cenTexture;
			newTri.normal(0) = cenNormal;

			newTri.world(1) = wPoints[i1];
			newTri.camera(1) = cPoints[i1];
			newTri.local(1) = lPoints[i1];
			newTri.texture(1) = tPoints[i1];
			newTri.normal(1) = normals[i1];

			newTri.world(2) = wPoints[i2];
			newTri.camera(2) = cPoints[i2];
			newTri.local(2) = lPoints[i2];
			newTri.texture(2) = tPoints[i2];
			newTri.normal(2) = normals[i2];

			newTri.plane() = pEq;
			newTri.setShader( shader );

			newTriangles.append( newTri );
		}
	}
*/
}

ostream& operator<< ( ostream& s, Face& f )
{
	s << "Number of points = " << f.size() << endl;

	s << "World: " << endl;
	s << f.wPoints << endl;

	s << "Camera: " << endl;
	s << f.cPoints << endl;

	s << "Local: " << endl;
	s << f.lPoints << endl;

	s << "Texture: " << endl;
	s << f.tPoints << endl;

	s << "Screen: " << endl;
	s << f.sPoints << endl;
	
	return s;

}








