#include <fstream.h>
#include "Image.h"
#include "Lerp.h"



/************************************************************************
							Image stuff
*************************************************************************/

//
// Allocates the image
//
Image::Image( vec2& r, vec3& bg ) 
	: res(r), bgColor(bg), regionDebugging(false)
{
	image = new float[(int)(3*res[VX]*res[VY])];
}

//
// Deallocates the image
//
Image::~Image()
{
	if( image ) {
		delete [] image;
	}
}

//
// Used for debugging
//
bool pointInRegion( int x, int y, vec4& region ) 
{
	return(	(x >= region[XL]) && (x <= region[XU]) && (y >= region[YL]) && (y <= region[YU]) );
}

//
// Converts pixel coords to array index.  Returns -1 for out of range pixels
//
inline int Image::posToIndex( int x, int y )
{
	if( (x < 0) || (x >= res[VX]) || (y < 0) || (y >= res[VY]) ) {

		// check for out of range
		return( -1 );

	} else {

		// compute index
		return( fastDblToInt((y*res[VX] + x)) );
	}
}

//
// Writes the image to the specified file in PPM binary format
//
void Image::write( const char *filename )
{
	int imageSize = (int)(res[VX]*res[VY]*3);
	char *byteImage = new char[imageSize];
	int imageIndex = 0;

	// write the image from the top scanline to the bottom, going left to right
	// across each scanline
	for( int y = (int)res[VY] - 1; y >= 0; y-- ) {
		for( int x = 0; x < (int)res[VX]; x++ ) {
			int arrayIndex = 3*((int)res[VX]*y + x);

			byteImage[imageIndex] = (unsigned char)(255.0*image[arrayIndex]);
			byteImage[imageIndex+1] = (unsigned char)(255.0*image[arrayIndex+1]);
			byteImage[imageIndex+2] = (unsigned char)(255.0*image[arrayIndex+2]);

			imageIndex += 3;
		}
	}

	// write the byte image to file
	ofstream outFile( filename, ios::out | ios::binary );
	outFile << "P6 " << (int)res[VX] << " " << (int)res[VY] << " 255 ";
	outFile.write( byteImage, imageSize );
	outFile.flush();

#ifdef DEBUG_RENDERDUDE

	// if region debugging, write the final colours of the region of interest
	if( regionDebugging ) {

		cout << "Debug Region: (" << debugRegion[XL] << "," << debugRegion[YL] << ") -> (" << debugRegion[XU] << "," << debugRegion[YU] << ")" << endl;

		for( int y = (int)debugRegion[YU]; y >= (int)debugRegion[YL]; y-- ) {
			for( int x = (int)debugRegion[XL]; x <= (int)debugRegion[XU]; x++ ) {
				int arrayIndex = y*(int)res[VX] + x;
				cout << "(" << image[3*arrayIndex] << "," << image[3*arrayIndex+1] << "," << image[3*arrayIndex+2] << ") ";
			} 
			cout << endl;
		}
	}
#endif

	delete [] byteImage;
}

//
// Sets the value of a single pixel
//
void Image::setPixel( int x, int y, vec3 &col )
{
	// find the index of the pixel in the image array
	int pos = posToIndex(x,y);

	if( pos >= 0 ) {
		// set the value
		image[3*pos] = (float)col[RED];
		image[3*pos+1] = (float)col[GREEN];
		image[3*pos+2] = (float)col[BLUE];
	}


#ifdef DEBUG_RENDERDUDE
	// if region debugging, watch for writes to pixels of interest
	if( regionDebugging && pointInRegion( x, y, debugRegion ) ) {
		cout << "Face " << debugFace << " setting pixel (" << x << "," << y << ") to " << col << endl;
	}
#endif


}

//
// Increments the value of a single pixel
//
void Image::incrementPixel( int x, int y, vec3 &col )
{
	// find pixel index
	int pos = posToIndex(x,y);

	if( pos >= 0 ) {
		// increment pixel
		image[3*pos] = (float)clamp(image[3*pos]+(float)col[RED], 0.0, 1.0 );
		image[3*pos+1] = (float)clamp(image[3*pos+1]+(float)col[GREEN], 0.0, 1.0 );
		image[3*pos+2] = (float)clamp(image[3*pos+2]+(float)col[BLUE], 0.0, 1.0 );
	}


#ifdef DEBUG_RENDERDUDE
	// if region debugging, watch for writes to pixels of interest
	if( regionDebugging && pointInRegion( x, y, debugRegion ) ) {
		cout << "Face " << debugFace << " incrementing pixel (" << x << "," << y << ") to " << vec3(image[3*pos], image[3*pos+1], image[3*pos+2]) << endl;
		cout << "Mask: " << endl << debugMask << endl;
	}
#endif


}


// 
// Sets a range of pixels, bilinearly interpolating from the 4 given colors.
// The color array should be ordered counterclockwise from the bottom left.
//
void Image::setPixels( int xLow, int yLow, int xHigh, int yHigh, vec3 colors[4] )
{
	// linearly interpolate along the leftmost and rightmost columns of the region
	LerpVec3 leftCol( colors[0], colors[3], vec2((double)yLow,(double)yHigh) );
	LerpVec3 rightCol( colors[1], colors[2], vec2((double)yLow,(double)yHigh) );

	// for each row, linearly interpolate along the pixels in the row from the 
	// values at the start and end of the row
	for( int y = yLow; y <= yHigh; y++ ) {

		// interpolate across the row from the column interpolators
		LerpVec3 curRow( leftCol.current(), rightCol.current(), vec2((double)xLow,(double)xHigh) );

		for( int x = xLow; x <= xHigh; x++ ) {
			setPixel( x, y, curRow.current() );


#ifdef DEBUG_RENDERDUDE
			if( regionDebugging && pointInRegion( x, y, debugRegion ) ) {
				cout << "Face " << debugFace << " setting pixel (" << x << "," << y 
					 << ") to " << curRow.current() << endl;
			}
#endif

			curRow.increment();
		}

		leftCol.increment();
		rightCol.increment();
	}
}

//
// Sets a range of pixels to a constant color
//
void Image::setPixels( int xLow, int yLow, int xHigh, int yHigh, vec3& color )
{
	for( int y = yLow; y <= yHigh; y++ ) {
		for( int x = xLow; x <= xHigh; x++ ) {
			setPixel( x, y, color );


#ifdef DEBUG_RENDERDUDE			
			if( regionDebugging && pointInRegion( x, y, debugRegion ) ) {
				cout << "Face " << debugFace << " setting pixel (" << x << "," << y << ") to " << color << endl;
			}
#endif


		}
	}
}

//
// Sets all pixels to the background colour
//
void Image::clear()
{
	clear( bgColor );
}

//
// Sets all pixels to something other than the background colour
//
void Image::clear( vec3& bg )
{

	if( (bgColor[RED] == 0.0) && (bgColor[RED] == bgColor[GREEN]) && (bgColor[GREEN] == bgColor[BLUE]) ) {

		// if bg is black, use memset to initialize to 0
		memset( image, 0, 3*fastDblToInt(res[VX]*res[VY]*sizeof(float)) );

	} else {

		// otherwise, set all pixels to the color
		int start = 0;
		int end = fastDblToInt(res[VX]*res[VY]);
		for( int i = start; i < end; i++ ) {
			image[3*i] = (float)bg[RED];
			image[3*i+1] = (float)bg[GREEN];
			image[3*i+2] = (float)bg[BLUE];
		}
	}
}

/************************************************************************
							ImageZB stuff
*************************************************************************/

const float ImageZB::FAR_AWAY = 99999.0f;

//
// Constructor, allocates the image and the depth buffer
//
ImageZB::ImageZB( vec2& res, vec3& bgColor ) 
: Image(res, bgColor)
{
	zbuffer = new float[(int)(res[VX]*res[VY])];

	for( int i = 0; i < (int)(res[VX]*res[VY]); i++ ) {
		zbuffer[i] = FAR_AWAY;
	}
}

//
// Virtual destructor deallocates zbuffer
//
ImageZB::~ImageZB()
{
	if( zbuffer ) {
		delete [] zbuffer;
	}
}

// 
// Clears the image and the depth buffer
//
void ImageZB::clear()
{
	Image::clear();

	for( int i = 0; i < (int)res[VX]*(int)res[VY]; i++ ) {
		zbuffer[i] = FAR_AWAY;
	}
}

//
// Sets a pixel and writes the depth value to the depth buffer, provided the
// specified depth is less than the stored depth
//
void ImageZB::setPixel( int x, int y, float depth, vec3& color )
{
	int pos = posToIndex(x,y);

	if( (pos > 0) && (depth < zbuffer[pos]) ) {
		zbuffer[pos] = depth;
		Image::setPixel( x, y, color );
	}
}

//
// Performs the depth test only
//
bool ImageZB::testDepth( int x, int y, float depth )
{
	int pos = posToIndex(x,y);

	if( (pos >= 0) && (depth < zbuffer[pos]) ) {
		return true;
	} else {
		return false;
	}
}

