#ifndef Image_h
#define Image_h

#include "algebra3.h"
#include "List.h"
#include "TriageMask.h"

class Tessellation;



/******************************************************************************

	Class:		Image

	Purpose:	Represents a PPM image.  The whole image is stored in memory.

*******************************************************************************/

class Image  
{
public:
	
	Image( vec2& res, vec3& bgColor );
	virtual ~Image();

	// writes the background colour to every pixel
	virtual void clear();

	// writes another colour to every pixel
	virtual void clear( vec3& bg );

	// writes the image to a file
	void write( const char *filename );

	// set the pixels in the specified region, bilinearly interpolating from the
	// 4 given corner colours
	void setPixels( int xLow, int yLow, int xHigh, int yHigh, vec3 colors[4] );

	// set the pixels in the given region to a constant colour
	void setPixels( int xLow, int yLow, int xHigh, int yHigh, vec3& color );

	// sets one pixel colour
	void setPixel( int x, int y, vec3& col );

	// increments one pixel colour
	void incrementPixel( int x, int y, vec3& color );


	//
	//	DEBUGGING FUNCTIONS
	//

	// is region debugging enabled?
	bool regionDebug() { return regionDebugging; };

	// sets the region of interest for debugging and turns on region debugging
	void setDebugRegion( vec4& region ) { regionDebugging = true; debugRegion = region; }
	
	// declares which face is modifying the image
	void setDebugFace( int face ) { debugFace = face; }

	// declares the pixel mask for the current face
	void setDebugMask( Bitmask& mask ) { debugMask = mask; }



protected:

	float *image;				// the actual image
	vec2 res;					// image resolution
	vec3 bgColor;				// background colour

	int posToIndex( int x, int y );	// given pixel position, returns index in 
									// image array
	
	//	
	//	DEBUGGING DATA
	//
	bool regionDebugging;		// is region debugging enabled?
	vec4 debugRegion;			// region of interest
	int debugFace;				// face currently writing to the image
	Bitmask debugMask;			// current mask for face currently writing to the image
};



/********************************************************************************

	Class:		ImageZB

	Purpose:	Same as Image, but also stores a depth buffer.  Depth comparisons 
				are performed before pixels are written.

**********************************************************************************/

class ImageZB : public Image
{
public:

	ImageZB( vec2& res, vec3& bgColor );
	virtual ~ImageZB();

	const static float FAR_AWAY;

	// clears the image and depth buffer
	void clear();

	// sets a pixel colour, if its depth is less than the currently stored depth
	void setPixel( int x, int y, float depth, vec3& col );

	// just performs the depth test
	bool testDepth( int x, int y, float depth );

protected:

	float *zbuffer;				// the actual depth buffer
};

#endif