#include "IntListList.h"
#include "algebra3.h"
#include "string.h"

IntListList::IntListList( int size )
: fFlatLists(6*size), fIndices(2*size)
{
}

//
//  Returns the number of lists.  Since each list is
//  represented by 2 entries in the fIndices array,
//  return the size of that array / 2.
//
int IntListList::numLists() const
{
	return fIndices.size()/2;
}

//
//  Gets the size of a particular list.  Returns the
//  second element of the (index,size) pair for the
//  specified list.
//
int IntListList::getListSize( int whichList ) const
{
	DB_ASSERT( (whichList < numLists()), "ERROR - invalid list index" );
	return fIndices[2*whichList+1];
}

//
//  Sets the size of the specified list.  The new list
//  size must be <= the old list size, otherwise the indices
//  for the new list wouldn't fit into the space allocated for
//  the old list in the fFlatLists array.
//
void IntListList::setListSize( int whichList, int newSize )
{
	DB_ASSERT( (whichList < numLists()), "ERROR - invalid list index" );
	DB_ASSERT( (newSize <= getListSize(whichList)), "ERROR - inserting list too big" );
	fIndices[2*whichList+1] = newSize;
}

//
//  Retrieves the index in the flat array of a particular list
//
int IntListList::getListIndex( int whichList ) const 
{
	DB_ASSERT( (whichList < numLists()), "ERROR - invalid list index" );
	return fIndices[2*whichList];
}

//
//  Retrieves a pointer to the specified list.  Just look up its
//  starting index in the index array and return a pointer to that
//  index in the flat list array.
//
int *IntListList::getList( int whichList )
{
	DB_ASSERT( (whichList < numLists()), "ERROR - invalid list index" );

	int faceStart = fIndices[2*whichList];

	// check for a freed list (index = -1)
	if( faceStart != -1 ) {
		return &(fFlatLists[faceStart]);
	} else {
		return NULL;
	}
}

//
//  Copies the specified new list over the old list having the 
//  specified index.
//
void IntListList::setList( int whichList, int *newList, int newSize )
{
	DB_ASSERT( (whichList < numLists()), "ERROR - invalid list index" );
	DB_ASSERT( (newSize <= getListSize(whichList)), "ERROR - inserting list too big" );

	int *oldList = getList(whichList);
	for( int i = 0; i < newSize; i++ ) {
		oldList[i] = newList[i];
	}

	setListSize( whichList, newSize );
}

//
//  Appends a new list and returns its index.
//
int IntListList::addList( int *list, int listSize )
{
	fIndices.append( fFlatLists.size() );
	fIndices.append( listSize );
	fFlatLists.append( list, listSize );
	return numLists()-1;
}

//
//  Frees a list by setting its index field to -1
//
void IntListList::freeList( int whichList )
{
	fIndices[2*whichList] = -1;
}

//
//  Appends a new IntListList to this one, and relocates the
//  new lists to be rooted at 'newRoot'.
//
void IntListList::appendAndRelocate( IntListList& other, int newRoot )
{
	int firstV = fFlatLists.size();
	int lastV = firstV + other.fFlatLists.size();

	int firstF = numLists();
	int lastF = firstF + other.numLists();
	int fStart = fFlatLists.size();

	// append the flat list array
	fFlatLists.append( other.fFlatLists );

	// append the index array
	fIndices.append( other.fIndices );

	// relocate the flat array to be rooted at the new root
	for( int v = firstV; v < lastV; v++ ) {
		fFlatLists[v] += newRoot;
	}

	// relocate the index array to be rooted at the end of the
	// old flat array
	for( int f = firstF; f < lastF; f++ ) {
		fIndices[2*f] += fStart;
	}
}


//
//  Shrink wraps the index and flat array so no space is wasted.
//
void IntListList::shrinkWrap()
{
	fFlatLists.shrinkWrap();
	fIndices.shrinkWrap();
}

//
//  Deep copy operator.
//
IntListList& IntListList::operator=( const IntListList& rhs )
{
	fFlatLists = rhs.fFlatLists;
	fIndices = rhs.fIndices;
	return (*this);
}

//
//  Read from file - format is { {1,2,3}, {0,1}, .... }
//
bool IntListList::read( istream& s )
{
	eatWS(s);

	char c;
	s >> c;

	// read the opening brace
	if( c != '{' ) {
		return false;
	}

	while( true ) {
		// read the individual lists
		char curList[100];

		eatWS(s);

		s >> c;
		if( c != '{' ) {
			return false;
		}

		s.get( curList, 100, '}' );

		s >> c;
		if( c != '}' ) {
			return false;
		}

		int curListSize = 0;
		char *curElement = strtok( curList, ", " );

		fIndices.append( fFlatLists.size() );

		while( curElement != NULL ) {
			curListSize++;
			int cur = atoi(curElement);
			fFlatLists.append(cur);
			curElement = strtok( NULL, ", " );
		}

		fIndices.append( curListSize );

		eatWS(s);
		char n = s.peek();
		if( n == ',' ) {
			s >> c;
			continue;
		} else if( n == '}' ) {
			break;
		} else {
			return false;
		}
	}

	s >> c;
	if( c != '}' ) {
		return false;
	}

	return true;
}


//
//  Write the list of lists in the same format as it was read.
//
void IntListList::write( ostream& s )
{
	int num = numLists();

	s << "{";

	for( int cur = 0; cur < num; cur++ ) {	
		
		int *curList = getList( cur );
		int curListSize = getListSize( cur );

		s << "{ ";

		for( int i = 0; i < curListSize; i++ ) {
			if( i != 0 ) {
				s << ",";
			}
			s << curList[i];
		}

		s << " }";

	}

	s << "}";
}
			

