#ifndef List_h
#define List_h

#include <stdlib.h>
#include <memory.h>
#include <iostream.h>
#include "Global.h"
#include "parseInput.h"

/****************************************************************

	Class:		List<T>
	Purpose:	Dynamically allocated list class.  List size doubles
				whenever it fills to capacity.

*****************************************************************/

template <class T> class List
{
	public:

		// constructor specifies initial size of list
		List( int size = 1 );
		~List();

		// nonconst indexing operator
		T& operator[]( int index );
		
		// const indexing operator
		const T& operator[]( int index ) const;

		// appends a single element to the end of the list
		void append( const T& newT );

		// appends an array (size 'numNew') of elements to the end of the list
		void append( T* newT, int numNew );

		// appends another list
		void append( const List<T>& newT );

		// size of the list (number of elements that have been appended)
		int size() const { return logicalSize; };

		// resizes the list
		void resize( int newSize );

		// makes the physical size equal to the logical size
		void shrinkWrap();

		// stack operations
		void push( const T& newT );
		T& pop();

		// assignment operator, copies the list
		List<T>& operator=( const List<T>& other );

		// pointer to a sublist starting at given index
		T* subList( int index );

		// initializes the entire list
		void setList( unsigned char init );

		// finds an element having name 'name', where 'compFunc' compares an
		// element with a string name
		int find( const char *name, int (*compFunc)(T&, const char *) ) const;

		// output operator
		friend ostream& operator<<( List<T>& l, ostream& s );

	private:
		int logicalSize;		// the actual number of elements allocated
		int physicalSize;		// the number of those elements that are used
		T* list;				// the list
};

template <class T> T* find( const char *name, const List<T *>& list )
{
	int ret = -1;
	int size = list.size();

	for( int i = 0; i < size; i++ ) {
		if( *(list[i]) == name ) {
			ret = i;
			break;
		}
	}

	if( ret == -1 ) {
		return NULL;
	} else {
		return list[ret];
	}
}

//
//	Constructor - allocates the list
//
template <class T>List<T>::List<T>( int size )
: logicalSize(0), physicalSize(size)
{
	DB_ASSERT( (size > 0), "ERROR - cannot init size 0 list" );
	if( size > 0 ) {
		list = new T[physicalSize];
	} else {
		list = NULL;
	}
}


//
//	Destructor - deletes the list
//
template <class T>List<T>::~List<T>()
{
	if( list != NULL ) {
		delete [] list;
	}
}

//
//	Returns a pointer to the sublist starting at 'index'
//
template <class T> inline T* List<T>::subList( int index ) 
{ 
	DB_ASSERT( ((index >= 0) && (index < logicalSize)), "ERROR - invalid sublist index" );
	return &(list[index]); 
}

//
//  List index operator
//
template <class T> inline T& List<T>::operator[]( int index ) 
{ 
	DB_ASSERT( ((index >= 0) && (index < logicalSize)), "ERROR - invalid list index" );
	return list[index]; 
}

//
//  List index operator
//
template <class T> inline const T& List<T>::operator[]( int index ) const
{ 
	DB_ASSERT( ((index >= 0) && (index < logicalSize)), "ERROR - invalid list index" );
	return list[index]; 
}


//
//	Appends an element to the end of the list, incrementing
//  the logical size.  If the logical size gets bigger than
//	the physical size, the physical size is doubled, and the
//	array is reallocated and copied, so that the list is
//	always represented as a single contiguous array.
//
template <class T> void List<T>::append( const T& newT )
{
	if( logicalSize == physicalSize ) {

		// if array is full, resize and copy it
		int oldLogicalSize = logicalSize;
		resize( 2*physicalSize );

		// insert the new element
		logicalSize = oldLogicalSize + 1;
		list[oldLogicalSize] = newT;
		

	} else {

		// just insert the new element
		list[logicalSize++] = newT;

	}
}


//
//	Appends 'numNew' entries from the array 'newT'
//
template <class T> void List<T>::append( T* newT, int numNew )
{
	for( int i = 0; i < numNew; i++ ) {
		append( newT[i] );
	}
}


//
//  Appends "newList" to this list
//
template <class T> void List<T>::append( const List<T>& newList )
{
	append( newList.list, newList.logicalSize );
}


//
//	Stack push - just append the element to the end of the list
//
template <class T> void List<T>::push( const T& newT )
{
	append( newT );
}


//
//	Stack pop - remove the last element and return it
//
template <class T> T& List<T>::pop()
{
	DB_ASSERT( (logicalSize > 0), "ERROR - pop from empty stack" );
	return list[(logicalSize--)-1];
}


//
//	Change the logical size of the list, altering the physical
//	size if necessary (ie if the list isn't big enough to accommodate
//	the new logical size).
//
template <class T> void List<T>::resize( int newSize )
{
	if( newSize > physicalSize ) {

		// need to reallocate the list
		if( list != NULL ) {

			// save old list, allocate new one
			T* oldList = list;
			list = new T[newSize];

			// copy old into new
			memcpy( list, oldList, logicalSize*sizeof(T) );
			
			// delete old
			delete [] oldList;

		} else {

			// list doesn't yet exist, so just allocate it
			list = new T[newSize];
		}

		// adjust physical and logical sizes
		logicalSize = physicalSize = newSize;

	} else {
		
		// physical size OK, just change logical size
		logicalSize = newSize;

	}
}

//
//	Resizes the list so that the physical size is exactly equal to
//	the logical size, so no extra space is being used if we know
//	the list is not going to be growing
//
template <class T> void List<T>::shrinkWrap()
{
	if( physicalSize > logicalSize ) {
		T* oldList = list;
		list = new T[logicalSize];
		memcpy( list, oldList, logicalSize*sizeof(T) );
		physicalSize = logicalSize;
	}
}

//
//	Assignment operator - does a shallow copy of the entire list
//
template <class T> List<T>& List<T>::operator=( const List<T>& other )
{
	// delete current list, if it exists
	if( list != NULL ) {
		delete [] list;
	}

	// set sizes, allocate new space
	logicalSize = other.logicalSize;
	physicalSize = other.physicalSize;
	list = new T[physicalSize];

	// copy list
	memcpy( list, other.list, logicalSize*sizeof(T) );

	return (*this);
}


//
//  Sets all the memory for the list to a particular character
//
template <class T> void List<T>::setList( unsigned char init )
{
	memset( list, init, sizeof(T)*physicalSize );
}

//
//	Find an element in the list that matches a string, where 'compFunc' determines
//	whether or not an element matches a string.
//
template <class T> int List<T>::find( const char *name, int (*compFunc)(T&, const char *) ) const
{
	int last = logicalSize - 1;

	for( int i = 0; i <= last; i++ ) {
		if( (*compFunc)(list[i], name ) == 0 ) {
			return i;
		}
	}

	return -1;
}


//
//	Output operator
//
template <class T> ostream& operator<<( ostream& s, List<T>& l )
{
	int n = l.size();

	for( int i = 0; i < n; i++ ) {
		s << l[i] << ",";
	}

	return s;
}

template <class T> bool readList( istream& s, List<T>& l )
{
	T cur;
	int key;

	enum {  OPENBR=1,
			CLOSEBR=2,
			COMMA=3,
			UNKNOWN=4
	};

	static char *key_words[] = { "{",
								 "}",
								 ",",
								 0
	};

	key = parseInput( s, key_words );
	if( key != OPENBR ) {
		return false;
	}

	for (;;) {
		
		eatWS( s );		
		
		s >> cur;
		l.append(cur);

		key = parseInput( s, key_words );
		if( key == COMMA ) {
			continue;
		} else if( key == CLOSEBR ) {
			break;
		} else {
			return false;
		}

	}

	if( key != CLOSEBR ) {
		key = parseInput( s, key_words );
		if( key != CLOSEBR ) {
			return false;
		}
	}
	return true;
}

#endif