#include "Model.h"
#include "Scene.h"
#include <string.h>

class Shape;

//
//  Attribute index definition
//
int Model::kTranslate;
int Model::kRotate;
int Model::kScale;
int Model::kParent;

//
//	Default constructor, everything is identity
//
Model::Model() 
: localTform( identity3D() ), localNormTform( identity3D() ),
  cumulativeTform( identity3D() ), cumulativeNormTform( identity3D() ),
  parent(NULL)
{
	Model::initializeAttributes();
}

//
//  Attribute initialization
//
void Model::initializeAttributes()
{
	NamedEntity::initializeAttributes();

	kTranslate =	addAttr( new AttributeVec3( "translate", ANIMATABLE, vec3(0.0) ) );
	kRotate	=		addAttr( new AttributeVec3( "rotate", ANIMATABLE, vec3(0.0) ) );
	kScale =		addAttr( new AttributeVec3( "scale", ANIMATABLE, vec3(1.0) ) );
	kParent =		addAttr( new AttributeString( "parent", CONSTANT, "" ) );
}

//
//  Post-read func, called after a Model has been read from a file,
//  tries to find the parent model.
//
bool Model::postReadFunc( const Scene& scene )
{
	if( !NamedEntity::postReadFunc(scene) ) {
		return false;

	} else {

		// try to find the parent model in the scene
		if( strlen( attrParent() ) ) {
			Shape *parentPtr = scene.findShape( attrParent() );
	
			if( parentPtr == NULL ) {
				cout << "ERROR - can't find parent " << attrParent() << endl;
				return false;
			} else {	
				parent = (Model *)parentPtr;
			}
		}
		return true;
	}
}

//
//  Determines whether or not the transform for this model is animated
//
bool Model::animatedTransform()
{
	// check for this model's animated transform
	bool localAnimated =	findAttr(kTranslate)->isAnimated() ||
							findAttr(kRotate)->isAnimated() ||
							findAttr(kScale)->isAnimated();

	// check for parent as well
	bool parentAnimated = false;
	if( parent != NULL ) {
		parentAnimated = parent->animatedTransform();
	}

	return (localAnimated | parentAnimated);
		
}

//
//	Recomputes the cumulative transform and cumulative norm transform.  We also
//  store copies of just the rotation and translation parts of the matrix, for
//  use in parenting transforms
//
void Model::updateTransform()
{
	// update the local transform
	mat4 scaleMat = scaling3D( attrScale() );

	vec3 rotateVec = attrRotate();

	mat4 rotateMat = rotation3D( vec3(1.0,0.0,0.0), rotateVec[VX] ) *
					 rotation3D( vec3(0.0,1.0,0.0), rotateVec[VY] ) *
					 rotation3D( vec3(0.0,0.0,1.0), rotateVec[VZ] );

	mat4 translateMat = translation3D( attrTranslate() );

	localTformTR = translateMat * rotateMat;
	localNormTformTR = localTformTR.inverse().transpose();

	localTform = localTformTR * scaleMat;
	localNormTform = localTform.inverse().transpose();

	// check for a parent
	if( parent != NULL ) {

		// has a parent, so update the parent transforms and append the
		// local transforms to get the cumulative transform.  We only use
		// the translate and rotate parts of the parent transform.
		parent->updateTransform();
		
		cumulativeTform = parent->cumulativeTformTR * localTform;
		cumulativeNormTform = parent->cumulativeNormTformTR * localNormTform;

		// store translate/rotate matrices
		cumulativeTformTR = parent->cumulativeTformTR * localTformTR;
		cumulativeNormTformTR = parent->cumulativeNormTformTR * localNormTformTR;

	} else {

		// no parent, local transforms same as cumulative transforms
		cumulativeTform = localTform;
		cumulativeNormTform = localNormTform;

		// store translate/rotate matrices
		cumulativeTformTR = localTformTR;
		cumulativeNormTformTR = localNormTformTR;

	}
}

