#include "Octree.h"
#include "Tessellation.h"
#include "BspTree.h"

// definition of octree node max size.  Increasing this size makes
// octree construction faster, but will slow BSP creation
const int Octree::MAX_FACES = 100;

// This array is a set of precomputed front-to-back octant orderings, indexed
// by the octant in which the eyepoint lies.  
static OctantType octreeOrders[8][8] = 
{
	{xNyNzN,xNyNzP,xNyPzN,xNyPzP,xPyNzN,xPyNzP,xPyPzN,xPyPzP},	// xNyNzN
	{xNyNzP,xNyNzN,xNyPzP,xNyPzN,xPyNzP,xPyNzN,xPyPzP,xPyPzN},	// xNyNzP
	{xNyPzN,xNyPzP,xNyNzN,xNyNzP,xPyPzN,xPyPzP,xPyNzN,xPyNzP},	// xNyPzN
	{xNyPzP,xNyPzN,xNyNzP,xNyNzN,xPyPzP,xPyPzN,xPyNzP,xPyNzN},	// xNyPzP
	{xPyNzN,xPyNzP,xPyPzN,xPyPzP,xNyNzN,xNyNzP,xNyPzN,xNyPzP},	// xPyNzN
	{xPyNzP,xPyNzN,xPyPzP,xPyPzN,xNyNzP,xNyNzN,xNyPzP,xNyPzN},	// xPyNzP
	{xPyPzN,xPyPzP,xPyNzN,xPyNzP,xNyPzN,xNyPzP,xNyNzN,xNyNzP},	// xPyPzN
	{xPyPzP,xPyPzN,xPyNzP,xPyNzN,xNyPzP,xNyPzN,xNyNzP,xNyNzN}	// xPyPzP
};


/****************************************************************

	Class:		OctreeNode
	Purpose:	Represents a node in an octree.  Stores a list
				of the faces contained in this node, and a list
				of pointers to child nodes.

*****************************************************************/
class OctreeNode 
{
public:

	// create an octree node based on a particular bounding box
	OctreeNode( BoundingBox& box, int numFaces );
	~OctreeNode();

	// adds a particular face to this node (no splitting)
	void addFace( int face );

	friend class Octree;
	friend ostream& operator<<( ostream& s, Octree& o );

private:

	int numFaces;		// number of faces in this node
	BoundingBox bb;		// axis-aligned bounding box for this node
	int *children;		// array of indices of children of this node
	List<int> *faces;	// list of faces contained in this node
	BspTree *bspTree;	// BSP tree for the faces contained in this node
};

// 
//  Constructor just initializes the face list - no children, no BSP tree
//
OctreeNode::OctreeNode( BoundingBox& box, int num ) 
: bb(box), children(NULL), numFaces(0), bspTree(NULL)
{
	faces = new List<int>(num);
}

//
//  Clean up.
// 
OctreeNode::~OctreeNode()
{
	if( faces ) { 
		delete faces;
	}

	if( children ) {
		delete [] children;
	}

	if( bspTree ) {
		delete bspTree;
	}
}

//
//  Adds a face to an octree node.
//
void OctreeNode::addFace( int face )
{
	faces->append(face);
	numFaces++;
}


/***********************************************************************
							OCTREE STUFF
************************************************************************/

//
//	Constructor builds the octree by inserting every
//  face of the specified tessellation
//
Octree::Octree( Tessellation *t, BoundingBox& bb )
: tess(t), nodes(20)
{
	// how many faces to add?
	int numFaces = t->size();

	// create the root node
	nodes.append( new OctreeNode( bb, MAX_FACES ) );
	
	// insert each face, if it is valid
	for( int face = 0; face < numFaces; face++ ) {
		if( tess->fList.getList(face) != NULL ) {
			insertFace( face );
		}
	}
}

//
//  Deletes the octree
//
Octree::~Octree()
{
	for( int i = 0; i < nodes.size(); i++ ) {

		if( nodes[i] != NULL ) {
			delete nodes[i];
		}

	}
}

//
//  Splits a node into 8 children, clips the faces of the node to the
//  children, and distributes the clipped faces to the children.
// 
void Octree::splitNode( int nodeIndex )
{
	// get the node and its face list
	OctreeNode *node = nodes[nodeIndex];
	List<int> &faces = *(node->faces);
	int numFaces = faces.size();

	// initialize the child array to all -1's (meaning 'empty children')
	node->children = new int[8];
	for( OctantType i = 0; i < 8; i++ ) {
		node->children[i] = -1;
	}

	// split each face to the 8 children and insert the pieces into the
	// children
	for( int faceIndex = 0; faceIndex < numFaces; faceIndex++ ) {
		splitAndInsertFace( faces[faceIndex], nodeIndex );
	}

	// deallocate the face list for the parent node
	delete node->faces;
	node->faces = NULL;
}

//
//  "Initializes" a particular child of a node.  Creates a new node
//  for the child, and initializes the child pointer in the parent.
//
void Octree::addChild( int nodeIndex, OctantType whichChild )
{
	// find the parent and compute the child bounding box
	OctreeNode *node = nodes[nodeIndex];
	BoundingBox subBox = node->bb.subBox( whichChild );

	// create a node for the child
	OctreeNode *newNode = new OctreeNode( subBox, MAX(1,(MAX_FACES/4)) );

	// add the node to the tree and initialize the parent's child pointer
	nodes.append( newNode );
	node->children[whichChild] = nodes.size() - 1;
}

//
//  Inserts a face into a node, possibly splitting the face and/or the 
//  node.
//
void Octree::insertFace( int face, int nodeIndex )
{
	OctreeNode* node = nodes[nodeIndex];	

	// if the node has already been split, split the face and distribute 
	// the pieces to the children of the node
	if( node->children != NULL ) {
		splitAndInsertFace( face, nodeIndex );

	} else {
		
		// the node hasn't been split, so add the face to the node
		node->addFace( face );

		// if the node has exceeded the maximum number of faces allowed in a
		// node, split the node
		if( node->numFaces >= MAX_FACES ) {
			splitNode( nodeIndex );
		}
	}
}


//
//  Splits a face and distributes the pieces to the node's children, 
//  initializing the child nodes if necessary.  New faces go into the
//  tessellation data structure.
//
void Octree::splitAndInsertFace( int face, int nodeIndex )
{
	int newFaces[8];
	OctreeNode* node = nodes[nodeIndex];

	// split the face along the octree median planes
	tess->octreeSplitFace( face, node->bb, newFaces );

	// loop through the child octants
	for( OctantType i = 0; i < 8; i++ ) {
		if( newFaces[i] != -1 ) {
			// there is a piece of the face in this octant,
			// so check to see if the child exists
			if( node->children[i] == -1 ) {

				// the child doesn't exist, so add it
				addChild( nodeIndex, i );
			}

			// otherwise, insert the face piece into the child node
			insertFace( newFaces[i], node->children[i] );
		}
	}
}

//
//  Builds a BSP tree out of each of the leaf nodes of the octree.  Any new
//  faces go into the tessellation.
//
void Octree::constructBSP()
{
	// for each node, grab its face list, and create a BSP tree for that 
	// list of faces
	int numNodes = nodes.size();
	for( int i = 0; i < numNodes; i++ ) {

		OctreeNode *node = nodes[i];
		if( node->children == NULL ) {

			node->bspTree = new BspTree( tess, *(node->faces) );

		}
	}
}

//
//  Compute a front-to-back ordering of the faces in the octree for a particular
//  viewpoint.  Use the precomputed octant ordering to order the octree nodes, and
//  use the BSP trees to order the faces within each octree node.
//
void Octree::orderFaces( vec3& viewPoint, List<int>& faces, int nodeIndex )
{
	OctreeNode* node = nodes[nodeIndex];

	if( node->children == NULL ) {
		node->bspTree->orderFaces( viewPoint, faces );
	} else {
		OctantType whichOctant = node->bb.whichOctant( viewPoint );
		OctantType *order = octreeOrders[whichOctant];

		for( OctantType i = 0; i < 8; i++ ) {
			if( node->children[order[i]] != -1 ) {
				orderFaces( viewPoint, faces, node->children[order[i]] );
			}
		}
	}
}

// 
//  Output operator
//
ostream& operator<<( ostream& s, Octree& octree ) 
{
	for( int i = 0; i < octree.nodes.size(); i++ ) {
		if( octree.nodes[i]->children != NULL ) {
			cout << "Node " << i << " already split" << endl;
			cout << "xNyNzN = " << octree.nodes[i]->children[xNyNzN] << endl;
			cout << "xNyNzP = " << octree.nodes[i]->children[xNyNzP] << endl;
			cout << "xNyPzN = " << octree.nodes[i]->children[xNyPzN] << endl;
			cout << "xNyPzP = " << octree.nodes[i]->children[xNyPzP] << endl;
			cout << "xPyNzN = " << octree.nodes[i]->children[xPyNzN] << endl;
			cout << "xPyNzP = " << octree.nodes[i]->children[xPyNzP] << endl;
			cout << "xPyPzN = " << octree.nodes[i]->children[xPyPzN] << endl;
			cout << "xPyPzP = " << octree.nodes[i]->children[xPyPzP] << endl;
		} else {
			if( octree.nodes[i]->bspTree ) {
				cout << "Node " << i << endl << *(octree.nodes[i]->bspTree) << endl << endl;
			} else {
				cout << "Node " << i << " - no BSP tree" << endl;
			}
		}
	}
	return s;
}



































