#ifndef Octree_h
#define Octree_h

#include <iostream.h>
#include "BoundingBox.h"

class Tessellation;
class BspTree;
class OctreeNode;


/****************************************************************

	Class:		Octree	
	Purpose:	Represents an octree spatial partition of a given
				tessellation.

				Faces are only stored in the leaf nodes of the 
				octree, and faces straddling multiple octree nodes
				are clipped to each intersected node.  Each octree
				node can also store a BSP of its contents.  In this
				manner, a front-to-back traversal of the entire contents
				of the octree can be performed.

				Nodes in the octree are identified by integers, with
				node 0 being the root of the tree.

				The octree uses a regular subdivision scheme starting
				from a bounding box that is provided to the constructor.

				The 'OctantType' enumeration is used to identify the 8
				octants of a cube, such as xNyPzN (meaning the octant in
				which x <= xm, y >= ym, z <= zm, where x=xm, y=ym, and z=zm
				are the median planes of the cube.

*****************************************************************/
class Octree
{
public:

	// create an octree based on a tessellation, starting from a given 
	// bounding box
	Octree( Tessellation *tess, BoundingBox& bb );
	~Octree();

	// inserts a face into a given node, possibly splitting the face and 
	// distributing the pieces to the children of the node, if the node has
	// already split
	void insertFace( int face, int node=0 );

	// splits a face along the 3 dividing planes of the node and distributes
	// the pieces to the children of the node
	void splitAndInsertFace( int face, int node=0 );
	
	// splits a node into 8 subnodes, splitting all faces in the node and 
	// distributing them to the children of the node
	void splitNode( int node );

	// adds a child of a node corresponding to a particular octant of that
	// node
	void addChild( int nodeIndex, OctantType whichChild );

	// constructs BSP trees out of each of the leaf nodes
	void constructBSP();

	// constructs a front-to-back ordering of all the faces in the octree,
	// starting at the specified node
	void orderFaces( vec3& viewPoint, List<int>& faces, int nodeIndex = 0 );

	// output operator
	friend ostream& operator<<( ostream& s, Octree& octree );


private:

	// determines the maximum number of faces allowed in an octree
	// cell before it is split
	static const int MAX_FACES;

	// the tessellation on which the octree is based
	Tessellation *tess;

	// the actual octree itself
	List<OctreeNode *> nodes;
};

#endif