#include <assert.h>
#include "Poly.h"
#include "Triangle.h"
#include "Face.h"
#include "glut.h"
#include "Tessellation.h"

//
//	Attribute index definitions
//
int Poly::kObjPoints;
int Poly::kTexPoints;
int Poly::kNormals;
int Poly::kFaceList;

//
//  Constructor just initializes attributes
//
Poly::Poly()
{
	Poly::initializeAttributes();
}

//
//  Attribute initialization
//
void Poly::initializeAttributes()
{
	Shape::initializeAttributes();

	kObjPoints = addAttr( new AttributeVec3List( "objpoints", CONSTANT ) );
	kTexPoints = addAttr( new AttributeVec2List( "texpoints", CONSTANT ) );
	kNormals   = addAttr( new AttributeVec3List( "normals", CONSTANT ) );
	kFaceList  = addAttr( new AttributeIntListList( "facelist", CONSTANT ) );
}

//
//  Post-read function determines whether or not normals/tex coords have
//  been supplied.
//
bool Poly::postReadFunc( const Scene& scene )
{
	if( !Shape::postReadFunc( scene ) ) {
		return false;
	}

	fHasNormals = attrNormals().size() ? true : false;
	fHasTexCoords = attrTexPoints().size() ? true : false;

	return true;
}

//
//  Tessellation method.  Tessellates the poly into its local
//  tessellation structure.  If the 'space' parameter is WORLD,
//  the vertices are converted into world space.  
//
//	The options provided to the function are merged with any other
//  options specified in the file format for this poly.
//
void Poly::localTessellate( TessOptions options, SpaceType space )
{
	// merge the supplied options with the options from the file
	TessOptions allOptions = options | getTessOptions();

	cout << "Local tessellating " << attrName() << " - " << endl;
	if( allOptions & kTessNormals ) {
		cout << " normals ";
	}
	if( allOptions & kTessTex ) {
		cout << " textures ";
	}
	if( allOptions & kTessPlaneEqs ) {
		cout << " planeeqs ";
	}
	cout << endl;

	// get the points and faces for the poly
	List<vec3>& objPoints = attrObjPoints();
	IntListList& faces = attrFaceList();

	// allocate the local tessellation
	localTess = new Tessellation( allOptions, faces.numLists(), objPoints.size() );

	// store the shape pointer
	localTess->sList.append( this );

	// copy the object-space vertices into the tessellation
	localTess->vList = attrObjPoints();

	// copy the normals into the tessellation, if necessary
	if( allOptions & kTessNormals ) {
		DB_ASSERT( attrHasNormals(), "Can't provide normals" );
		localTess->nList = attrNormals();
	}

	// copy the texture coords into the tessellation, if necessary
	if( allOptions & kTessTex ) {
		DB_ASSERT( attrHasTex(), "Can't provide tex" );
		localTess->tList = attrTexPoints();
	}

	// copy the face list into the tessellation
	localTess->fList = faces;

	// if worldspace tessellation is required, transform the vertices and
	// normals
	if( space == WORLD ) {
		localTess->transformVertices( getCtransform(), NULL );
		localTess->transformNormals( getCNtransform(), NULL );
	}

	// if plane equations are required, compute them here
	if( allOptions & kTessPlaneEqs ) {
		localTess->computePlaneEqs(NULL,NULL);
	}
}

//
// Count the faces
//
int Poly::faceCount()
{
	return( attrFaceList().numLists() );
}

//
//  Compute the bounding box from the vertices.  Apply the given
//  transformation to each vertex.
//
void Poly::computeBB( const mat4& transform )
{
	// get the vertex position list
	const List<vec3>& objPoints = attrObjPoints();
	int numPoints = objPoints.size();

	// loop through the vertices, expanding the bounding box to 
	// fit each new vertex
	for( int curPoint = 0; curPoint < numPoints; curPoint++ ) {
		if( curPoint == 0 ) {
			// initialize the bounding box for the first vertex
			vec3 point0 = transform * objPoints[0];
			fBoundingBox = BoundingBox( point0, point0 );
		} else {
			fBoundingBox.expand( transform * objPoints[curPoint] );
		}
	}
}


//
//  Computes the tessellation options for this shape.  Combine the
//  data that was provided (indicated by fHasNormals, fHasTex) with
//  any additional criteria specified through the hasNormals and
//  hasTex attributes.
//
TessOptions Poly::getTessOptions()
{
	TessOptions ret = kTessNone;

	if( attrHasNormals() && fHasNormals ) {
		ret |= kTessNormals;
	}

	if( attrHasTex() && fHasTexCoords ) {
		ret |= kTessTex;
	}

	return ret;
}


















