#include "RefMesh.h"
#include <fstream.h>
#include <string.h>

//
//  Attribute index definition
//
int RefMesh::kFilename;
int RefMesh::kFaceCount;
int RefMesh::kTriangulate;

//
//  Constructor just initializes attributes
//
RefMesh::RefMesh()
{
	RefMesh::initializeAttributes();
}

//
//  Attribute initialization
//
void RefMesh::initializeAttributes()
{
	Shape::initializeAttributes();

	kFilename = addAttr( new AttributeString( "filename", CONSTANT, "" ) );
	kFaceCount = addAttr( new AttributeInt( "facecount", CONSTANT, 500 ) );
	kTriangulate = addAttr( new AttributeBool( "triangulate", CONSTANT, false ) );
}

//
//  Read the mesh from an OBJ file.
//
bool RefMesh::postReadFunc( const Scene& scene )
{
	if( !Shape::postReadFunc( scene ) ) {
		return false;
	}

	// temp buffers to read in points, normals, tex coords
	List<vec3> tempV(100);
	List<vec3> tempN(100);
	List<vec2> tempT(100);

	// input stream for the input file
	ifstream refStream( attrFilename() );

	// temp vec3 and vec2
	vec3 data3;
	vec2 data2;

	// read the point/normal/texcoord data
	while( !refStream.eof() ) {

		char line[100];
		refStream >> line;

		if( strcmp(line,"vn") == 0 ) {

			// read the normal
			refStream >> data3;
			tempN.append(data3);

		} else if( strcmp(line,"vt") == 0 ) {

			// read the texture coord
			refStream >> data2;
			tempT.append(data2);

		} else if( strcmp(line,"v") == 0 ) {

			// read the vertex position
			refStream >> data3;
			tempV.append(data3);

		} else if( strcmp(line,"f") == 0 ) {

			// done with vertex info, starting faces
			refStream.putback('f');
			break;

		} else {

			// unknown, just ignore it
			refStream.getline(line,100);
		}
	}

	// now we read  the face list - lists of vertex/normal/texcoord
	// triples

	int numV = tempV.size();			// number of vertices
	List<int> lastUsedVertex( numV );	// the last index in the face list
										// where each vertex was used
	for( int i = 0; i < numV; i++ ) {
		lastUsedVertex.append(-1);
	}

	List<int> vertexN(10);				// list of normal indices
	List<int> vertexT(10);				// list of texture indices
	List<int> prevPtr(10);				// pointers to previous uses of each vertex

	// figure out which tessellation options this object provides
	TessOptions options = kTessNone;
	int faceVertSize = 1;
	if( tempN.size() ) {
		
		if( attrHasNormals() ) {
			options |= kTessNormals;
		}

		faceVertSize++;
	}
	if( tempT.size() ) {
		
		if( attrHasTex() ) {
			options |= kTessTex;
		}

		faceVertSize++;
	}

	// allocate the local tessellation
	localTess = new Tessellation( options, attrFaceCount(), numV );
	List<vec3>& objPoints = localTess->vList;
	List<vec3>& normals = localTess->nList;
	List<vec2>& texPoints = localTess->tList;
	IntListList& faceList = localTess->fList;

	while( !refStream.eof() ) {

		char line[200];
		refStream.getline(line,200);
		
		if( line[0] != 'f' ) continue;

			int indices[3];
			int i = 0;

			int faceSize = 0;
			int faceIndices[20];

			// read the indices on the line
			char *token = strtok( line+1, " /" );
			while( token != NULL ) {

				if( strlen(token) ) {
					indices[i++] = atoi(token);
				}

				if( i == faceVertSize ) {

					// we've got a v/t/n triple
					int v,t,n;
					int cur = 0;
					v = indices[cur++]-1;

					if( options & kTessTex ) {
						t = indices[cur++]-1;
					} else {
						t = -1;
					}

					if( options & kTessNormals ) {
						n = indices[cur++]-1;
					} else {
						n = -1;
					}

					// get head of chain
					int lu = lastUsedVertex[v];

					// loop back through the prevPtr list, looking for
					// a previous use of this v/t/n combination
					while( lu != -1 ) {

						if( (vertexN[lu] == n) && (vertexT[lu] == t) ) {
							break;
						} else {
							lu = prevPtr[lu];
						}

					}

					if( lu == -1 ) {
						// we didn't find this v/t/n combo, so add it

						objPoints.append( tempV[v] );

						if( options & kTessTex ) {
							texPoints.append( tempT[t] );
						}

						if( options & kTessNormals ) {
							normals.append( tempN[n] );
						}

						vertexN.append(n);
						vertexT.append(t);

						prevPtr.append( lastUsedVertex[v] );
						lastUsedVertex[v] = objPoints.size()-1;

						faceIndices[faceSize++] = objPoints.size()-1;
					
					} else {

						// we already have this v/t/n combo
						faceIndices[faceSize++] = lu;
					}

					i = 0;
				}
				token = strtok( NULL, " /" );
			}

			DB_ASSERT( (faceSize <= MAX_VERTICES), "ERROR - read face too large" );


			if( attrTriangulate() ) {
				for( int curTri = 0; curTri < faceSize-2; curTri++ ) {
					int indices[3];

					indices[0] = faceIndices[0];
					indices[1] = faceIndices[curTri+1];
					indices[2] = faceIndices[curTri+2];

					faceList.addList( indices, 3 );
				}
			} else {
				faceList.addList( faceIndices, faceSize );
			}
	}

	// resize the face list so as not to waste any space
	localTess->fList.shrinkWrap();

	return true;
}

//
//	Local tessellation - since we already have the tessellation from
//  when the file was read, we just perform the transformation to 
//  worldspace and the plane equation computation, if necessary. 
//
void RefMesh::localTessellate( TessOptions options, SpaceType space )
{
	localTess->getOptions() |= options;

	if( space == WORLD ) {
		localTess->transformVertices( getCtransform(), NULL );
		localTess->transformNormals( getCNtransform(), NULL );
	}

	if( localTess->getOptions() & kTessPlaneEqs ) {
		localTess->computePlaneEqs(NULL,NULL);
	}
}

//
//  Count the faces
//
int RefMesh::faceCount()
{
	return localTess->fList.numLists();
}

//
//  Compute bounding box from vertices
//
void RefMesh::computeBB( const mat4& transform )
{
	int numVertices = localTess->vList.size();

	for( int i = 0; i < numVertices; i++ ) {
		if( i == 0 ) {
			vec3 point0 = localTess->vList[i];
			fBoundingBox = BoundingBox( point0, point0 );
		} else {
			fBoundingBox.expand( transform * localTess->vList[i] );
		}
	}
}
