#ifndef Scene_h
#define Scene_h

#include <iostream.h>
#include "List.h"
#include "BoundingBox.h"
#include "Tessellation.h"
#include "Cam.h"
#include "Shape.h"
#include "Light.h"
#include "Shader.h"
#include "Texture.h"
#include "AnimCurve.h"

/****************************************************************

	Class:		Scene
	Purpose:	Represents a RenderDude(tm) scene.		

*****************************************************************/
class Scene
{
public:
	Scene() {};
	~Scene(){};

	//
	//  Scene input/output methods
	//
	bool readScene( istream& s );
	virtual void write( ostream& s );


	//
	//  Scene rendering methods
	//

	// preProcess() gets called after the scene has been read, but before
	// rendering takes place
	void preProcess() {};

	// render() renders all the renderable cameras in the scene
	void render();

	// postProcess() gets called after all the renderable cameras have been 
	// rendered
	void postProcess() {};


	//
	//  Tessellation-related methods
	//

	// constructs the local tessellation for all the objects in the scene
	void localTessellate( TessOptions localOptions, SpaceType space );

	// constructs the worldspace bounding box for the scene
	void buildBB();

	// bounding box accessor
	BoundingBox getBB() { return sceneBB; };

	// maximum vertices in any local tessellation in the scene
	int maxVertices();

	// total number of faces in the scene
	int faceCount();


	// 
	//  Animation methods
	//

	// sets the current animation time - marks all animated attributes as
	// dirty
	void setCurTime( double newTime );

	// get current animation time
	double getCurTime() { return Attribute::curTime(); };

	// updates all the entities in the scene at the current animation time
	void updateTextures();	// updates texture transforms
	void updateGeometry();	// updates geometry transforms
	void updateShaders();	// nothing for now
	void updateLights();	// updates light transforms
	void updateAnimCurves();	// evaluates animation curves at the current time


	//
	//  Entity accessors - finds the various named entities in the scene
	//					   based on their names.
	//
	Cam	*findCam( char *camName ) const { return find<Cam>(camName, cameraList); };
	Light *findLight( char *lightName ) const { return find<Light>(lightName, lightList); };
	Shape *findShape( char *shapeName ) const { return find<Shape>(shapeName,objList); };
	Shader *findShader( char *shaderName ) const { return find<Shader>(shaderName,shaderList); };
	Texture *findTexture( char *textureName ) const { return find<Texture>(textureName,textureList); };
	AnimCurve *findAnimCurve( char *animCurveName ) const { return find<AnimCurve>(animCurveName,animCurveList); };

	// cameras have access to scene internals
	friend class CamOpenGL;
	friend class CamRT;
	friend class CamZB;
	friend class CamHCM;
	friend class Cam;

private:
	List<Cam *>		cameraList;
	List<Light *>	lightList;
	List<Shape *>	objList;
	List<Shader *>	shaderList;
	List<Texture *>	textureList;
	List<AnimCurve *> animCurveList;

	BoundingBox sceneBB;
};

#endif