#ifndef Shape_h
#define Shape_h

#include "Model.h"
#include "BoundingBox.h"
#include "Tessellation.h"

class Shader;

/****************************************************************

	Class:		Shape
	Purpose:	Represents an object in space.  The object must 
				provide a tessellate method and be able to compute
				a bounding box for itself. 

	File Attributes:
				shadername <string>	- the name of the surface shader for
									  this object
				hidden     <bool>	- whether or not this object is visible
				hasnormals <bool>	- whether or not normals should be stored
									  with the object's tessellation
				hastex	   <bool>	- whether or not texture coordinates should
									  be stored with the object's tessellation

*****************************************************************/
class Shape : public Model
{
public:

	Shape();

	// attribute indices
	static int kShaderName;
	static int kHidden;
	static int kHasNormals;
	static int kHasTex;

	// attribute initialization
	virtual void initializeAttributes();

	// attribute accessors
	char	*attrShaderName() { return findAttr(kShaderName)->asString(); };
	bool&	attrHidden()	  { return findAttr(kHidden)->asBool(); };
	bool&   attrHasNormals()  { return findAttr(kHasNormals)->asBool(); };
	bool&   attrHasTex()	  { return findAttr(kHasTex)->asBool(); };



	// post-read function
	virtual bool postReadFunc( const Scene& scene );


	//
	//  Tessellation methods
	//

	// retrieves tessellation options
	virtual TessOptions getTessOptions();

	// tessellates the shape into its local tessellation object
	virtual void localTessellate( TessOptions options, SpaceType space ) {};

	// local tessellation accessor
	Tessellation *localTessellation() { return localTess; };

	// deletes the local tessellation
	virtual void deleteLocalTessellation();

	// counts the number of faces in the object's tessellation
	virtual int faceCount() = 0;



	// retrieves the surface shader for this object
	Shader* getShader() { return fShader; };

	// bounding box methods
	virtual void computeBB( const mat4& transform ) = 0;
	BoundingBox getBB() const {return fBoundingBox;};

	// determines whether or not the shape's geometry is animated
	virtual bool geometryAnimated() { return fGeometryAnimated; };

protected:

	Shader* fShader;			// surface shader
	BoundingBox fBoundingBox;	// bounding box

	bool fGeometryAnimated;		// is geometry animated?
	Tessellation *localTess;	// local tessellation
};

#endif