#include <assert.h>
#include "Tessellation.h"
#include "BoundingBox.h"
#include "Triangle.h"
#include "Cam.h"
#include "Edge.h"
#include "Shape.h"

// dummy values - used as defaults
static vec4 dummyPlaneEq(0.0);
static vec3 dummyNormal(0.0);
static vec2 dummyTex(0.0);

// temporary storage lists
int dummyV[MAX_VERTICES];
int dummyN[MAX_VERTICES];
int dummyT[MAX_VERTICES];

//
//  Constructor initializes the lists that will be needed (according to the 
//  tessellation options), based on the estimated vertex and face sizes provided.
//
Tessellation::Tessellation( TessOptions options, int faceSize, int vertexSize )
: fTessOptions(options), vList(1), nList(1), tList(1), eList(1), oList(1), fList(faceSize)
{
	// if a vertex size is not provided, estimate one
	if( vertexSize == -1 ) {
		vertexSize = faceSize*3;
	}

	// estimate number of vertices
	vList.resize( vertexSize );
	vList.resize(0);
	
	// estimate number of normals
	if( hasNormals() ) {
		nList.resize( vertexSize );
	}
	nList.resize(0);

	// estimate number of texture coords
	if( hasTex() ) {
		tList.resize( vertexSize );
	}
	tList.resize(0);

	// if we have plane equations, initialize the list to store them
	if( hasPlaneEqs() ) {
		eList.resize( faceSize );
	}
	eList.resize(0);

	// if the tessellation will store multiple shapes, initialize the
	// list of shape ptrs
	if( hasObjIds() ) {
		oList.resize( faceSize );
	}
	oList.resize(0);

	// initialize the free face list 
	for( int i = 0; i < MAX_VERTICES; i++ ) {
		freeList[i] = -1;
	}
}

//
//  Appends the local tessellation of the specified shape to this tessellation.
//  It is assumed that this tessellation's options are a superset of those of
//  the tessellation being added.  Dummy values will be filled in for any data
//  that the new tessellation cannot provide, but that is required by this 
//  tessellation (for example, if this tessellation has normals, but the
//  tessellation being added does not, we will add dummy normals instead)
//
//  This method only works on "multi-object" tessellations, as specified by the
//  kObjIds tessellation option.
//
void Tessellation::appendLocalTess( Shape *newShape )
{
	// make sure this is a multi-object tessellation
	DB_ASSERT( (fTessOptions | kTessObjIds), "ERROR - can't append to non-multi tessellation" );

	// get the local tesselation for the new shape
	Tessellation *newTess = newShape->localTessellation();

	// find the points in the vertex and face list where the new tessellation will begin
	int firstPoint = vList.size();
	int firstNewFace = fList.numLists();

	// get the new face list and tessellation options
	IntListList& newFaces = newTess->fList;
	int numNewFaces = newFaces.numLists();
	TessOptions options = newTess->getOptions();

	// add the shape ptr for the new object
	unsigned char newObjId = sList.size();
	sList.append(newShape);
	hnList.append( newTess->hasNormals() );

	// append the vertex position list
	int numV = newTess->vList.size();
	vList.append( newTess->vList );

	// append the vertex normal list, if necessary
	if( fTessOptions & kTessNormals ) {
		if( options & kTessNormals ) {

			// the other tessellation has normals, so append them
			nList.append( newTess->nList );

		} else {

			// the other tessellation has no normals, so append 
			// dummy normals
			int numN = newTess->vList.size();
			for( int i = 0; i < numN; i++ ) {
				nList.append( dummyNormal );
			}

		}
	}

	// append the vertex texture coords, if necessary
	if( fTessOptions & kTessTex ) {
		if( options & kTessTex ) {

			// the other tessellation has texture coords, so append them
			tList.append( newTess->tList );
		} else {

			// the other tessellation has no tex coords, so append dummy
			// coords
			int numT = newTess->vList.size();
			for( int i = 0; i < numT; i++ ) {
				tList.append( dummyTex );
			}
		}
	}
			
	// append the face list for the new tessellation, relocating it to be
	// relative to the first vertex index corresponding to the new tessellation
	fList.appendAndRelocate( newTess->fList, firstPoint );

	// add the object ids for the new faces
	oList.resize(oList.size()+numNewFaces);
	for( int f = 0; f < numNewFaces; f++ ) {
		oList[f+firstNewFace] = newObjId;
	}

	// append the plane equations, if necessary
	if( fTessOptions & kTessPlaneEqs ) {
		if( options & kTessPlaneEqs ) {

			// the other tessellation has plane equations, so append them
			eList.append( newTess->eList );

		} else {

			// the other tessellation has no plane equations, so append dummy
			// equations
			int numE = newTess->eList.size();
			for( int i = 0; i < numE; i++ ) {
				eList.append( dummyPlaneEq );
			}

		}
	}
}

//
//	Creates one "super-tessellation" from the local tessellations of several shapes.  
//  The local tessellations are deleted if the 'del' parameter is true.  
//
//  The new tessellation will have all the options of the subtessellations, in addition
//  to any options specified by the 'options' parameter
//
//	Returns the index of the new face in the face list.
//
Tessellation *Tessellation::mergeLocalTessellations( List<Shape *>& shapeList, TessOptions options, bool del )
{
	Tessellation *result;
	int fSize = 0;
	int vSize = 0;
	TessOptions allOptions = options;

	// loop through the visible shapes, accumulate their tessellation options, and
	// count the total number of faces and vertices
	for( int curShape = 0; curShape < shapeList.size(); curShape++ ) {

		if( shapeList[curShape]->attrHidden() ) continue;

		Tessellation *objTess = shapeList[curShape]->localTessellation();
		allOptions |= objTess->getOptions();
		fSize += objTess->size();
		vSize += objTess->vList.size();
	}

	// create the super tessellation 
	result = new Tessellation( allOptions, fSize, vSize );

	// append the local tessellations to the super tessellation, deleting them as we go
	// if required
	for( curShape = 0; curShape < shapeList.size(); curShape++ ) {

		// skip hidden shapes
		if( shapeList[curShape]->attrHidden() ) continue;

		result->appendLocalTess( shapeList[curShape] );

		// delete the local tessellation, if required
		if( del ) {
			shapeList[curShape]->deleteLocalTessellation();
		}
	}

	// shrink wrap the result
	result->shrinkWrap();

	return result;
}

//
//  Shrink wraps the vertex/texture/normal/face lists
//
void Tessellation::shrinkWrap()
{
	vList.shrinkWrap();
	nList.shrinkWrap();
	tList.shrinkWrap();
	fList.shrinkWrap();
	eList.shrinkWrap();
	sList.shrinkWrap();
	oList.shrinkWrap();
	hnList.shrinkWrap();
}

//
//  Count the number of triangles in this tessellation - each n-sided face
//  gets tessellated into n-2 triangles
//
int Tessellation::numTriangles()
{
	int numFaces = fList.numLists();
	int numTri = 0;

	for( int i = 0; i < numFaces; i++ ) {
		numTri += fList.getListSize(i) - 2;
	}

	return numTri;
}


//
//  Adds a face to the tessellation.  The face consists of 'size' vertices, whose indices
//  (in order, clockwise or counterclockwise) are given by 'vertices'.  The object id and
//  plane equation of the face can also be specified.
//
int Tessellation::addFace( int size, int *vertices, ObjId id, vec4 *planeEq )
{
	DB_ASSERT( (size < MAX_VERTICES), "ERROR - face too big" );

	// where the new face indices will be filled in
	int *vIndices;

	// check the free list to see if there is a spot big enough to hold this face
	int freeFace = freeList[size];
	if( freeFace != -1 ) {

		// there is a free face - the 'index' field of this face is -1 and
		// the 'size' field is -(index of the next free face)
		DB_ASSERT( (fList.getListSize(freeFace) <= 0), "ERROR - invalid free list index" );

		// advance the free list ptr to the next free list of the required size
		freeList[size] = fList.getListSize(freeFace);

		// get the index of the free chunk of vertices
		vIndices = fList.getList(freeFace);

	} else {

		// we add a new list and fill in the vertex indices here
		vIndices = NULL;
		int newFace = fList.addList( vertices, size );
	}

	// fill in the vertex indices, if they haven't been already
	if( vIndices ) {
		for( int i = 0; i < size; i++ ) {
			vIndices[i] = vertices[i];
		}
	}

	// append the object id for the new face, if necessary
	if( hasObjIds() ) {
		if( freeFace != -1 ) {
			oList[freeFace] = id;
		} else {
			oList.append( id );
		}
	}

	// append the plane equation for the new face, if necessary
	if( hasPlaneEqs() ) {
		if( freeFace != -1 ) {
			eList[freeFace] = planeEq ? (*planeEq) : dummyPlaneEq;
		} else {
			eList.append( planeEq ? (*planeEq) : dummyPlaneEq );
		}
	}

	// return the index of the new face
	return (freeFace != -1) ? freeFace : fList.numLists()-1;
}

//
//  Replaces a specified face with a new one, given by a new size and a new set
//  of vertex indices.  The new face can be at most as large as the face being
//  replaced.
//
//  For operations that add many faces to the tessellation that are not all necessary
//  in the end (repeated clipping during BSP creation, for example), this operation
//  will help keep the face list from growing too large.
//
int Tessellation::replaceFace( int face, int size, int *vertices )
{
	DB_ASSERT( (size < MAX_VERTICES), "ERROR - face too big" );

	// make sure the new face is at most as large as the old one
	DB_ASSERT( (size <= fList.getListSize(face)), "ERROR - can't replace face with a bigger one" );

	// replace the face
	fList.setList( face, vertices, size );
	return face;
}

//
//  Puts a face on the free list, so that its allocated space in the face list
//  may be reused.
//
void Tessellation::freeFace( int face )
{
	// add the index of this face to the chain of face indices for this face's size
	int numVertices = fList.getListSize(face);

	// the chain starts at freeList[size] and continued through the size fields of 
	// the chained faces.  We insert the newly freed face at the head of the list, 
	// ie in the free list, and bump the old list head to the size field of the specified
	// face
	fList.setListSize( face, freeList[numVertices] );

	// invalidates the face in the face list
	fList.freeList(face);

	// store the newly freed face index at the head of the free list
	freeList[numVertices] = face;
}


//
//  Splits a face along a given plane.  The two pieces of the split face are put into
//  the tessellation.  One of the new faces will replace the original face.  The indices
//  of the two new faces are returned through 'negFaceIndex' and 'posFaceIndex'.  These
//  values are -1 if the face lies entirely on one side of the plane
//
IsectType Tessellation::splitFace( int face, vec4& clipPlane, int& negFaceIndex, int& posFaceIndex )
{
	// get the list of vertices for the current face
	int numVertices = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	DB_ASSERT( (numVertices <= MAX_VERTICES), "ERROR - face too large" );

	// object id of current face
	ObjId curId = hasObjIds() ? oList[face] : 0;

	// dummy lists used to store split faces
	int posFace[MAX_VERTICES];
	int negFace[MAX_VERTICES];

	// stores which side of the splitting plane each vertex lies on
	SideType inOuts[MAX_VERTICES];

	// check to see if all points lie on one side of the plane
	bool allInside = true;
	bool allOutside = true;
	for( int i = 0; i < numVertices; i++ ) {

		SideType curInOut = inside( vList[vIndices[i]], clipPlane );
		inOuts[i] = curInOut;
		
		allInside &= ((curInOut == INSIDE) || (curInOut == INPLANE));
		allOutside &= ((curInOut == OUTSIDE) || (curInOut == INPLANE));
	}

	// check to see if all points lay on one side of the plane
	if( allInside ) {

		// all points on the negative side
		negFaceIndex = face;
		posFaceIndex = -1;
		return NEG;

	} else if( allOutside ) {

		// all points on the positive side
		negFaceIndex = -1;
		posFaceIndex = face;
		return POS;

	} else {

		// some on positive side, some on negative side, so split the face
		int posSize = 0;
		int negSize = 0;

		for( int i = 0; i < numVertices; i++ ) {

			int v1, v2, newPoint;

			v1 = vIndices[i];
			v2 = vIndices[(i+1)%numVertices];

			SideType v1where = inOuts[i];
			SideType v2where = inOuts[(i+1)%numVertices];

			int newVertex = -1;

			// if the current edge crosses the plane, find the intersection with the plane and
			// add this point to the tessellation
			if( (v1where != v2where) && (v1where != INPLANE) && (v2where != INPLANE) ) {

				double t;
				linePlaneIsect( vList[v1], vList[v2], clipPlane, t );

				// add the new point
				newPoint = addPoint( v1, v2, t );
			}

			// based on where the first vertex of this edge is, update the positive and
			// negative faces that are being built
			switch( v1where ) {

			case INSIDE:

				negFace[negSize++] = v1;

				if( v2where == OUTSIDE ) {

					negFace[negSize++] = newPoint;
					posFace[posSize++] = newPoint;
				}

				break;

			case INPLANE:

				negFace[negSize++] = v1;
				posFace[posSize++] = v1;

				break;

			case OUTSIDE:

				posFace[posSize++] = v1;

				if( v2where == INSIDE ) {
					negFace[negSize++] = newPoint;
					posFace[posSize++] = newPoint;
				}

				break;
			}
		}

		DB_ASSERT( ((posSize > 0) && (negSize > 0)), "ERROR - face split error" );

		// determine which face is the smaller one.  This one will always be smaller
		// than the original face, so we can replace the original face with it
		if( posSize <= negSize ) {

			// positive face is smaller, so replace the original face with the 
			// positive one and add the negative face as a new face
			vec4 planeEq = eList[face];
			posFaceIndex = replaceFace( face, posSize, posFace );
			negFaceIndex = addFace( negSize, negFace, curId, &planeEq );

		} else {

			// negative face is smaller, so replace the original face with the 
			// negative one and add the positive face as a new face
			vec4 planeEq = eList[face];
			negFaceIndex = replaceFace( face, negSize, negFace );
			posFaceIndex = addFace( posSize, posFace, curId, &planeEq );
		}

		return BOTH;
	}
}


//
//  Splits a face along the 3 median planes of a bounding box.  The resulting
//  faces are stored in the 'faces' array, indexed by the OctantType 
//  enumeration
//
void Tessellation::octreeSplitFace( int face, BoundingBox& bb, int *faces )
{
	// plane equations of the median planes
	vec4 planeEqs[3];

	int newXfaces[2];		// new faces resulting from the first (x) split
	int newYfaces[4];		// new faces resulting from the second (y) split

	// get the median planes
	bb.clipPlanes( xPyPzP, planeEqs );

	// split in X
	(void)splitFace( face, planeEqs[0], newXfaces[0], newXfaces[1] );

	// split pieces in Y
	for( int curFace=0; curFace < 2; curFace++ ) {
		if( newXfaces[curFace] != -1 ) {
			(void)splitFace( newXfaces[curFace], planeEqs[1], 
							 newYfaces[2*curFace], newYfaces[2*curFace+1] );
		} else {
			newYfaces[2*curFace] = -1;
			newYfaces[2*curFace+1] = -1;
		}
	}

	// split pieces in Z
	for( curFace=0; curFace < 4; curFace++ ) {
		if( newYfaces[curFace] != -1 ) {
			(void)splitFace( newYfaces[curFace], planeEqs[2],
							 faces[2*curFace], faces[2*curFace+1] );
		} else {
			faces[2*curFace] = -1;
			faces[2*curFace+1] = -1;
		}
	}
}

//
//  Tessellates a face into triangles suitable for rendering.  If the 'worldPoints'
//  list is provided, the world points are taken fron it, otherwise we use the vertex
//  positions from the tessellation.  Similarly for normals.  Screenspace points must
//  be provided
//
int Tessellation::tessellateFace( int face, bool reverse, 
								  List<vec3> *worldPoints,
								  List<vec3>& ssPoints, 
								  List<vec3> *normals, 
								  Triangle *triangles )
{
	// get the vertices of the face to be triangulated
	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	// determine where to get our worldspace vertices and normals from
	List<vec3> *wTemp = worldPoints ? worldPoints : &vList;
	List<vec3> *nTemp = normals ? normals : &nList;
	List<vec3> &w = (*wTemp);
	List<vec3> &n = (*nTemp);

	// which face normal to use (if the vertices don't have normals
	vec3 faceNormal = hasPlaneEqs() ? vec3(eList[face], VW) : dummyNormal;
	if( reverse ) {
		faceNormal = -faceNormal;
	}

	for( int tri = 0; tri < faceSize-2; tri++ ) {
		int indices[3];

		// figure out the current triangle
		if( reverse ) {
			indices[0] = tri+2;
			indices[1] = tri+1;
			indices[2] = 0;
		} else {
			indices[0] = 0;
			indices[1] = tri+1;
			indices[2] = tri+2;
		}

		for( int i = 0; i < 3; i++ ) {

			// always fill in world and screenspace points
			triangles[tri].world(i) = w[vIndices[indices[i]]];
			triangles[tri].screen(i) = ssPoints[vIndices[indices[i]]];

			// fill in normals, if necessary
			if( faceHasNormals(face) ) {
				triangles[tri].normal(i) = n[vIndices[indices[i]]];
			} else {
				triangles[tri].normal(i) = faceNormal;
			}


			// fill in texture coords, if necessary
			if( hasTex() ) {
				triangles[tri].texture(i) = tList[vIndices[indices[i]]];
			} else {
				triangles[tri].texture(i) = dummyTex;
			}
		}
	}
	return faceSize-2;
}

//
//  Computes a triage mask for a given face in a given region of space.  Edges for the 
//  current face are stored in the edge cache, since we process one face for the entire
//  image at a time.
//
TriageMask Tessellation::triageMask( int face, bool reverse, List<vec3>& ssPoints, vec4& screenBox, bool correct )
{
	// get the current face
	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	// check to see if the edges for this face are already in the edge cache
	if( face != lastFace ) {
	
		// new face, must add edges to edge cache
		for( int e = 0; e < faceSize; e++ ) {

			int v1, v2;
			if( reverse ) {
				v1 = vIndices[(e+1)%faceSize];
				v2 = vIndices[e];
			} else {
				v1 = vIndices[e];
				v2 = vIndices[(e+1)%faceSize];
			}

			// construct the edge
			edgeCache[e] = Edge( ssPoints[v1], ssPoints[v2] );
		}
	}

	// loop through the edges of the current face, computing the triage mask as we go
	bool first = true;
	TriageMask tmask;

	for( int edge = 0; edge < faceSize; edge++ ) {

		int coords[2];			// screen box intersection coords
		TriageMask edgeMask;	// triage mask for current edge

		// compute the intersection points of the current edge with the screen box
		OverlapType overlap = edgeCache[edge].boxIntersect( screenBox, coords, correct );

		// process the results
		switch( overlap ) {

		case ALL:

			// the edge covers the box, so the edge mask is full
			edgeMask = TriageMask(FULL);
			break;

		case NONE:

			// the edge is completely outside the box, so the edge mask is empty
			edgeMask = TriageMask(EMPTY);
			break;

		case SOME:

			// the edge partially covers the box, so look up the precomputed edge
			// mask for this edge
			edgeMask = TriageMask( coords[0], coords[1] );
			break;

		}

		// composite the new edge mask into the total mask
		if( first ) {
			tmask = edgeMask;
			first = false;
		} else {
			tmask += edgeMask;
		}
	}

	// keep track of which face we just did
	lastFace = face;
	return tmask;
}


//
//  Same as above, but computes a bitmask instead of a triage mask
//
Bitmask Tessellation::bitMask( int face, bool reverse, List<vec3>& ssPoints, vec4& screenBox, bool correct ) 
{
	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	if( face != lastFace ) {
		for( int e = 0; e < faceSize; e++ ) {

			int v1, v2;

			if( reverse ) {

				v1 = vIndices[(e+1)%faceSize];
				v2 = vIndices[e];

			} else {

				v1 = vIndices[e];
				v2 = vIndices[(e+1)%faceSize];

			}
			edgeCache[e] = Edge( ssPoints[v1], ssPoints[v2] );
		}
	}

	bool first = true;
	Bitmask bmask;

	for( int edge = 0; edge < faceSize; edge++ ) {
		int coords[2];
		Bitmask edgeMask;

		OverlapType overlap = edgeCache[edge].boxIntersect( screenBox, coords, correct );

		switch( overlap ) {

		case ALL:
			edgeMask = Bitmask(FULL);
			break;

		case NONE:
			edgeMask = Bitmask(EMPTY);
			break;

		case SOME:
			edgeMask = Bitmask( coords[0], coords[1] );
			break;

		}

		if( first ) {
			bmask = edgeMask;
			first = false;
		} else {
			bmask &= edgeMask;
		}
	}

	return bmask;
}


//
//  Retrieves the shape pointer for the given face
//
Shape *Tessellation::getFaceShape( int face )
{
	if( hasObjIds() ) {
		return sList[oList[face]];
	} else {
		return sList[0];
	}
}

//
//  Determines whether or not a face has normals
//
bool Tessellation::faceHasNormals( int face )
{
	if( hasObjIds() ) {

		// hnList indicates whether the object from
		// which this face cam had normals
		return hnList[oList[face]];
	
	} else {

		// this is a single-object tessellation
		return hasNormals();

	}
}


//
//  Adds a point to the tessellation.  Returns index of new point
//
int Tessellation::addPoint( const vec3& v, const vec3& n, const vec2& t )
{
	// always add the position
	vList.append( v );

	// add normal if required
	if( hasNormals() ) {
		nList.append( n );
	} 

	// add texture coord if required
	if( hasTex() ) {
		tList.append( t );
	}

	// return new index
	return vList.size()-1;
}

//
//  Adds a new point to the tessellation that is an affine combination
//  of two other points, (1-t)*v1 + t*v2
//
int Tessellation::addPoint( int v1, int v2, double t )
{
	// the interpolated position, normal, and texture coords
	vec3 newV(0.0), newN(0.0), newT(0.0);

	// interpolate position
	newV = (1.0-t)*vList[v1] + t*vList[v2];

	// interpolate normals, if necessary
	if( hasNormals() ) {
		newN = (1.0-t)*nList[v1] + t*nList[v2];
	}

	// interpolate texture coords, if necessary
	if( hasTex() ) {
		newT = (1.0-t)*tList[v1] + t*tList[v2];
	}

	// add the point
	return addPoint( newV, newN, newT );
}

//
//  Transforms the vertex positions using the given matrix, storing the results
//  in 'vTrans' if it is non-NULL or in place if 'vTrans' is NULL.
//
void Tessellation::transformVertices( const mat4& vMatrix, List<vec3>* vTrans )
{
	// find where the results will go
	List<vec3>& resultList = vTrans ? (*vTrans) : vList;

	// resize result list
	int numVertices = vList.size();
	resultList.resize(numVertices);

	// transform the points
	for( int i = 0; i < numVertices; i++ ) {
		resultList[i] = vMatrix * vList[i];
	}
}

//
//  Transforms the normals using the given matrix, storing the results
//  in 'nTrans' if it is non-NULL or in place if 'nTrans' is NULL.
//
void Tessellation::transformNormals( const mat4& nMatrix, List<vec3> *nTrans )
{
	// only transform normals if they are present
	if( hasNormals() ) {

		// find where to store results
		List<vec3>& resultList = nTrans ? (*nTrans) : nList;

		// resize result list
		int numVertices = nList.size();
		resultList.resize(numVertices);

		// transform the normals
		for( int i = 0; i < numVertices; i++ ) {
			resultList[i] = vec3( nMatrix * (vec4(nList[i],0.0)), VW);
			resultList[i].normalize();
		}
	}
}

//
//  Computes the plane equations for the faces, based on the 'points' vertex
//  array if it is specified, or based on the tessellation vertex positions if
//  it is not.  The results are stored in the 'results' array if it is specified
//  or in place if it is not.
//
void Tessellation::computePlaneEqs( List<vec3> *points, List<vec4> *results )
{
	// only do this if the tessellation has plane equations
	if( hasPlaneEqs() ) {

		// find which vertices to use
		List<vec3>& inputs = points ? (*points) : vList;

		// find where to store results
		List<vec4>& outputs = results ? (*results) : eList;
		
		// resize outputs
		int numFaces = fList.numLists();
		outputs.resize( numFaces );

		// compute each face plane equation
		for( int f = 0; f < numFaces; f++ ) {

			// stores vertex positions of this face
			vec3 vertices[MAX_VERTICES];

			// get the list of vertex indices
			int faceSize = fList.getListSize(f);
			int *vIndices = fList.getList(f);

			// store the vertices
			for( int v = 0; v < faceSize; v++ ) {
				vertices[v] = inputs[vIndices[v]];
			}

			// compute the plane equation
			outputs[f] = computePlaneEq( vertices, faceSize );
		}
	}
}

//
//  Prints a face
//
void Tessellation::printFace( ostream& s, int face, List<vec3> *ssPoints )
{

	cout << "Face " << face << endl;

	cout << "Plane Eq = " << (hasPlaneEqs() ? eList[face] : dummyPlaneEq) << endl;
	cout << "Shape ptr = " << (hasObjIds() ? sList[oList[face]] : NULL) << endl;

	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	for( int j = 0; j < faceSize; j++ ) {
		int v = vIndices[j];

		cout << vList[v] << (hasNormals() ? nList[v] : -1) << (hasTex() ? tList[v] : -1);
		
		if( ssPoints ) {
			cout << (*ssPoints)[v];
		}

		cout << endl;

	}

	cout << endl;

}
	












































/*
void Tessellation::octreeSplitFace( int face, BoundingBox& bb, int *faces )
{
	vec4 planeEqs[3];

	int newXfaces[2];
	int newYfaces[4];

	bb.clipPlanes( xPyPzP, planeEqs );

	// split in X
	(void)splitFace( face, planeEqs[0], newXfaces[0], newXfaces[1] );

	// split pieces in Y
	for( int curFace=0; curFace < 2; curFace++ ) {
		if( newXfaces[curFace] != -1 ) {
			(void)splitFace( newXfaces[curFace], planeEqs[1], 
							 newYfaces[2*curFace], newYfaces[2*curFace+1] );
		} else {
			newYfaces[2*curFace] = -1;
			newYfaces[2*curFace+1] = -1;
		}
	}

	// split pieces in Z
	for( curFace=0; curFace < 4; curFace++ ) {
		if( newYfaces[curFace] != -1 ) {
			(void)splitFace( newYfaces[curFace], planeEqs[2],
							 faces[2*curFace], faces[2*curFace+1] );
		} else {
			faces[2*curFace] = -1;
			faces[2*curFace+1] = -1;
		}
	}
}



IsectType Tessellation::clipFace( int face, vec4& clipPlane )
{
	int numVertices = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	SideType inOuts[MAX_VERTICES];
	assert( numVertices <= MAX_VERTICES );

	bool hasNormals = bList[face];

	bool allInside = true;
	bool allOutside = true;

	// check to see if all points lie on one side of the plane
	for( int i = 0; i < numVertices; i++ ) {

		SideType curInOut = inside( vList[vIndices[i]], clipPlane );
		inOuts[i] = curInOut;
		
		allInside &= ((curInOut == INSIDE) || (curInOut == INPLANE));
		allOutside &= ((curInOut == OUTSIDE) || (curInOut == INPLANE));
	}

	if( allInside ) {
		return POS;

	} else if( allOutside ) {
		return NEG;

	} else {

		int newFace[MAX_VERTICES];
		int newSize = 0;

		// otherwise, some are inside, some are outside
		for( i = 0; i < numVertices; i++ ) {

			double t;
			int i1 = vIndices[i];
			int i2 = vIndices[(i+1)%numVertices];

			int v1where = inOuts[i];
			int v2where = inOuts[(i+1)%numVertices];

			if( linePlaneIsect( vList[i1], vList[i2], clipPlane, t ) ) {

				int newPoint = addPoint( i1, i2, t );

				switch( v1where ) {

				case INSIDE:

					newFace[newSize++] = i1;

					if( v2where == OUTSIDE ) {
						newFace[newSize++] = newPoint;
					}

					break;

				case INPLANE:

					newFace[newSize++] = i1;

					break;

				case OUTSIDE:

					if( v2where == INSIDE ) {
						newFace[newSize++] = newPoint;
					}

					break;
				}

			} else {

				if( v1where == INSIDE ) {
					newFace[newSize++] = i1;
				}
			}
		}

		assert( newSize > 0 );

		if( newSize <= numVertices ) {

			replaceFace( newFace, newSize, face );
		} else {
			freeFace( face );
			addFace( newFace, newSize, sList[face], hasNormals, &(eList[face]) );
		}

		return BOTH;
	}
}


int Tessellation::tessellateFace( int face, bool reverse, List<vec3>& ssPoints, Triangle *triangles )
{
	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	bool hasNormals = bList[face];
	vec3 planeEq;

	for( int tri = 0; tri < faceSize-2; tri++ ) {
		int indices[3];

		if( reverse ) {
			indices[0] = tri+2;
			indices[1] = tri+1;
			indices[2] = 0;

		} else {
			indices[0] = 0;
			indices[1] = tri+1;
			indices[2] = tri+2;
		}

		for( int i = 0; i < 3; i++ ) {
			triangles[tri].world(i) = vList[vIndices[indices[i]]];
			triangles[tri].texture(i) = tList[vIndices[indices[i]]];
			triangles[tri].screen(i) = ssPoints[vIndices[indices[i]]];

			if( hasNormals ) {
				triangles[tri].normal(i) = nList[vIndices[indices[i]]];
			} else {
				triangles[tri].normal(i) = vec3( eList[face], VW );
			}
		}
	}
	return faceSize-2;


}

TriageMask Tessellation::triageMask( int face, bool reverse, List<vec3>& ssPoints, vec4& screenBox, bool correct )
{
	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	if( face != lastFace ) {
		// this is the first time we've rendered this face, so preprocess the edges
		for( int e = 0; e < faceSize; e++ ) {

			int v1, v2;

			if( reverse ) {

				v1 = vIndices[(e+1)%faceSize];
				v2 = vIndices[e];

			} else {

				v1 = vIndices[e];
				v2 = vIndices[(e+1)%faceSize];

			}
			edgeCache[e] = Edge( ssPoints[v1], ssPoints[v2] );
		}
	}

	bool first = true;
	TriageMask tmask;

	for( int edge = 0; edge < faceSize; edge++ ) {
		int coords[2];
		TriageMask edgeMask;

		OverlapType overlap = edgeCache[edge].boxIntersect( screenBox, coords, correct );

		switch( overlap ) {

		case ALL:
			edgeMask = TriageMask(FULL);
			break;

		case NONE:
			edgeMask = TriageMask(EMPTY);
			break;

		case SOME:
			// round to outside
			edgeMask = TriageMask( coords[0], coords[1] );
			break;

		}

		if( first ) {
			tmask = edgeMask;
			first = false;
		} else {
			tmask += edgeMask;
		}
	}

	lastFace = face;
	return tmask;
}


Bitmask Tessellation::bitMask( int face, bool reverse, List<vec3>& ssPoints, vec4& screenBox, bool correct ) 
{
	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	if( face != lastFace ) {
		// this is the first time we've rendered this face, so preprocess the edges
		for( int e = 0; e < faceSize; e++ ) {

			int v1, v2;

			if( reverse ) {

				v1 = vIndices[(e+1)%faceSize];
				v2 = vIndices[e];

			} else {

				v1 = vIndices[e];
				v2 = vIndices[(e+1)%faceSize];

			}
			edgeCache[e] = Edge( ssPoints[v1], ssPoints[v2] );
		}
	}

	bool first = true;
	Bitmask bmask;

	for( int edge = 0; edge < faceSize; edge++ ) {
		int coords[2];
		Bitmask edgeMask;

		OverlapType overlap = edgeCache[edge].boxIntersect( screenBox, coords, correct );

		switch( overlap ) {

		case ALL:
			edgeMask = Bitmask(FULL);
			break;

		case NONE:
			edgeMask = Bitmask(EMPTY);
			break;

		case SOME:
			edgeMask = Bitmask( coords[0], coords[1] );
			break;

		}

		if( first ) {
			bmask = edgeMask;
			first = false;
		} else {
			bmask &= edgeMask;
		}
	}

	return bmask;
}

void Tessellation::printFace( ostream& s, int face, List<vec3> *ssPoints )
{
	cout << "Face " << face << endl;

	cout << "Plane Eq = " << eList[face] << endl;
	cout << "Shape ptr = " << sList[face] << endl;

	int faceSize = fList.getListSize(face);
	int *vIndices = fList.getList(face);

	for( int j = 0; j < faceSize; j++ ) {
		int v = vIndices[j];

		cout << vList[v] << nList[v] << tList[v];
		
		if( ssPoints ) {
			cout << (*ssPoints)[v];
		}

		cout << endl;

	}

	cout << endl;
}
	

ostream& operator<<( ostream& s, Tessellation& t )
{
	int numFaces = t.size();

	for( int i = 0; i < numFaces; i++ ) {

		cout << "Face " << i << endl;

		cout << "Plane Eq = " << t.eList[i] << endl;
		cout << "Shape ptr = " << t.sList[i] << endl;

		int faceSize = t.fList.getListSize(i);
		int *vIndices = t.fList.getList(i);

		for( int j = 0; j < faceSize; j++ ) {
			int v = vIndices[j];

			cout << t.vList[v] << t.nList[v] << t.tList[v] << endl;

		}

		cout << endl;
	}
	return s;
}
*/


/*
void Tessellation::tessObjStart( Shape *curShape )
{
	DB_ASSERT( (fFirstObjFace == -1), "ERRPR - tessObjStart not re-entrant" );

	fCurShape = curShape;
	fFirstObjV = vList.size();
	fFirstObjFace = fList.numLists();
}

void Tessellation::tessObjEnd()
{
	DB_ASSERT( (fFirstObjFace != -1), "ERROR - no tess object started" );

	fFirstObjFace = -1;
}

int Tessellation::tessObjAddFace( int size, int *vertices )
{
	int dummyPoints[MAX_VERTICES];

	// remap vertex indices
	for( int i = 0; i < size; i++ ) {
		dummyV[i] = vertices[i] + fFirstObjV;
	}

	return addFace( size, dummyPoints, fCurShape, NULL );
}
	
void Tessellation::tessObjTransform( const mat4& vMat, const mat4& nMat )
{
	int lastV = vList.size();
	for( int i = fFirstObjV; i < lastV; i++ ) {
		vList[i] = vMat * vList[i];
	}

	if( fHasNormals ) {
		int lastN = nList.size();
		for( int i = fFirstObjN; i < lastN; i++ ) {
			nList[i] = nMat * nList[i];
		}
	}
}

void Tessellation::tessObjPlaneEqs()
{
	if( fHasPlaneEqs ) {
		int lastF = fList.numLists();

		for( int f = fFirstObjFace; f < lastF; f++ ) {
			vec3 vertices[MAX_VERTICES];

			int faceSize = fList.getListSize(f);
			int *vIndices = fList.getList(f);

			for( int v = 0; v < faceSize; v++ ) {
				vertices[v] = vList[vIndices[v]];
			}
			
			eList[f] = computePlaneEq( vertices, faceSize );
		}
	}
}
*/