#ifndef Tessellation_h
#define Tessellation_h

#include <ostream.h>
#include "algebra3.h"
#include "List.h"
#include "TriageMask.h"
#include "Edge.h"
#include "IntListList.h"

class Shape;
class BoundingBox;
class Triangle;
class Cam;

//  The TessOptions bitfield determines
//  which fields are stored in the tessellation
//  data.  Vertex information is always stored.
typedef unsigned char TessOptions;
enum {
		kTessNone=0x00,			// no additional data
		kTessNormals=0x01,		// per-vertex normals
		kTessTex=0x02,			// per-vertex texture coords
		kTessPlaneEqs=0x04,		// per-face plane equations
		kTessObjIds=0x08		// per-face object identifiers
};

// ObjIds are used to identify the object to which each face belongs.
typedef unsigned char ObjId;

// maximum allowed face size
const int MAX_VERTICES = 45;

/****************************************************************

	Class:		Tessellation
	Purpose:	Represents a tessellation of a surface or several
				surfaces into polygonal facets.  Can store normals,
				texture coordinates, and plane equations as well as
				vertices.

				The per-vertex data is stored in several lists, one
				for vertex positions, one for normals, and one for 
				texture coordinates.  There is a one-to-one correspondance
				between each array, ie the position,normal, and texture
				coordinates of a particular vertex will each be stored
				at the same index in the respective arrays.

				Facet information is stored as a list of lists of
				vertex indices, each sublist denoting a particular
				facet of the tessellation.

				A tessellation object can be used to store the 
				tessellation of a single shape, or it can be used
				to accumulate the tessellations of several shapes.
				In this case, object ids are stored for each face.

*****************************************************************/
class Tessellation
{
public:

	// constructor specifies options to be stored, estimated number of
	// faces, and estimated number of vertices
	Tessellation( TessOptions tessOptions, int faceSize, int vertSize=-1 );
	~Tessellation() {};


	/*******************************************************************
	    Tessellation option methods - query which pieces of information
		  							  are stored in this tessellation.
	*******************************************************************/

	// get the tessellation options for this tessellation
	TessOptions& getOptions() { return fTessOptions; };

	// option accessors
	bool hasNormals() { return (fTessOptions & kTessNormals) ? true : false; };
	bool hasTex() { return (fTessOptions & kTessTex) ? true : false; };
	bool hasPlaneEqs() { return (fTessOptions & kTessPlaneEqs) ? true : false; };
	bool hasObjIds() { return (fTessOptions & kTessObjIds) ? true : false; };



	/**********************************************************************
	    Tessellation level methods - used to combine several tessellations 
									 into one.
	***********************************************************************/

	// appends the local tessellation for a given shape into this tessellation
	void appendLocalTess( Shape *newShape );

	// merges local tessellations for several shapes into one
	static Tessellation *mergeLocalTessellations( List<Shape *>& shapes, TessOptions options, bool del = true );

	// returns the number of faces in the tessellation
	int size() { return fList.numLists(); };

	// returns the number of triangles in the tessellation
	int numTriangles();

	// shrinkwraps all the lists in the tessellation so no space is wasted
	void shrinkWrap();
	

	/**********************************************************************
		Face level methods - operations on faces of the tessellation
	**********************************************************************/

	// adds a face to the tessellation
	int addFace( int size, int *points, ObjId objId, vec4 *planeEq );

	// replaces a specified face with another face of equal or smaller size
	int replaceFace( int face, int size, int *points );

	// marks a face as deleted from the tessellation
	void freeFace( int face );

	// splits a face along a particular plane and returns the indices of the new
	// pieces.
	IsectType splitFace( int face, vec4& planeEq, int& negFaceIndex, int& posFaceIndex );

	// clips a face to the positive side of a specified plane
	IsectType clipFace( int face, vec4& planeEq );
	
	// splits a face along the 3 median planes of a given bounding box
	void octreeSplitFace( int face, BoundingBox& bb, int *faces );

	// tessellates a face into triangles
	int tessellateFace( int face, bool reverse, List<vec3> *worldPoints, List<vec3>& ssPoints, List<vec3> *normals, Triangle *triangles );

	// compute bitmasks and triagemasks for faces
	Bitmask bitMask( int face, bool reverse, List<vec3>& ssPoints, vec4& imageRegion, bool correct );
	TriageMask triageMask( int face, bool reverse, List<vec3>& ssPoints, vec4& imageRegion, bool correct );

	// retrieves a pointer to the shape to which the face belongs
	Shape *getFaceShape( int face );

	// determines whether a face has normals or not
	bool faceHasNormals( int face );

	

	/***********************************************************************
		Vertex level methods - operations on vertices in the tessellation
	************************************************************************/

	// adds a point having a specified position, normal, and texture coordinates
	int addPoint( const vec3& v, const vec3& n, const vec2& t );

	// adds a point that is an affine combination of two other points already
	// in the tessellation
	int addPoint( int v1, int v2, double t );

	// transforms all the vertices of the tessellation
	void transformVertices( const mat4& vMatrix, List<vec3> *vTrans = NULL);

	// transforms all the normals of the tessellation
	void transformNormals ( const mat4& nMatrix, List<vec3> *nTrans = NULL);

	// computes plane equations for all the faces in the tessellation
	void computePlaneEqs( List<vec3> *points, List<vec4> *results );



	/************************************************************************
		Tessellation data members
	*************************************************************************/

	// vertex data
	List<vec3>	vList;		// vertex position list
	List<vec3>  nList;		// vertex normal list
	List<vec2>  tList;		// vertex texture coord list

	// face data
	IntListList fList;		// vertex indices of faces
	List<vec4>  eList;		// plane equation list
	
	// shape data
	List<Shape *> sList;	// list of pointers to shapes stored in this
							// tessellation
	List<ObjId> oList;		// object id list - list of indices into the 
							// shape pointer array

	List<bool> hnList;		// indicates which shapes in the tessellation 
							// have normals


	// output routines
	void printFace( ostream& s, int face, List<vec3> *ssPoints );

private:

	int freeList[MAX_VERTICES];			// keeps track of free segments in the face list
	Edge edgeCache[MAX_VERTICES];		// caches the edges of the current triangle
	int lastFace;						// the last face that was rendered from the tessellation
										// (helps determine when the cache is invalidated)

	TessOptions fTessOptions;			// the options that determine what data is stored in 
										// this tessellation
};

#endif