#include "TriageMask.h"
#include "stdio.h"

// definitions of subregion/subinterval sizes
const int TriageMask::subIntervals = 16;
const int Bitmask::subIntervals = 16;
const int TriageMask::maskSize = 8;
const int Bitmask::maskSize = 8;

// definition of precomputed triage mask table
TriageMask TriageMask::pre[NUMSUB][NUMSUB];

// definition of precomputed bitmask table
Bitmask Bitmask::pre[NUMSUB][NUMSUB];

// definition of preset rectangular region bitmasks for region bit setting
Bitmask Bitmask::preSumAbove[MASKSIZE][MASKSIZE];
Bitmask Bitmask::preSumBelow[MASKSIZE][MASKSIZE];

// definition of precomputed bitcount table
unsigned char Bitmask::preBitcounts[256];

// initialization flag definitions
bool TriageMask::initialized = false;
bool Bitmask::initialized = false;

// 
//  The following macros are used to implement a bit-counting operation
//  for 32-bit integers.  MASKn() masks out the n'th block of 8 bits and
//  shifts them to the low end of the word.  This result is then used as
//  an index into the precomputed 'preBitCounts' table, which stores the 
//  number of bits set in every possible 8-bit integer.
//
//  The BITCOUNT() macro simply counts the bits by doing 4 table lookups
//  on the 4 8-bit sections of the work and sums the results to arrive at
//  a bit count for the entire word.
//
#define MASK1(b)  ((b)&0x000000FF)
#define MASK2(b)  (((b)&0x0000FF00)>>8)
#define MASK3(b)  (((b)&0x00FF0000)>>16)
#define MASK4(b)  (((b)&0xFF000000)>>24)
#define BITCOUNT(b) (preBitcounts[MASK1(b)]+preBitcounts[MASK2(b)]+preBitcounts[MASK3(b)]+preBitcounts[MASK4(b)])

//
//  Initializes an empty mask
//
Bitmask::Bitmask() : b1(0), b2(0)
{
}

//
//  Copy constructor
//
Bitmask::Bitmask( Bitmask& b ) : b1(b.b1), b2(b.b2)
{
}

//
//  Initializes with the pre-computed mask for the specified edge
//
Bitmask::Bitmask( int in, int out )
{
	DB_ASSERT( initialized, "ERROR - Bitmask tables not initialized" );
	(*this) = pre[in][out];
}

//
//  Initializes an empty or full mask
//
Bitmask::Bitmask( FillType which ) 
{
	if( which == FULL ) {
		b1 = b2 = ~0;
	} else {
		b1 = b2 = 0;
	}
}

//
//  AND operator ANDs two bitmasks together
//
Bitmask operator& ( Bitmask& left, Bitmask& right ) 
{
	return( Bitmask(left) &= right );
}

//
//  In-place AND operator ANDs this bitmask with another one.
//
Bitmask& Bitmask::operator&= ( Bitmask& opnd )
{
	b1 &= opnd.b1;
	b2 &= opnd.b2;

	return *this;
}

//
//  OR operator ORs two bitmasks together
//
Bitmask operator| ( Bitmask& left, Bitmask& right ) 
{
	return( Bitmask(left) |= right );
}

//
//  In-place OR operator ORs this bitmask with another one
//
Bitmask& Bitmask::operator|= ( Bitmask& opnd )
{
	b1 |= opnd.b1;
	b2 |= opnd.b2;

	return *this;
}

//
//  Complement operator flips the bits of the specified bitmask
//
Bitmask operator~ ( Bitmask& opnd ) 
{
	Bitmask n(opnd);
	return n.complement();
}

//
//  In-place complement operator flips this bitmask
//
Bitmask& Bitmask::complement(void)
{
	b1 = ~b1;
	b2 = ~b2;
	return (*this);
}

//
//  Tests the specified bit.
//
int Bitmask::testBit( int x, int y ) 
{
	// flatten the index
	int index = maskSize*y + x;
	int retVal;

	// figure out which word the index lies in and
	// test the appropriate position in that word
	if( index < 32 ) {
		int b = 1<<index;
		retVal = b&b1;
	} else {
		int b = 1<<(index-32);
		retVal = b&b2;
	}

	return retVal;
}

//
//  Sets the specified bit
//
void Bitmask::setBit( int x, int y ) 
{
	// flatten the index
	int index = maskSize*y + x;

	// set the specified bit
	if( index < 32 ) {
		int b = 1<<index;
		b1 |= b;
	} else {
		int b = 1<<(index-32);
		b2 |= b;
	}
}

//
//  Clears the specified bit
//
void Bitmask::clearBit( int x, int y )
{
	// flatten the index
	int index = maskSize*y + x;

	// figure out where the bit lies and clear it
	if( index < 32 ) {
		int b = ~(1<<index);
		b1 &= b;
	} else {
		int b = ~(1<<(index-32));
		b2 &= b;
	}
}


//
//  Sets the block of bits from (x1,y1) to (x2,y2)
//
void Bitmask::setBits( int x1, int y1, int x2, int y2 ) 
{
	if( (x1 <= x2) && (y1 <= y2) ) {
	
		DB_ASSERT( ((x1>=0)&&(x1<=7)&&(x2>=0)&&(x2<=7)&&
					(y1>=0)&&(y1<=7)&&(y2>=0)&&(y2<=7)),
					"ERROR - bitmask index out of range" );

		// to set the bits from (x1,y1)->(x2,y2), just AND together
		// the precomputed mask with all bits above (x1,y1) set with
		// the precomputed mask with all bits below (x2,y2) set.
		(*this) |= preSumAbove[x1][y1] & preSumBelow[x2][y2];
	}
}

//
//  Clears the bitmask
//
void Bitmask::clear(void)
{
	b1 = b2 = 0;
}

//
//  Tests for an empty bitmask
//
bool Bitmask::empty(void)
{
	return (b1|b2) ? false : true;
}

//
//  Fills the bitmask
//
void Bitmask::fill(void)
{
	b1 = b2 = ~0;
}

//
//  Tests for a full bitmask
//
bool Bitmask::full(void)
{
	return (~b1|~b2) ? false : true;
}

//
//  Counts the number of bits in the mask - just sum the bitcounts
//  for the 2 words.
//
int Bitmask::numBits(void) 
{
	return BITCOUNT(b1)+BITCOUNT(b2);
}

//
//  Initializes the precomputed tables required by the Bitmask class, 
//  ie the 'pre' table, the 'preSumAbove' and 'preSumBelow' tables, and
//  the 'preBitCounts' table.  For the 'pre' table (the precomputed bitmasks),
//  it is read from the file "Bitmask.data" in the current directory if it exists.
//  Otherwise the table is recomputed and stored in that file.
//
void Bitmask::initialize()
{
	int in, out;

	// initialize bitcount table - preBitcounts[i] is the number
	// of bits set in the 8-bit character i.
	for( int i = 0; i < 256; i++ ) {

		int total = 0;
		for( int bit = 0; bit < 8; bit++ ) {
			if( i & (1<<bit) ) total++;
		}

		preBitcounts[i] = total;
	}


	// initialize pre-summed tables - preSumAbove[i][j] has all
	// bits above (i,j) set, preSumBelow[i][j] has all bits below
	// (i,j) set.
	for( i = 0; i < maskSize; i++ ) {
		for( int j = 0; j < maskSize; j++ ) {

			// compute the mask with all bits above this one set
			for( int ii = i; ii < maskSize; ii++ ) {
				for( int jj = j; jj < maskSize; jj++ ) { 

					preSumAbove[i][j].setBit(ii,jj);

				}
			}

			// compute the mask with all bits below this one set
			for( ii = 0; ii <= i; ii++ ) {
				for( int jj = 0; jj <= j; jj++ ) { 

					preSumBelow[i][j].setBit(ii,jj);

				}
			}

		}
	}

	//
	//  Now we compute the 'pre' table, which stores precomputed bitmasks
	//  for all possible edge entry and exit points on the square.
	//

	// first, try to read the file
	FILE *bmaskReadFile = fopen( "Bitmask.data", "rb" );
	int numIntervals = 4 * maskSize * subIntervals;

	if( bmaskReadFile == NULL ) {

		// bitmask file doesn't exist, so compute the bitmasks
		cout << "Computing bitmasks..." << endl;
		for( in = 0; in <= numIntervals; in++ ) {
			for( out = 0; out <= numIntervals; out++ ) {
				pre[in][out] = computeMask( in, out );
			}
		}

		cout << "Done computation." << endl;
		cout << "Writing bitmasks..." << endl;

		// write the precomputed bitmasks to the file
		FILE *bmaskWriteFile = fopen( "Bitmask.data", "wb" );
		for( in = 0; in <= numIntervals; in++ ) {
			fwrite( (void *)(pre[in]), sizeof(Bitmask), numIntervals, bmaskWriteFile );
		}

		fclose( bmaskWriteFile );
		cout << "Done writing." << endl;

	} else {

		// the file exists, so read in the data one row at a time
		cout << "Reading bitmasks..." << endl;
		for( in = 0; in <= numIntervals; in++ ) {
			fread( (void *)(pre[in]), sizeof(Bitmask), numIntervals, bmaskReadFile );
		}
		fclose( bmaskReadFile );

		cout << "Done reading." << endl;
	}

	Bitmask::initialized = true;
}

//
//  Computes the bitmask for the edge entering the square at index
//  'in' on the perimeter, and exiting the square at index 'out' on
//  the perimeter.
//
Bitmask Bitmask::computeMask( int in, int out )
{
	// empty bitmask for degenerate case
	if( in == out ) {
		return Bitmask();
	}

	// get positions of entry and exit points
	vec2 inXY = indexToPos(in);
	vec2 outXY = indexToPos(out);
	double dy = outXY[VY] - inXY[VY];
	double dx = outXY[VX] - inXY[VX];

	Bitmask C;

	if( dx == 0.0 ) {

		// the edge is vertical, so determine which side of the edge is 'inside'
		int leftInside = (dy > 0.0);

		if( leftInside ) {
			// the inside is on the left, so set the bits to the left of the 
			// edge
			C.setBits(0,0,fastDblToInt(floor(inXY[VX]-0.5)),maskSize-1);

		} else {
			// the inside is on the right, so set the bits to the right of the
			// edge
			C.setBits( fastDblToInt(floor(inXY[VX]+0.5)),0,maskSize-1,maskSize-1);
		}

	} else if ( dy == 0.0 ) {

		// the edge is horizontal, so determine which side of the edge is 'inside'
		int topInside = (dx > 0.0);

		if( topInside ) {
			// inside is above the edge, so set all the bits above the edge
			C.setBits(0,fastDblToInt(floor(inXY[VY]+0.5)),maskSize-1,maskSize-1);

		} else {
			// inside is below the edge, so set all the bits below the edge
			C.setBits(0,0,maskSize-1,fastDblToInt(floor(inXY[VY]-0.5)));
		}

	} else {
		
		// edge is neither horizontal nor vertical
		double slope = dy / dx;

		vec2 low, high;		// store low and high positions in Y
		int underCov = (dx < 0.0);	// determines whether samples that are under
									// the edge (in Y) are considered inside or not
		int leftInside;				// determines whether samples that are to the left
									// of the edge are considered inside or not

		// sort the perimeter intersection points in Y
		if( dy >= 0.0 ) {
			low = inXY;
			high = outXY;
		} else {
			low = outXY;
			high = inXY;
		}

		// figure out the first and last rows of samples intersected by the edge
		int firstRow = fastDblToInt(floor(low[VY]+0.5));
		int lastRow = fastDblToInt(floor(high[VY]-0.5));

		// set the samples that are covered but lie outside the y-range of the edge
		if( underCov ) {

			// samples below the edge are inside, so set them
			for( int row = 0; row < firstRow; row++ ) {
				C.setBits(0,row,maskSize-1,row);
			}

			// if the slope is negative, samples to the left are inside
			leftInside = (slope < 0.0);

		} else {

			// samples above the edge are inside, so set them
			for( int row = maskSize-1; row > lastRow; row-- ) {
				C.setBits(0,row,maskSize-1,row);
			}

			// if the slope is positive, samples to the left are inside
			leftInside = (slope > 0.0);
		}

		// now, loop through the sample rows that are intersected by the edge

		double curX, xIncr;	// current x-position on the row and the amount by
							// which x gets incremented between rows

		xIncr = 1.0 / slope;	// x increments by 1/slope between scanlines
		curX = low[VX] + (((double)firstRow + 0.5) - low[VY])*xIncr;	// starting x value on 
																		// the first scanline intersected

		// loop through the scanlines, filling in the appropriate samples
		for( int curRow = firstRow; curRow <= lastRow; curRow++ ) {
			
			// figure out the first sample column to the right of the edge
			int firstCol = fastDblToInt(floor(curX-0.5));
			
			// fill in all the samples to the left or right of this column, depending
			// on which side is 'inside'
			if( leftInside ) {
				C.setBits( 0, curRow, firstCol, curRow );
			} else {
				C.setBits( firstCol+1, curRow, maskSize-1, curRow );
			}

			// increment for the next scanline
			curX += xIncr;
		}
	}

	return C;
}

//
//  Coverts a perimeter index to a 2d position
//
vec2 Bitmask::indexToPos( int index )
{
	vec2 pos;

	// how many intervals per side?
	int side = maskSize * subIntervals;

	// figure out which side the point is on, and compute the position
	switch( index/side ) {
	case 0:
		pos[VX] = index;
		pos[VY] = 0.0;
		break;
	case 1:
		pos[VX] = side;
		pos[VY] = index - side;
		break;
	case 2:
		pos[VX] = (3*side) - index;
		pos[VY] = side;
		break;
	case 3:
		pos[VX] = 0.0;
		pos[VY] = (4*side) - index;
		break;
	}

	pos /= subIntervals;

	return pos;
}

//
//  Output operator
//
ostream& operator<< ( ostream& os, Bitmask& b )
{
	for( int i = Bitmask::maskSize-1; i >= 0; i-- ) {
		for( int j = 0; j < Bitmask::maskSize; j++ ) { 
			os << (b.testBit(j,i) ? 1 : 0);
		}
		os << endl;
	}

	return os;
}


// 
//  Constructor initializes an empty triage mask (C empty, V full)
//
TriageMask::TriageMask()
{
	v.fill();
}

//
//  Constructor initializes C and V masks
//
TriageMask::TriageMask( Bitmask& cov, Bitmask& vac ) 
: c(cov), v(vac)
{
}

//
//  Constructor initializes C mask, assumes V = ~C
//
TriageMask::TriageMask( Bitmask& cov ) : c(cov)
{
	v = ~c;
}

// 
//  Copy constructor
//
TriageMask::TriageMask( TriageMask& t ) 
: c(t.c), v(t.v)
{
}

//
//  Constructor initializes full or empty mask
//
TriageMask::TriageMask( FillType which ) 
{
	if( which == FULL ) {
		c.fill();
		v.clear();
	} else {
		c.clear();
		v.fill();
	}
}

//
//  Constructor initializes with precomputed triage mask for the edge
//  having the given entry and exit points
//
TriageMask::TriageMask( int in, int out )
{
	(*this) = pre[in][out];
}

//
//  Infix compositing operator
//
TriageMask operator+( TriageMask& l, TriageMask& r )
{
	return TriageMask(l)+=r;
}

//
//  In-place compositing operator.  Composites the new mask with this
//  one, by ANDing together the covered masks and ORing together the
//  vacant masks.
//
TriageMask& TriageMask::operator+= ( TriageMask& t )
{
	c &= t.c;
	v |= t.v;

	return (*this);
}

//
//  Initializes the precomputed triage mask table.  First tries to
//  read the table from the file "TriageMask.data", otherwise computes
//  all the masks and writes the file.
//
void TriageMask::initialize()
{
	// compute number of intervals around the perimeter of a square
	int numIntervals = 4 * maskSize * subIntervals;

	// try to open the precomputed triage mask file
	FILE *tmaskReadFile = fopen( "TriageMask.data", "rb" );
	if( tmaskReadFile == NULL ) {

		// file doesn't exist, so compute the triage masks
		cout << "Computing TriageMasks..." << endl;
		for( int in = 0; in <= numIntervals; in++ ) {
			for( int out = 0; out <= numIntervals; out++ ) {
				pre[in][out] = computeMask( in, out );
			}
		}

		cout << "Done computation." << endl;
		cout << "Writing TriageMasks..." << endl;

		// write the masks to the file
		FILE *tmaskWriteFile = fopen( "TriageMask.data", "wb" );
		for( in = 0; in <= numIntervals; in++ ) {
			fwrite( (void *)(pre[in]), sizeof(TriageMask), numIntervals, tmaskWriteFile );
		}

		fclose( tmaskWriteFile );
		cout << "Done writing." << endl;

	} else {

		// the file exists, so read the masks from the file
		cout << "Reading TriageMasks..." << endl;
		for( int in = 0; in <= numIntervals; in++ ) {
			fread( (void *)(pre[in]), sizeof(TriageMask), numIntervals, tmaskReadFile );
		}

		fclose( tmaskReadFile );
		cout << "Done reading." << endl;
	}

	TriageMask::initialized = true;
}

//
//  Computes the triage mask for the edge entering at index 'in' and exiting
//  at index 'out'.
//
TriageMask TriageMask::computeMask( int in, int out )
{
	// empty mask for degenerate case
	if( in == out ) {
		Bitmask C, V;
		V.setBits(0,0,maskSize-1,maskSize-1);
		return TriageMask(C,V);
	}

	// find perimeter intersection points
	vec2 inXY = indexToPos(in);
	vec2 outXY = indexToPos(out);
	double dy = outXY[VY] - inXY[VY];
	double dx = outXY[VX] - inXY[VX];

	Bitmask C,V;

	if( dx == 0.0 ) {

		// edge is vertical
		int leftInside = (dy > 0.0);
		
		// set the bits in the C mask for the subregions that are completely inside
		// the edge, and set the bits in the V mask for the subregions that are 
		// completely outside the edge
		if( leftInside ) {
			C.setBits(0,0,fastDblToInt(floor(inXY[VX]))-1,maskSize-1);
			V.setBits(fastDblToInt(ceil(inXY[VX])),0,maskSize-1,maskSize-1);
		} else {
			V.setBits(0,0,fastDblToInt(floor(inXY[VX]))-1,maskSize-1);
			C.setBits(fastDblToInt(ceil(inXY[VX])),0,maskSize-1,maskSize-1);
		}

	} else if ( dy == 0.0 ) {

		// edge is horizontal
		int topInside = (dx > 0.0);

		// set the bits in the C mask for the subregions that are completely inside
		// the edge, and set the bits in the V mask for the subregions that are 
		// completely outside the edge
		if( topInside ) {
			C.setBits(0,fastDblToInt(ceil(inXY[VY])),maskSize-1,maskSize-1);
			V.setBits(0,0,maskSize-1,fastDblToInt(floor(inXY[VY]))-1);
		} else {
			V.setBits(0,fastDblToInt(ceil(inXY[VY])),maskSize-1,maskSize-1);
			C.setBits(0,0,maskSize-1,fastDblToInt(floor(inXY[VY]))-1);
		}

	} else {

		// edge is neither horizontal nor vertical, so scan convert the edge and
		// construct the mask
		double slope = dy / dx;

		vec2 low, high;
		int underCov = (dx < 0.0);
		int leftInside;

		// sort the intersection points in Y
		if( dy >= 0.0 ) {
			low = inXY;
			high = outXY;
		} else {
			low = outXY;
			high = inXY;
		}

		// find the first and last subregions partially intersected by the edge
		int firstRow = fastDblToInt(floor(low[VY]));
		int lastRow = fastDblToInt(ceil(high[VY])-1);

		if( underCov ) {

			// set the C bits for the regions that are completely below the edge
			for( int row = 0; row < firstRow; row++ ) {
				C.setBits(0,row,maskSize-1,row);
			}

			// set the V bits for the regions that are completely above the edge
			for( row = maskSize-1; row > lastRow; row-- ) {
				V.setBits(0,row,maskSize-1,row);
			}

			// determine whether regions on the left are inside or outside
			leftInside = (slope < 0.0);

		} else {

			// set the V bits for the regions that are completely below the edge
			for( int row = 0; row < firstRow; row++ ) {
				V.setBits(0,row,maskSize-1,row);
			}
			
			// set the C bits for the regions that are completely above the edge
			for( row = maskSize-1; row > lastRow; row-- ) {
				C.setBits(0,row,maskSize-1,row);
			}

			// determine whether regions on the left are inside or outside
			leftInside = (slope > 0.0);
		}

		
		// now, scan convert the actual edge

		// x increment between regions
		double xIncr = 1.0 / slope;

		// I honestly can't remember what the heck this is doing
		double xMin, xMax;
		if( slope > 0.0 ) {
			xMin = low[VX];
			xMax = xMin + xIncr * (floor(low[VY]+1.0) - low[VY]);
		} else {
			xMax = low[VX];
			xMin = xMax + xIncr * (floor(low[VY]+1.0) - low[VY]);
		}

		// loop through the rows and fill in the appropriate bits in the C and V masks
		for( int curRow = firstRow; curRow <= lastRow; curRow++ ) {
			if( leftInside ) {
				C.setBits( 0, curRow, fastDblToInt(floor(xMin)-1), curRow );
				V.setBits( fastDblToInt(ceil(xMax)), curRow, maskSize-1, curRow );
			} else {
				V.setBits( 0, curRow, fastDblToInt(floor(xMin)-1), curRow );
				C.setBits( fastDblToInt(ceil(xMax)), curRow, maskSize-1, curRow );
			}

			if( slope > 0.0 ) {
				xMin = xMax;
				xMax += xIncr;
			} else {
				xMax = xMin;
				xMin += xIncr;
			}
		}
	}

	return TriageMask(C,V);

}

//
//  Coverts a perimeter index to a 2d position
//
vec2 TriageMask::indexToPos( int index )
{
	vec2 pos;

	// how many intervals per side?
	int side = maskSize * subIntervals;

	// figure out which side the point is on, and compute the position
	switch( index/side ) {
	case 0:
		pos[VX] = index;
		pos[VY] = 0.0;
		break;
	case 1:
		pos[VX] = side;
		pos[VY] = index - side;
		break;
	case 2:
		pos[VX] = (3*side) - index;
		pos[VY] = side;
		break;
	case 3:
		pos[VX] = 0.0;
		pos[VY] = (4*side) - index;
		break;
	}

	pos /= subIntervals;

	return pos;
}

//
//  Output operator.
//
ostream& operator << ( ostream& os, TriageMask& t )
{
	for( int i = TriageMask::maskSize-1; i >= 0; i-- ) {
		for( int j = 0; j < TriageMask::maskSize; j++ ) {
			if( t.C().testBit(j,i) ) {
				os << "C";
			} else if( t.V().testBit(j,i) ) {
				os << "V";
			} else {
				os << "A";
			}
		}

		os << endl;
	}
	
	return os;
}