#include "Triangle.h"
#include "Cam.h"
#include "Sample.h"
#include "TriageMask.h"
#include "Edge.h"

// 
//  Fills a shading sample for a specified vertex
//
void Triangle::fillSample( int whichVertex, Sample& sample )
{
	sample.worldCoords = wPoints[whichVertex];
	sample.worldNormal = normals[whichVertex];
	sample.worldNormal.normalize();
	sample.texCoords2d = tPoints[whichVertex];
	sample.texCoords3d = sample.worldCoords;
}

// 
//  Fills a shading sample that is an affine combination of the
//  3 triangle vertices.
//
void Triangle::fillSample( vec3& ac, Sample& sample )
{
	sample.worldCoords = ac[0]*wPoints[0] + ac[1]*wPoints[1] + ac[2]*wPoints[2];
	sample.worldNormal = ac[0]*normals[0] + ac[1]*normals[1] + ac[2]*normals[2];
	sample.worldNormal.normalize();
	sample.texCoords2d = ac[0]*tPoints[0] + ac[1]*tPoints[1] + ac[2]*tPoints[2];
	sample.texCoords3d = sample.worldCoords;
}

//
//  Computes a triage mask for the triangle in the given screen box.  Exactly
//  like Tessellation::triageMask
//
TriageMask Triangle::triageMask( vec4 screenBox, bool correct )
{
	bool first = true;
	TriageMask tmask;

	for( int edge = 0; edge < 3; edge++ ) {
		int coords[2];
		Edge e( sPoints[edge], sPoints[(edge+1)%3] );
		TriageMask edgeMask;

		OverlapType overlap = e.boxIntersect( screenBox, coords, correct );

		switch( overlap ) {

		case ALL:
			edgeMask = TriageMask(FULL);
			break;

		case NONE:
			edgeMask = TriageMask(EMPTY);
			break;

		case SOME:
			edgeMask = TriageMask( coords[0], coords[1] );
			break;

		}

		if( first ) {
			tmask = edgeMask;
			first = false;
		} else {
			tmask += edgeMask;
		}
	}

	return tmask;
}

//
//  Computes a bitmask for the triangle in the given screen box.  Exactly
//  like Tessellation::bitMask
//
Bitmask Triangle::bitMask( vec4 screenBox, bool correct ) 
{
	bool first = true;
	Bitmask bmask;

	for( int edge = 0; edge < 3; edge++ ) {
		int coords[2];
		Edge e( sPoints[edge], sPoints[(edge+1)%3] );
		Bitmask edgeMask;

		OverlapType overlap = e.boxIntersect( screenBox, coords, correct );

		switch( overlap ) {

		case ALL:
			edgeMask = Bitmask(FULL);
			break;

		case NONE:
			edgeMask = Bitmask(EMPTY);
			break;

		case SOME:
			edgeMask = Bitmask( coords[0], coords[1] );
			break;

		}

		if( first ) {
			bmask = edgeMask;
			first = false;
		} else {
			bmask &= edgeMask;
		}
	}

	return bmask;
}

//
//  Output operator
//
ostream& operator<< (ostream& s, Triangle& t)
{
	s << endl << "Triangle {" << endl;

	s << "World:   ";
	s << t.wPoints[0] << "," << t.wPoints[1] << "," << t.wPoints[2] << endl;

	s << "Screen:  ";
	s << t.sPoints[0] << "," << t.sPoints[1] << "," << t.sPoints[2] << endl;

	s << "Texture: ";
	s << t.tPoints[0] << "," << t.tPoints[1] << "," << t.tPoints[2] << endl;

	s << "Normals: ";
	s << t.normals[0] << "," << t.normals[1] << "," << t.normals[2] << endl;

	s << endl;
	return s;
}
